// ------------------------------------------------------------------
//
//  GenericMinVar.h
//  Qdos2 - Generic Constrained Minimum Variance Algorithm for. Based on 
//  the original MinVariance.h.
//
//  Janet Barnes
//  May 2000
//
//  This version of Minimum Variance contains an additional projection
//  matrix P, which, when constrain is enabled, is projected from both 
//  sides on the covariance  matrix, thus creating a conversion matrix 
//  (eigenvecs) which  ensures that the time series used to constrain has 
//  a zero average along the direction corresponding to the normal 
//  direction. See blue ISSI book on Multispacecraft Analysis, p193. 
//
//  Cloned & modified by 
//  SEH,    May 2002 - cloned from original MinVariance.h
//  SEH,  6 Sep 2003 - Generic version version, 
//  SEH, 11 May 2005 - corrected RH correction; 
//                     
// ------------------------------------------------------------------

#include "DvObj.h"

class GenericMinVar
{
    DvObject_var _eigvals;       // Matrix Dim(3) of eigenvalues
    DvObject_var _eigvecs;       // Matrix Dim(3,3) of eigenvectors (in rows)
    DvObject_var _eigqual;       // Matrix Dim(3) eigenvalue quality
    double _orgqual;             // orthogonality quality
    DvObject_var _covariance;    // covariance matrix
    DvObject_var temp;           // temporary storage matrix
    int i,j;
    
public:
    GenericMinVar(int constrained, DvObject_var &Q, DvObject_var &P)
    {        
        _covariance = new DvObject(Q);        
		// force hermitian
        _covariance->dbl(0,1,0) = _covariance->dbl(0,0,1);
        _covariance->dbl(0,2,0) = _covariance->dbl(0,0,2);
        _covariance->dbl(0,2,1) = _covariance->dbl(0,1,2);
        
	// 
	// Constrained MVA : find eigenvec of [P # C # P] * n = lamda * n
	// instead of [C] * n = lambda * n. 
	//
	//   0 - standard MVA
	//   1 - constrain so that Bn == 0 (see p 209 in blue ISSI book)
        //       P matrix calculated here
	//   2 - constrain using given P matrix
	//
	// SEH, 14 May 2002
	//      28 May 2002, MFR version
	//      26 Mar 2003, MFL version
	//

	if (constrained >= 1) {
        temp = _covariance*P;
        _covariance = P*temp;

        // force symmetric covariance matrix again
        _covariance->dbl(0,1,0) = _covariance->dbl(0,0,1);
        _covariance->dbl(0,2,0) = _covariance->dbl(0,0,2);
        _covariance->dbl(0,2,1) = _covariance->dbl(0,1,2);
	}

        // get eigenvectors and values  
        DvMatrixEig eig(_covariance);
        _eigvecs=eig.eigenvectors();
        _eigvals=eig.eigenvalues();


	// If constrained : eigval 0 == 0. Swap 1st and 2nd eigenvec

        if (constrained >= 1) {
          _eigvals->dbl(0) = _eigvals->dbl(1);
          _eigvals->dbl(1) = _eigvals->dbl(2);
          _eigvals->dbl(2) = 0;                                // by definition
          temp = new DvObject(_eigvecs);
          for (i=0; i<3; i++) _eigvecs->dbl(0,0,i) = temp->dbl(0,1,i);
          for (i=0; i<3; i++) _eigvecs->dbl(0,1,i) = temp->dbl(0,2,i);
          for (i=0; i<3; i++) _eigvecs->dbl(0,2,i) = temp->dbl(0,0,i);
        }

        
        // check the positive sense of the eigenvector associated with
        // the minimum eigenvalue and if necessary reverse the signs
        // of all three components i.e. vint x vmax = vector perpendicular
        // to both vint and vmax. If vmin is perpendicular to vint 
        // and vmax (i.e. system is ortogonal), then (vinx x vmax).vmin=1
        // If < 0 then not a right-handed system, so reverse signs of vmin
        
        // get vectors out
        DvObject_var eigV0 = _eigvecs->sliceDim(0,0);
        DvObject_var eigV1 = _eigvecs->sliceDim(0,1);
        DvObject_var eigV2 = _eigvecs->sliceDim(0,2);
        
        eigV0->change_xref(FRAME, "vector>mv_xyz");
        eigV1->change_xref(FRAME, "vector>mv_xyz");
        eigV2->change_xref(FRAME, "vector>mv_xyz");
        
        DvObject_var rh = eigV1->vec(eigV2);
        rh = rh->dot(eigV0);
        
        double rh_factor = rh->dbl();
        if (rh_factor < 0.) *eigV0 *= -1.0;
        _orgqual = fabs(fabs(rh_factor) - 1.0);

                
        // ensure that the sign of the average minimum variance component
        // is positive by computing average of sum of B-field data.min
        // eigenvector. If less than zero negate both vmin and vint (so
        // preserving right-handed system)


        // SEH 11 May 2005 : this was already commented out; a positive normal comp
        // does not make sense for us.

        _eigqual = new DvObject(_eigvals);
        *_eigqual /= _eigvals->dbl(2);
     
        // SEH 2003_03_26: for MFL, MFR, pr definition, the eigenvector describing
        // the normal component should have a positive X-value; If not,
        // flip the eigenvector matrix, conserve RH system

        // SEH 11 May 2005 : Correction; to preserve RH sys, only two eigvals should be flipped.
        if (eigV0->dbl(0) < 0.0){
            *eigV0 *= -1.0;
            *eigV1 *= -1.0;
        }

       
        // put vectors back
        slice ds(0, eigV0->totalSize(), 1); // array_slice of all data in vector
        _eigvecs->setSliceDim(0, 0, eigV0->getDSlice(ds));
        _eigvecs->setSliceDim(0, 1, eigV1->getDSlice(ds));
        _eigvecs->setSliceDim(0, 2, eigV2->getDSlice(ds));

        
    }



    // methods for returning members and calculating variance
    DvObject_var covariance()
    { return _covariance; }
    
    DvObject_var eigenvalues()
    { return _eigvals;}
     
    DvObject_var eigenvectors()
    { return _eigvecs;}   
    
    DvObject_var eigenquality()
    { return _eigqual;}
    
    double& orthogonality_quality()
    { return _orgqual;}    // orthogonality quality     

        
};
    
