#include "qcalcCalcItem.h"
#include "qcalcSlots.h"
#include "qcalcScene.h"


VCalcItem::VCalcItem()
{
	inputs.clear();
	outputs.clear();
	name="";
    container = NULL;
	
	calc_id=calc_counter++;
	CalcScene::currentItems.push_back(this);
	isBlackBox=false;

	temp_id=-1;
	selected=false;
	positionInBlackBox=QPointF(0.,0.);
	page=CalcScene::currentPage;
}

VCalcItem::~VCalcItem()
{
	CalcScene::currentItems.removeAll(this);

	if(!isBlackBox)
	{
		qDeleteAll(inputs);
		qDeleteAll(outputs);
	}
	inputs.clear();
	outputs.clear();
}

QList<VCalcItem*> VCalcItem::connectedItems(bool parents, bool children) const
{
	QList<VCalcItem*> res;
	res.clear();
	if(parents)
	{
		VCalcItem* temp;
		foreach(VInSlot* in, inputs)
		{
			temp=in->connectedItem();
			if(!res.contains(temp) && temp!=NULL) res.push_back(temp);
		}
	}
	
	if(children)
	{
		foreach(VOutSlot* out, outputs)
		{
			foreach(VCalcItem* item, out->connectedItems())
			{
				if(!res.contains(item)) res.push_back(item);
			}
		}
	}
	
	return res;
}

bool VCalcItem::isParentOf(VCalcItem* i) const
{
	if(i==this) return true;
	
	foreach(VOutSlot* out, outputs)
	{
		QList<VCalcItem*> items=out->connectedItems();
		foreach(VCalcItem* item, items)
		{
			if(item->isParentOf(i))
				return true;
		}
	}
	
	return false;
}

// bool VCalcItem::setConnection(int outNumber, VCalcItem* target, int inNumber)
// {
// 	if(outNumber>=outputs.size() || outNumber<0) return false;
// 	
// 	if(inNumber>=target->inputs.size() || inNumber<0) return false;
// 	
// 	if(target->isParentOf(this)) return false;
// 	
// 	VOutSlot* from=outputs[outNumber];
// 	VInSlot* to=target->inputs[inNumber];
// 	from->connectTo(to);
// 	return true;
// }

void VCalcItem::flagOutSlotsForReevaluation()
{
	// tells the output slots they will need reevaluation
	foreach(VOutSlot* out, outputs)
	{
		out->setReevaluationFlag(true);
	}
	// spread to the (children) connected items
	QList<VCalcItem*> items=connectedItems();
	foreach(VCalcItem* item, items)
	{
		item->flagOutSlotsForReevaluation();
	}
}

void VCalcItem::fromXMLElement(QDomElement& el, QGraphicsScene* scene)
{
	QString val;
	if(getXMLNodeText(el, "ID", val))
	{
		temp_id=val.toInt();
	}
	return;
}


CalcItem::CalcItem(QGraphicsItem* parent):QGraphicsRectItem(parent)
{
	// Buttons
	rzButton=new RotoZoomButton(this);
	closeButton=new CloseButton(this);
	errButton=new ErrorButton(this);
	errButton->setVisible(false);

	// Qt is the one doing the job to move and select items
	setFlag(QGraphicsItem::ItemIsMovable, true);
	setFlag(QGraphicsItem::ItemIsSelectable, true);
	
	inSlots.clear();
	outSlots.clear();

#if QT_VERSION >= 0x040600
	this->setFlag(QGraphicsItem::ItemSendsGeometryChanges, true);
#endif
	
}

// Update the connectors positions when the object is moved
QVariant CalcItem::itemChange(GraphicsItemChange change, const QVariant& value)
{
	if (change == ItemPositionChange && scene()) {
		QPointF newPos = value.toPointF();
		
		updateConnectorsPositions();
		
		return newPos;
	}
	else if(change==ItemSelectedChange && scene())
	{
		if(value.toBool())
		{
			setZValue(1.);
		}
		else setZValue(0.);
	}
	return QGraphicsItem::itemChange(change, value);
}

void CalcItem::updateConnectorsPositions()
{
	foreach(InSlot* slot, inSlots)
	{
		slot->updateConnectorPosition();
	}
	foreach(OutSlot* slot, outSlots)
	{
		slot->updateConnectorsPositions();
	}
}

void CalcItem::paint(QPainter* painter, const QStyleOptionGraphicsItem* , QWidget* )
{
	painter->setRenderHints(QPainter::Antialiasing);
	QPen pen(Qt::SolidLine);
	pen.setWidthF(1.);
	pen.setColor(outlineColor);
	painter->setPen(pen);
	
	QLinearGradient gradient(QPointF(0., -height/2.), QPointF(0., -height/2.+HEADER_HEIGHT));
	gradient.setColorAt(0., color1);
	gradient.setColorAt(1., color2);
	QBrush brush(gradient);
	painter->setBrush(brush);
	painter->drawRoundRect(boundingRect(), 5, 5);
	
	QRectF box(QPointF(-width/2.+2.*HEADER_HEIGHT, -height/2.), QPointF(width/2.-2.*HEADER_HEIGHT, -height/2.+HEADER_HEIGHT));
	painter->drawText(box, Qt::AlignCenter, title);
	
	// dashed outer box
	if(isSelected())
	{
		pen.setColor(Qt::black);
		pen.setStyle(Qt::DashLine);
		pen.setWidthF(2.);
		painter->setPen(pen);
		brush.setStyle(Qt::NoBrush);
		painter->setBrush(brush);
		QRectF r=boundingRect();
		r.adjust(1., 1., -1., -1.);
		painter->drawRoundRect(r, 5, 5);
	}
}

void CalcItem::setErrMessage(bool err, const QString& message)
{
	errButton->setVisible(err);
	if(err)
	{
		errButton->setToolTip(message);
	}
	else{
		errButton->setToolTip(" ");
	}
}

QDomElement CalcItem::toXMLElement(QDomDocument& document) const
{
	QDomElement tag=document.createElement("GUI");
    QRectF bounding=boundingRect();

    QDomElement pos_x=document.createElement("Pos_x");
    pos_x.appendChild(document.createTextNode(QString("%1").arg(scenePos().x())));
    tag.appendChild(pos_x);

    QDomElement pos_y=document.createElement("Pos_y");
    pos_y.appendChild(document.createTextNode(QString("%1").arg(scenePos().y())));
	tag.appendChild(pos_y);
    
	// Save the transformation matrix parameters
    QTransform t=transform();
    if(t!=QTransform())
    {
        QDomElement trans1=document.createElement("Transform_1");
        trans1.appendChild(document.createTextNode(QString("%1").arg(t.m11())));
        tag.appendChild(trans1);

        QDomElement trans2=document.createElement("Transform_2");
        trans2.appendChild(document.createTextNode(QString("%1").arg(t.m12())));
        tag.appendChild(trans2);
    }
	return tag;
}

void CalcItem::fromXMLElement(QDomElement& el)
{
	QString pos_x="0";
    QString pos_y="0";
    QString trans1="1.";
    QString trans2="0.";
    
    QDomElement pos_x_el=el.firstChildElement("Pos_x");
    if(!pos_x_el.isNull())
    {
        pos_x=pos_x_el.firstChild().nodeValue();
    }
    QDomElement pos_y_el=el.firstChildElement("Pos_y");
    if(!pos_y_el.isNull())
    {
        pos_y=pos_y_el.firstChild().nodeValue();
    }

    QDomElement trans1_el=el.firstChildElement("Transform_1");
    if(!trans1_el.isNull())
    {
        trans1=trans1_el.firstChild().nodeValue();
    }
    QDomElement trans2_el=el.firstChildElement("Transform_2");
    if(!trans2_el.isNull())
    {
        trans2=trans2_el.firstChild().nodeValue();
    }

    if(trans1!="1." || trans2!="0.")
    {
        QTransform t(trans1.toDouble(), trans2.toDouble(), 0., -trans2.toDouble(), trans1.toDouble(), 0., 0., 0., 1.);
                     
        setTransform(t);
    }

    setPos(pos_x.toDouble(), pos_y.toDouble());
}

