#include "qcalcScene.h"
#include "qcalcSlots.h"
#include "qcalcCalcItem.h"
#include "qcalcInputItem.h"
#include "qcalcOperatorItem.h"
#include "qcalcOutputItem.h"
#include "qcalcExtractionItem.h"
#include <cmath>
#include "qcalcWindow.h"
#include <qtextedit.h>
#include <qtransform.h>
#include "qdutil.h"
#include "DvObj.h"
#include "DvRecord.h"

using namespace QSAS;


int CalcScene::currentPage;
QList<VCalcItem*> CalcScene::currentItems; // all the items in the calculator
extern TimeLineHandler timeLineHandler;

CalcScene::CalcScene(QObject* parent):
	QGraphicsScene(parent)
{	
	currentConnector=NULL;
	movingConnector=false;
	currentItem=NULL;
	rotoZooming=false;
	draggingSelectionRect=false;
	selectionRect=new QGraphicsRectItem();
    this->addItem(selectionRect);
	selectionRect->setVisible(false);
	QPen pen(Qt::DotLine);
	tracking1=new QGraphicsLineItem();
    this->addItem(tracking1);
	tracking1->setPen(pen);
	tracking1->setVisible(false);
	tracking2=new QGraphicsLineItem();
    this->addItem(tracking2);
	tracking2->setPen(pen);
	tracking2->setVisible(false);
	tracking=false;
	trackingSuccess=false;
	
	currentPage=1;
	
	addPage();
	setTimeLine("Automatic");
	
}

CalcScene::~CalcScene()
{
	pages.clear();
	QplugSetTextDisplay((QTextEdit*)NULL);
	delete selectionRect;
}

void CalcScene::setTracking()
{
	tracking=true;
	trackingSuccess=false;
	emit trackingStarted();
}

void CalcScene::dragEnterEvent(QGraphicsSceneDragDropEvent * event)
{
	event->setAccepted(event->mimeData()->hasText());
}

void CalcScene::dragMoveEvent(QGraphicsSceneDragDropEvent * event)
{
	event->setAccepted(event->mimeData()->hasText());
}

void CalcScene::updateVisibleItems(bool restore)
{
	foreach(QGraphicsItem* item, items())
	{
			InputItem* in=qgraphicsitem_cast<InputItem*>(item);
			OutputItem* out=qgraphicsitem_cast<OutputItem*>(item);
			ExtractionItem* ext=qgraphicsitem_cast<ExtractionItem*>(item);
			OperatorItem* op=qgraphicsitem_cast<OperatorItem*>(item);
			Connector* conn=qgraphicsitem_cast<Connector*>(item);
			JoinItem* join=qgraphicsitem_cast<JoinItem*>(item);
			if(in!=NULL) {
				in->setVisible(in->page()==currentPage);
				if(restore) connect(in, SIGNAL(newInput(const QString&)), this, SIGNAL(newInput(const QString&)));
			}
			if(out!=NULL) out->setVisible(out->page()==currentPage);
			if(ext!=NULL) ext->setVisible(ext->page()==currentPage);
			if(op!=NULL) op->setVisible(op->page()==currentPage);
			if(conn!=NULL) conn->setVisible(conn->page==currentPage);
			if(join!=NULL) join->setVisible(join->page()==currentPage);
	}
	
}

int CalcScene::hasJoinItems(int i)
{
	foreach(QGraphicsItem* item, items())
	{
		JoinItem* join=qgraphicsitem_cast<JoinItem*>(item);
		if(join!=NULL && join->page()==i) return 1;
		
		OperatorItem* op=qgraphicsitem_cast<OperatorItem*>(item);
		if(op!=NULL && op->page()==i ) {
			VOperatorItem * vop = op->parentItem;
			if(vop->isBlackBox){
				QList<VCalcItem*> BBoxList = vop->getBBList();
				foreach(VCalcItem* bbItem, BBoxList){
					VJoinItem* BBjoin=dynamic_cast<VJoinItem*>(bbItem);
					if(BBjoin) return 2;
				}
			}
		}
		
	}
	return 0;
}

void CalcScene::dropEvent(QGraphicsSceneDragDropEvent * event)
{
// 	std::cout << "CalcScene::dropEvent " << event->mimeData()->hasText() << std::endl;
	event->setAccepted(event->mimeData()->hasText());
	QPointF point=event->scenePos();
	QPointF nearpoint=point + QPoint(10,10);
	
	// Set item selected so newInput() propagates
	QRectF rect(point, nearpoint);
	QGraphicsRectItem itemRect;
	itemRect.setRect(rect);
	setSelectionArea(itemRect.shape());

	if(event->mimeData()->hasFormat("application/wlitems"))
	{
		InputItem* inItem=NULL;
		
		QList<QGraphicsItem*> pointedItems=items(point);
		InSlot* inslot=NULL;
		for(QList<QGraphicsItem*>::iterator iter=pointedItems.begin(); iter!=pointedItems.end() && inItem==NULL; ++iter)
		{
			inItem=qgraphicsitem_cast<InputItem*>(*iter);
		}
		
		QStringList items=QString(event->mimeData()->data("application/wlitems").data()).split("//");

		if(inItem!=NULL)
		{
			inItem->setValue(items[0]);
		}
		else
		{
			foreach(QString itemString, items)
			{
				VInputItem* item=new VInputItem;
				InputItem* item_gui=item->getGUI();
				addItem(item_gui);
				connect(item_gui, SIGNAL(newInput(const QString&)), this, SIGNAL(newInput(const QString&)));
				item_gui->setPos(point);
				point=point+QPointF(20,20);
				item_gui->setValue(itemString);
			}
		}
	}
	else if(event->mimeData()->hasText())
	{
		InputItem* inItem=NULL;
		QList<QGraphicsItem*> pointedItems=items(event->scenePos());
		InSlot* inslot=NULL;
		for(QList<QGraphicsItem*>::iterator iter=pointedItems.begin(); iter!=pointedItems.end() && inItem==NULL; ++iter)
		{
			inItem=qgraphicsitem_cast<InputItem*>(*iter);
		}
		
		if(inItem!=NULL)
		{
			inItem->setValue(event->mimeData()->text());
		}
		else if(pointedItems.isEmpty())
		{
			VInputItem* item=new VInputItem;
			InputItem* item_gui=item->getGUI();
			addItem(item_gui);
			connect(item_gui, SIGNAL(newInput(const QString&)), this, SIGNAL(newInput(const QString&)));
			item_gui->setPos(event->scenePos());
			item_gui->setValue(event->mimeData()->text());
		}
	}

	// reset item selected to update tabs
	setSelectionArea(itemRect.shape());

}

void CalcScene::keyPressEvent(QKeyEvent* keyEvent)
{
	if(tracking && keyEvent->key()==Qt::Key_Escape)
	{
		trackingSuccess=false;
		tracking=false;
		tracking1->setVisible(false);
		tracking2->setVisible(false);
		emit trackingEnded();
		return;
	}	
	QGraphicsScene::keyPressEvent(keyEvent);
}

void CalcScene::escPressed()
{
	if(tracking)
	{
		trackingSuccess=false;
		tracking=false;
		tracking1->setVisible(false);
		tracking2->setVisible(false);
		emit trackingEnded();
		return;
	}
}

void CalcScene::mouseDoubleClickEvent(QGraphicsSceneMouseEvent* mouseEvent)
{
	qApp->processEvents();
	emit hadDoubleClick();
}
	
void CalcScene::mousePressEvent(QGraphicsSceneMouseEvent* mouseEvent)
{	
	if(mouseEvent->button()!=Qt::LeftButton)
	{

		QGraphicsScene::mousePressEvent(mouseEvent);
		return;
	}
	if(tracking)
	{
		trackingPosition=mouseEvent->scenePos();
		trackingSuccess=true;
		tracking=false;
		tracking1->setVisible(false);
		tracking2->setVisible(false);
		emit trackingEnded();
		return;
	}
	
#if QT_VERSION >= 0x040600
	QGraphicsItem* pointedItem=itemAt(mouseEvent->scenePos(), QTransform());
#else
	QGraphicsItem* pointedItem=itemAt(mouseEvent->scenePos());
#endif
	
	if(pointedItem!=NULL && !movingConnector && !rotoZooming)
	{
		InSlot* inslot=qgraphicsitem_cast<InSlot*>(pointedItem);
		OutSlot* outslot=qgraphicsitem_cast<OutSlot*>(pointedItem);
		RotoZoomButton* rzButton=qgraphicsitem_cast<RotoZoomButton*>(pointedItem);
		ExpandButton* expandButton=qgraphicsitem_cast<ExpandButton*>(pointedItem);
		CloseButton* closeButton=qgraphicsitem_cast<CloseButton*>(pointedItem);
		
		if(inslot!=NULL)
		{
			// Move end
			if(inslot->getConnector()!=NULL)
			{
				currentConnector=inslot->getConnector();
				movingConnector=true;
				currentConnector->setEnd(mouseEvent->scenePos());
				return;
			}
		}
		else if(outslot!=NULL)
		{
			// Create connector
			currentConnector=new Connector(outslot, NULL, mouseEvent->scenePos());
			addItem(currentConnector);
			movingConnector=true;
			return;
		}
		else if(rzButton!=NULL)
		{
			rotoZooming=true;
			currentItem=rzButton->getItem();
			reference=mouseEvent->scenePos();
			referenceMatrix=currentItem->matrix();
			currentItem->setSelected(true);
		}
		else if(expandButton!=NULL)
		{
			OperatorItem* item=(OperatorItem*)(expandButton->getItem());
			if(!item->isExpanded()) item->expand();
			else item->collapse();
			item->setSelected(true);
		}
		else if(closeButton!=NULL)
		{
			currentItem=closeButton->getItem();
			
			closing=true;
		}
		
	}
	else // start of selection
	{
		draggingSelectionRect=true;
		reference=mouseEvent->scenePos();
		selectionRect->setRect(QRectF(reference, reference));
		selectionRect->setVisible(true);
	}
	QGraphicsScene::mousePressEvent(mouseEvent);
}

void CalcScene::mouseMoveEvent(QGraphicsSceneMouseEvent* mouseEvent)
{
	if(tracking)
	{
		if(!tracking1->isVisible())
		{
			tracking1->setVisible(true);
			tracking2->setVisible(true);
		}
		QRectF rect=sceneRect();
		tracking1->setLine(rect.left(), mouseEvent->scenePos().y(), rect.right(), mouseEvent->scenePos().y());
		tracking2->setLine(mouseEvent->scenePos().x(),rect.bottom(), mouseEvent->scenePos().x(),rect.top());
	}
	else if(movingConnector)
	{
		if(currentConnector!=NULL)
		{
			currentConnector->setEnd(mouseEvent->scenePos());
		}
		else
		{
			movingConnector=false;
		}
	}
	else if(rotoZooming)
	{
		QPointF centre=currentItem->scenePos();
		double refx(reference.x()-centre.x()), refy(reference.y()-centre.y());
		double curx(mouseEvent->scenePos().x()-centre.x()), cury(mouseEvent->scenePos().y()-centre.y());
		double norm1=sqrt(refx*refx+refy*refy);
		double norm2=sqrt(curx*curx+cury*cury);

		double refxn=refx/norm1;
		double refyn=refy/norm1;
		double curxn=curx/norm2;
		double curyn=cury/norm2;
		
		double angle=acos(refxn*curxn+refyn*curyn);
		if(asin(refxn*curyn-refyn*curxn)<0.) angle=-angle;
		
		QMatrix m(referenceMatrix);
		m.rotate(angle/M_PI*180.);
		m.scale(norm2/norm1, norm2/norm1);
		currentItem->setMatrix(m);
		currentItem->updateConnectorsPositions();
		return;
		
	}
	else if(draggingSelectionRect)
	{
		QRectF rect(reference, mouseEvent->scenePos());
		selectionRect->setRect(rect);
		setSelectionArea(selectionRect->shape());
	}

	QGraphicsScene::mouseMoveEvent(mouseEvent);
}

void CalcScene::mouseReleaseEvent(QGraphicsSceneMouseEvent* mouseEvent)
{
	if(tracking) return;

	if(mouseEvent->button()==Qt::LeftButton && movingConnector)
	{
		QList<QGraphicsItem*> pointedItems=items(mouseEvent->scenePos());
		InSlot* inslot=NULL;
		for(QList<QGraphicsItem*>::iterator iter=pointedItems.begin(); iter!=pointedItems.end() && inslot==NULL; ++iter)
		{
			inslot=qgraphicsitem_cast<InSlot*>(*iter);
		}
		
		if(inslot!=NULL)
		{
			if(!inslot->slot->parent()->isParentOf(currentConnector->getFromSlot()->slot->parent()))
			{
				currentConnector->setEnd(inslot);
			}
			else
			{
				delete currentConnector;
			}
		}
		else
		{
			delete currentConnector;
		}
		movingConnector=false;
		currentConnector=NULL;
	}
	else if(mouseEvent->button()==Qt::LeftButton && rotoZooming && currentItem!=NULL)
	{
		rotoZooming=false;
		currentItem=NULL;
	}
	else if(mouseEvent->button()==Qt::LeftButton && closing && currentItem!=NULL)
	{
#if QT_VERSION >= 0x040600
		QGraphicsItem* pointedItem=itemAt(mouseEvent->scenePos(), QTransform());
#else
		QGraphicsItem* pointedItem=itemAt(mouseEvent->scenePos());
#endif

		CloseButton* closeButton=qgraphicsitem_cast<CloseButton*>(pointedItem);
		if(closeButton!=NULL)
		{
			if(closeButton->getItem()==currentItem)
			{
				OperatorItem* opItem=qgraphicsitem_cast<OperatorItem*>(currentItem);
				InputItem* inItem=qgraphicsitem_cast<InputItem*>(currentItem);
				OutputItem* outItem=qgraphicsitem_cast<OutputItem*>(currentItem);
                ExtractionItem* extItem=qgraphicsitem_cast<ExtractionItem*>(currentItem);
				JoinItem* joinItem=qgraphicsitem_cast<JoinItem*>(currentItem);
				if(opItem!=NULL)
				{
					VOperatorItem* vopItem=opItem->parentItem;
					vopItem->removeGUI();
					delete vopItem;
				}
				else if(inItem!=NULL)
				{
					VInputItem* vinItem=inItem->parentItem;
					vinItem->removeGUI();
					delete vinItem;
				}
				else if(outItem!=NULL)
				{
					VOutputItem* voutItem=outItem->parentItem;
					voutItem->removeGUI();
					delete voutItem;
				}
                else if(extItem!=NULL)
                {
                    VExtractionItem* vextItem=extItem->parentItem;
                    vextItem->removeGUI();
                    delete vextItem;
                }
				else if(joinItem!=NULL)
                {
                    VJoinItem* vjoinItem=joinItem->parentItem;
                    vjoinItem->removeGUI();
                    delete vjoinItem;
                }
				clearSelection(); // tell scene to deselect (not always up to date after delete)
				emit refreshContextWidget();
			}
		}
		closing=false;
		currentItem=NULL;
	}
	else if(draggingSelectionRect)
	{
		draggingSelectionRect=false;
		selectionRect->setVisible(false);
		emit selectionChanged();
	}

	QGraphicsScene::mouseReleaseEvent(mouseEvent);

}

void CalcScene::setTimeLine(const QString &name){
	if(currentPage <= pages.size()) pages[currentPage-1]->timeLineName = name;
}


QString CalcScene::evaluatePage(int i)
{
	if (i == 0 || i > pages.size() ) i = currentPage; // default
	int hasJoin = hasJoinItems(i);

	if( hasJoin != 0 && pages[i-1]->timeLineName == "Automatic"){
		
		if(hasJoin == 1) return QString("Join object found: Timeline must be selected");
		else return QString("Join object in Black Box: Timeline must be selected");
	}
	
	
	vector <VInputItem*> inputs;
	vector <VJoinItem*> joins;
	
	foreach(VCalcItem* item, currentItems)
	{
		VInputItem* in=dynamic_cast<VInputItem*>(item);
		if(in!=NULL && in->page==i ) {
			inputs.push_back(in);
			// setNoJoin also flags for reevaluation
			if(pages[i-1]->timeLineName == "Automatic") in->setNoJoin(true);
			else in->setNoJoin(false); 
		}
		
		VJoinItem* join=dynamic_cast<VJoinItem*>(item);
		if(join!=NULL && join->page==i){
			joins.push_back(join);
			join->setNoJoin(false); // also flags for reevaluation
		}
	}
	if(pages[i-1]->timeLineName != "Automatic"){
		// get time line for joining
		TimeLine* tl=timeLineHandler.getTimeLine(pages[i-1]->timeLineName);

		if(tl==NULL) tl=timeLineHandler.getDefaultTimeLine();

	
		DvObject_var tags = getPageTags(inputs, tl);
        
        if(tags.is_nil() || tags->seqSize() < 1){
            QplugAppendTextDisplay( "     WARNING: No time tags in data range");
            if(tags->nError() > 0){
                for(size_t n=0; n<tags->nError(); n++) QplugAppendTextDisplay(tags->getError(n).c_str());
            }
        }
        
		for(int n=0; n<inputs.size(); n++)
		{
			inputs[n]->setTags(tags);		
		}
		
		for(int n=0; n<joins.size(); n++)
		{
			joins[n]->setTags(tags);
		}
	}
	
	bool failed = false;
	foreach(VCalcItem* item, currentItems)
	{
		VOutputItem* out=dynamic_cast<VOutputItem*>(item);
		if(out!=NULL)
		{
			if(out->page==i)
			{
				failed += out->computeValue();
			}
		}
	}
	if(!failed) QwuiClearTemp();
	
	return QString("");
}

DvObject_var CalcScene::getPageTags(vector <VInputItem*> &inputs, TimeLine* tl)
{
    DvObject_var tags; 
    double Spacing;
    double Interval;
		
    DvObject_var ivl = getPageIvl(inputs, tl);
	if(ivl.is_nil()) return tags;

	if( tl->regularTT )
	{
		// create regular tags
		if(ivl->is_event()) Interval = ivl->event().duration();
		else  Interval = ivl->asDouble(ivl->seqSize()-1) - ivl->asDouble(0);
		
		bool OK;
		Spacing = tl->TTString.toDouble(&OK);
		
		if(!OK || fabs(Spacing) < 1.e-20) return tags;

		if(Interval < 0) Spacing = -fabs(Spacing); // handle decreasing sequence
			
		int Ntags = (int) (fabs(Interval/Spacing) + 1.0);

		if(ivl->is_event()) tags = new DvObject(ivl->event().start(), (size_t) Ntags);
		else tags = new DvObject(ivl.ptr(), (size_t) Ntags);
		tags->assign_regular(Spacing);

	}
	else{
		// from object
		DvObject_var tag_obj = getObjectAndComp(tl->TTString.toLatin1().data());
		DvObject_var D0 = tag_obj->getDep0();

		Interval = D0->asDouble(0) - D0->asDouble(ivl->seqSize()-1);
		if(D0->is_time()){
			DvEvent tRange = ivl->getTimeRange();
			tags = D0->subSequence(tRange);
		}
		else{
			double min, max;
			int start, end;
   			ivl->getDataRange(min, max); // get interval min, max
   			D0->getRecBounds(start, end, min, max); // get enclosing recs
			tags = D0->subSequence(start, end);
		}
	}
    return tags;
}

DvObject_var CalcScene::getPageIvl(vector <VInputItem*> &inputs, TimeLine* tl)
{
    DvObject_var ivl; // event object or two scalars
	
    if(tl->fromData)
    {
 		// interval from data
        ivl = getPageIvlFromData(inputs, tl);
    }
    else
    {
    	// interval from object
        DvObject_var ivl_obj = getObjectAndComp(tl->fromObjString.toLatin1().data());
        if ( ivl_obj.is_ok() ) {
            if(ivl_obj->is_event()){
                ivl = ivl_obj;
            }
			else if(ivl_obj->is_time()){
				DvEvent tRange = ivl_obj->getTimeRange();
                ivl = new DvObject(tRange);
            }
			else{
				DvObject_var D0 = ivl_obj->getDep0();
				if(D0->is_ok()){
                	if(D0->is_time()){
               	 		ivl = new DvObject( D0->getTimeRange() );
            		}
					else{
						double min, max;
						if(D0->getDataRange(min, max))
						{
							ivl = new DvObject(min, 2);
							ivl->dbl(1) = max;
						}
						// else returns is nil
					}
				}
				else{
					double min, max;
					if(ivl_obj->getDataRange(min, max))
					{
						ivl = new DvObject(min, 2);
						ivl->dbl(1) = max;
					}
					// else returns is nil
				}
			}
        }

    }
    return ivl; 
}

DvObject_var CalcScene::getPageIvlFromData(vector <VInputItem*> &inputs, TimeLine* tl)
{
    DvEvent ivl;
    DvEvent ivlIntersect;
	double min=1.e30, max=-1.e30;
	double mn, mx;
	bool ttFound = false;
	
    for(int i=0; i<inputs.size(); i++)
    {
		
        DvObject_var obj = inputs[i]->getNotJoinedValue();
        DvObject_var ivl_obj = obj->getDep0();
		if(ivl_obj->is_time()){
        	if ( ivl_obj.is_ok())
       	 	{
            	ivl = ivl_obj->getTimeRange();

        		if(!ttFound) {
					ivlIntersect = ivl;
					ttFound = true;
				}
        		else
        		{
					ivlIntersect = ivlIntersect.event_intersect(ivl);
        		}
        	}
		}
		else{
			if( ivl_obj->getDataRange(mn, mx) )
        	{
				if(mn < min) min = mn;
				if(mx > max) max = mx;
			}
		}
    }
	DvObject_var ret;
  	if( ttFound )  {
		ret = new DvObject(ivlIntersect); // found time tags
	}
	else {
		ret = new DvObject(min, 2);
		ret->dbl(1) = max;
	}
    return ret;
}


bool CalcScene::saveAs(const QString& path, bool onlySelectedItems)
{
    ofstream file;
	QString actualPath(path);
	if(!actualPath.endsWith(".qx_calc")) actualPath.append(".qx_calc");
    file.open(actualPath.toLatin1().data(),ios::out);

    if(file.is_open())
    {
		QString XMLString=toXML(onlySelectedItems).toString();
        file << XMLString.toLatin1().data();
        file.close();
		return true;
    }
	return false;
}

bool CalcScene::exportAs(const QString& path)
{
	saveAs(path, true);
	return true;
}

void CalcScene::open(const QString& path, bool restore)
{
    QDomDocument doc;
    QFile myFile(path);
    doc.setContent(&myFile);
    myFile.close();

	fromXML(doc, restore);
}

void CalcScene::selectAll()
{
	foreach(QGraphicsItem* item, items())
	{
		OperatorItem* opItem=qgraphicsitem_cast<OperatorItem*>(item);
		InputItem* inItem=qgraphicsitem_cast<InputItem*>(item);
		OutputItem* outItem=qgraphicsitem_cast<OutputItem*>(item);
        ExtractionItem* extItem=qgraphicsitem_cast<ExtractionItem*>(item);
		JoinItem* joinItem=qgraphicsitem_cast<JoinItem*>(item);
		if(opItem!=NULL) opItem->setSelected(opItem->page()==currentPage);
		if(inItem!=NULL) inItem->setSelected(inItem->page()==currentPage);
		if(outItem!=NULL) outItem->setSelected(outItem->page()==currentPage);
		if(extItem!=NULL) extItem->setSelected(extItem->page()==currentPage);
		if(joinItem!=NULL) joinItem->setSelected(joinItem->page()==currentPage);
	}
}

void CalcScene::clearPage()
{
	QList<CalcItem*> itemsToDelete;
	foreach(QGraphicsItem* item, items()) // First, accumulate objects to delete, because deleting one CalcItem deletes various QGraphItems from items() (e.g. close buttons)
	{
		OperatorItem* opItem=qgraphicsitem_cast<OperatorItem*>(item);
		InputItem* inItem=qgraphicsitem_cast<InputItem*>(item);
		OutputItem* outItem=qgraphicsitem_cast<OutputItem*>(item);
        ExtractionItem* extItem=qgraphicsitem_cast<ExtractionItem*>(item);
		JoinItem* joinItem=qgraphicsitem_cast<JoinItem*>(item);
		if(opItem!=NULL)
		{
			if(opItem->page()==currentPage)
			{
				itemsToDelete.push_back(opItem);
			}
		}
		else if(inItem!=NULL)
		{
			if(inItem->page()==currentPage)
			{
				itemsToDelete.push_back(inItem);
			}
		}
		else if(outItem!=NULL)
		{
			if(outItem->page()==currentPage)
			{
				itemsToDelete.push_back(outItem);
			}
		}
        else if(extItem!=NULL)
        {
			if(extItem->page()==currentPage)
			{
				itemsToDelete.push_back(extItem);
			}
		}
		else if(joinItem!=NULL)
        {
			if(joinItem->page()==currentPage)
			{
				itemsToDelete.push_back(joinItem);
			}
		}
	}
	
	// Now delete
	foreach(CalcItem* item, itemsToDelete)
	{
		OperatorItem* opItem=qgraphicsitem_cast<OperatorItem*>(item);
		InputItem* inItem=qgraphicsitem_cast<InputItem*>(item);
		OutputItem* outItem=qgraphicsitem_cast<OutputItem*>(item);
        ExtractionItem* extItem=qgraphicsitem_cast<ExtractionItem*>(item);
		JoinItem* joinItem=qgraphicsitem_cast<JoinItem*>(item);
		if(opItem!=NULL)
		{
			VOperatorItem* vopItem=opItem->parentItem;
			vopItem->removeGUI();
			delete vopItem;
		}
		else if(inItem!=NULL)
		{
			VInputItem* vinItem=inItem->parentItem;
			vinItem->removeGUI();
			delete vinItem;
		}
		else if(outItem!=NULL)
		{
			VOutputItem* voutItem=outItem->parentItem;
			voutItem->removeGUI();
			delete voutItem;
		}
        else if(extItem!=NULL)
        {
			VExtractionItem* vextItem=extItem->parentItem;
            vextItem->removeGUI();
            delete vextItem;
        
		}
		else if(joinItem!=NULL)
        {
			VJoinItem* vjoinItem=joinItem->parentItem;
            vjoinItem->removeGUI();
            delete vjoinItem;
        
		}

	}
	clearSelection();
	emit refreshContextWidget();
	
}

void CalcScene::clearAll()
{

	emit refreshContextWidget(); // stop deleted items refreshing
	qApp->processEvents();

	QList<CalcItem*> itemsToDelete;
	foreach(QGraphicsItem* item, items()) // First, accumulate objects to delete, because deleting one CalcItem deletes various QGraphItems from items() (e.g. close buttons)
	{
		OperatorItem* opItem=qgraphicsitem_cast<OperatorItem*>(item);
		InputItem* inItem=qgraphicsitem_cast<InputItem*>(item);
		OutputItem* outItem=qgraphicsitem_cast<OutputItem*>(item);
        ExtractionItem* extItem=qgraphicsitem_cast<ExtractionItem*>(item);
		JoinItem* joinItem=qgraphicsitem_cast<JoinItem*>(item);
		if(opItem!=NULL)
		{
			itemsToDelete.push_back(opItem);
		}
		else if(inItem!=NULL)
		{
			itemsToDelete.push_back(inItem);
		}
		else if(outItem!=NULL)
		{
			itemsToDelete.push_back(outItem);
		}
        else if(extItem!=NULL)
        {
			itemsToDelete.push_back(extItem);
		}
		else if(joinItem!=NULL)
        {
			itemsToDelete.push_back(joinItem);
		}
	}
	
	// Now delete
	foreach(CalcItem* item, itemsToDelete)
	{
		OperatorItem* opItem=qgraphicsitem_cast<OperatorItem*>(item);
		InputItem* inItem=qgraphicsitem_cast<InputItem*>(item);
		OutputItem* outItem=qgraphicsitem_cast<OutputItem*>(item);
        ExtractionItem* extItem=qgraphicsitem_cast<ExtractionItem*>(item);
        JoinItem* joinItem=qgraphicsitem_cast<JoinItem*>(item);
		if(opItem!=NULL)
		{
			VOperatorItem* vopItem=opItem->parentItem;
			vopItem->removeGUI();
			delete vopItem;
		}
		else if(inItem!=NULL)
		{
			VInputItem* vinItem=inItem->parentItem;
			vinItem->removeGUI();
			delete vinItem;
		}
		else if(outItem!=NULL)
		{
			VOutputItem* voutItem=outItem->parentItem;
			voutItem->removeGUI();
			delete voutItem;
		}
        else if(extItem!=NULL)
        {
			VExtractionItem* vextItem=extItem->parentItem;
            vextItem->removeGUI();
            delete vextItem;
		}
        else if(joinItem!=NULL)
        {
            VJoinItem* vjoinItem=joinItem->parentItem;
            vjoinItem->removeGUI();
            delete vjoinItem;
        }

	}
	clearSelection();
	
}

void CalcScene::deleteSelection()
{
	QList<QGraphicsItem *> items=selectedItems();
	
	foreach(QGraphicsItem* item, items)
	{
		OperatorItem* opItem=qgraphicsitem_cast<OperatorItem*>(item);
		InputItem* inItem=qgraphicsitem_cast<InputItem*>(item);
		OutputItem* outItem=qgraphicsitem_cast<OutputItem*>(item);
        ExtractionItem* extItem=qgraphicsitem_cast<ExtractionItem*>(item);
        JoinItem* joinItem=qgraphicsitem_cast<JoinItem*>(item);
		if(opItem!=NULL)
		{
			VOperatorItem* vopItem=opItem->parentItem;
			vopItem->removeGUI();
			delete vopItem;
		}
		else if(inItem!=NULL)
		{
			VInputItem* vinItem=inItem->parentItem;
			vinItem->removeGUI();
			delete vinItem;
		}
		else if(outItem!=NULL)
		{
			VOutputItem* voutItem=outItem->parentItem;
			voutItem->removeGUI();
			delete voutItem;
		}
        else if(extItem!=NULL)
        {
            VExtractionItem* vextItem=extItem->parentItem;
            vextItem->removeGUI();
            delete vextItem;
        }
		else if(joinItem!=NULL)
        {
            VJoinItem* vjoinItem=joinItem->parentItem;
            vjoinItem->removeGUI();
            delete vjoinItem;
        }
	}
	clearSelection();
	emit refreshContextWidget();
}

void CalcScene::getItemsBoundaries(bool onlySelectedItems, double& minX, double& minY, double& maxX, double& maxY)
{
	minX=-1.e31;
	minY=-1.e31;
	maxX=-1.e31;
	maxY=-1.e31;
	foreach(VCalcItem* item, currentItems)
    {
        if(!onlySelectedItems || item->isSelected())
		{
			if(item->hasGui())
			{
				VInputItem* vin=dynamic_cast<VInputItem*>(item);
				VOutputItem* vout=dynamic_cast<VOutputItem*>(item);
				VExtractionItem* vext=dynamic_cast<VExtractionItem*>(item);
				VOperatorItem* vop=dynamic_cast<VOperatorItem*>(item);
                VJoinItem* vjoin=dynamic_cast<VJoinItem*>(item);
				if(vin!=NULL)
				{
					QPointF pos=vin->getGUI()->pos();
					if(minX<-1.e30)
					{
						minX=maxX=pos.x();
						minY=maxY=pos.y();
					}
					else
					{
						if(minX>pos.x()) minX=pos.x();
						if(minY>pos.y()) minY=pos.y();
						if(maxX<pos.x()) maxX=pos.x();
						if(maxY<pos.y()) maxY=pos.y();
					}
				}
				else if(vout!=NULL)
				{
					QPointF pos=vout->getGUI()->pos();
					if(minX<-1.e30)
					{
						minX=maxX=pos.x();
						minY=maxY=pos.y();
					}
					else
					{
						if(minX>pos.x()) minX=pos.x();
						if(minY>pos.y()) minY=pos.y();
						if(maxX<pos.x()) maxX=pos.x();
						if(maxY<pos.y()) maxY=pos.y();
					}
				}
				else if(vext!=NULL)
				{
					QPointF pos=vext->getGUI()->pos();
					if(minX<-1.e30)
					{
						minX=maxX=pos.x();
						minY=maxY=pos.y();
					}
					else
					{
						if(minX>pos.x()) minX=pos.x();
						if(minY>pos.y()) minY=pos.y();
						if(maxX<pos.x()) maxX=pos.x();
						if(maxY<pos.y()) maxY=pos.y();
					}
				}
				else if(vjoin!=NULL)
				{
					QPointF pos=vjoin->getGUI()->pos();
					if(minX<-1.e30)
					{
						minX=maxX=pos.x();
						minY=maxY=pos.y();
					}
					else
					{
						if(minX>pos.x()) minX=pos.x();
						if(minY>pos.y()) minY=pos.y();
						if(maxX<pos.x()) maxX=pos.x();
						if(maxY<pos.y()) maxY=pos.y();
					}
				}
				else if(vop!=NULL)
				{
					QPointF pos=vop->getGUI()->pos();
					if(minX<-1.e30)
					{
						minX=maxX=pos.x();
						minY=maxY=pos.y();
					}
					else
					{
						if(minX>pos.x()) minX=pos.x();
						if(minY>pos.y()) minY=pos.y();
						if(maxX<pos.x()) maxX=pos.x();
						if(maxY<pos.y()) maxY=pos.y();
					}
				}
			}
		}
	}
	
	if(minX<-1.e30)
	{
		minX=0.;
		minY=0.;
		maxX=1.;
		maxY=1.;
	}
}

QPointF CalcScene::interactivePosition()
{
	QApplication::setOverrideCursor(Qt::CrossCursor);

	setTracking();
	do
	{
		QCoreApplication::processEvents(QEventLoop::AllEvents, 50);

	} while(tracking);
	QApplication::restoreOverrideCursor();
	return trackingPosition;
}

QDomDocument CalcScene::toXML(bool onlySelectedItems)
{
	QDomDocument doc("QSAS_calculator_save");
	QDomElement root=doc.createElement("QSAS_Calc");
	doc.appendChild(root);
	string guiID;
	guiID=QSAS_VERSION_ID;


        QDomElement version=createTextXMLNode(doc, "QSAS_Version", guiID);
        root.appendChild(version);

		root.appendChild(createTextXMLNode(doc, "Current_Page", currentPage));
		double minX, minY, maxX, maxY;
		getItemsBoundaries(onlySelectedItems, minX, minY, maxX, maxY);
		root.appendChild(createTextXMLNode(doc, "Centre_X", (minX+maxX)/2.));
		root.appendChild(createTextXMLNode(doc, "Centre_Y", (minY+maxY)/2.));
		
        QDomElement nodes = doc.createElement("Nodes");
        root.appendChild(nodes);

    foreach(VCalcItem* item, currentItems)
    {        
        if(!onlySelectedItems || item->isSelected()) nodes.appendChild(item->toXMLElement(doc));
    }

        QDomElement connectors = doc.createElement("Connectors");
		
		foreach(VCalcItem* item, currentItems)
		{
			if(!onlySelectedItems || item->isSelected())
			{
				foreach(VInSlot* to, item->inputs)
				{
					VOutSlot* from=to->connectedSlot();
					if(from!=NULL)
					{

						QDomElement connector=doc.createElement("Connector");
						connectors.appendChild(connector);
						QDomElement fromEl=doc.createElement("From");
						fromEl.setAttribute("Item", from->parent()->calc_id);
						fromEl.setAttribute("Slot_Index", from->parent()->outputs.indexOf(from));
						connector.appendChild(fromEl);
						QDomElement toEl=doc.createElement("To");
						toEl.setAttribute("Item", to->parent()->calc_id);
						toEl.setAttribute("Slot_Index", to->parent()->inputs.indexOf(to));
						connector.appendChild(toEl);
					}
				}
				
			}
		}
		
        root.appendChild(connectors);
		
		// Add all time Lines
		root.appendChild(timeLineHandler.toXMLElement(doc));
		
		// Add pages
		QDomElement elPages = doc.createElement("Pages");

		for(int i=0; i<pages.size(); i++){
			QDomElement pageEl=doc.createElement("Page");
			pageEl.setAttribute("timeLineName", pages[i]->timeLineName);
			elPages.appendChild(pageEl);
		}
		root.appendChild(elPages);
		
	return doc;
}

void CalcScene::fromXML(QDomDocument& doc, bool restore)
{
	if(restore)
	{
		clearAll();

		// reset counters
		VCalcItem::calc_counter=0;
		VInputItem::counter=0;
		VOperatorItem::counter=0;
		VOutputItem::counter=0;
		VJoinItem::counter=0;
	}

    QDomElement docElem = doc.documentElement();
	if(docElem.isNull()) return;

    // get All timeLines
    timeLineHandler.fromXMLElement(docElem);

	// Get pages
	pages.clear();
	QDomElement elPages=docElem.firstChildElement("Pages");
	if(!elPages.isNull())
	{
		QDomElement pageEl = elPages.firstChildElement("Page");
		while(!pageEl.isNull())
		{
			addPage();
			QString timeLineName=pageEl.attribute("timeLineName", "");
			pages[pages.size()-1]->timeLineName = timeLineName;
            emit selectTimeline(timeLineName);
            pageEl = pageEl.nextSiblingElement("Page");
		}
	}
	if(pages.size() == 0) {
		addPage(); // ensure one page exists
		setTimeLine("Automatic");
	}
	double centreX(0.), centreY(0.);
	QString val;

	if(getXMLNodeText(docElem, "Current_Page", val) && restore)
	{
		currentPage=val.toInt();
	}
	
	clearSelection();
	emit refreshContextWidget();
	
	if(getXMLNodeText(docElem, "Centre_X", val))
	{
		centreX=val.toDouble();
	}
	if(getXMLNodeText(docElem, "Centre_Y", val))
	{
		centreY=val.toDouble();
	}

		
	if(!restore) // if include, get position for inclusion
	{
		interactivePosition();
		if(!trackingSuccess) // Cancelled
		{
			return;
		}
	}
		
	QDomElement nodes=docElem.firstChildElement("Nodes");
	if(!nodes.isNull())
	{
		QDomElement n = nodes.firstChildElement();
		while(!n.isNull())
		{
			if(n.tagName()=="Operator")
			{
				VOperatorItem* item=new VOperatorItem;
				item->fromXMLElement(n, this);
				if(item->hasGui() && !restore)
				{
					item->getGUI()->moveBy(trackingPosition.x()-centreX, trackingPosition.y()-centreY);
				}
				if(!restore) item->page=currentPage;
			}
			else if(n.tagName()=="Input")
			{
				VInputItem* item=new VInputItem;
				item->fromXMLElement(n, this);

                if(item->hasGui() && !restore)
				{
					InputItem* item_gui=item->getGUI();
					item_gui->moveBy(trackingPosition.x()-centreX, trackingPosition.y()-centreY);
					connect(item_gui, SIGNAL(newInput(const QString&)), this, SIGNAL(newInput(const QString&)));
				}
				if(!restore) item->page=currentPage;
			}
			else if(n.tagName()=="Join")
			{
				VJoinItem* item=new VJoinItem;
				item->fromXMLElement(n, this);
				if(item->hasGui() && !restore)
				{
					item->getGUI()->moveBy(trackingPosition.x()-centreX, trackingPosition.y()-centreY);
				}
				if(!restore) item->page=currentPage;
			}
			else if(n.tagName()=="Output")
			{
				VOutputItem* item=new VOutputItem;
				item->fromXMLElement(n, this);
				if(item->hasGui() && !restore)
				{
					item->getGUI()->moveBy(trackingPosition.x()-centreX, trackingPosition.y()-centreY);
				}
				if(!restore) item->page=currentPage;
			}
			else if(n.tagName()=="Extractor")
			{
				VExtractionItem* item=new VExtractionItem;
				item->fromXMLElement(n, this);
				if(item->hasGui() && !restore)
				{
					item->getGUI()->moveBy(trackingPosition.x()-centreX, trackingPosition.y()-centreY);
				}
				if(!restore) item->page=currentPage;
			}

			n = n.nextSiblingElement();
		}
	}
		
	
	QDomElement connectors=docElem.firstChildElement("Connectors");
	if(!connectors.isNull())
	{
		QDomElement n = connectors.firstChildElement("Connector");
		while(!n.isNull())
		{
			QDomElement fromEl=n.firstChildElement("From");
			QDomElement toEl=n.firstChildElement("To");
			if(!fromEl.isNull() && !toEl.isNull())
			{

				int fromItemId, fromItemIndex, toItemId, toItemIndex;
				fromItemId=fromEl.attribute("Item", "-1").toInt();
				fromItemIndex=fromEl.attribute("Slot_Index", "-1").toInt();
				toItemId=toEl.attribute("Item", "-1").toInt();
				toItemIndex=toEl.attribute("Slot_Index", "-1").toInt();
				if(fromItemId>=0 && fromItemIndex>=0 && toItemId>=0 && toItemIndex>=0)
				{

					VCalcItem* fromItem=NULL;
					VCalcItem* toItem=NULL;
					for(QList<VCalcItem*>::const_iterator iter=currentItems.constBegin();
						iter!=currentItems.constEnd();
						++iter)
					{
						if((*iter)->temp_id==fromItemId)
						{
							fromItem=(*iter);
						}
						if((*iter)->temp_id==toItemId)
						{
							toItem=(*iter);
						}
					}

					if(fromItem!=NULL && toItem!=NULL && fromItemIndex<fromItem->outputs.size() && toItemIndex<toItem->inputs.size())
					{

						VOutSlot* fromSlot=fromItem->outputs[fromItemIndex];
						VInSlot* toSlot=toItem->inputs[toItemIndex];
                        
                        
						if(fromItem->hasGui() && toItem->hasGui() )
						{
							Connector* connector=new Connector(fromSlot->getGUI(), toSlot->getGUI());
							addItem(connector);
							connector->page=fromItem->page;
							fromSlot->getGUI()->updateConnectorsPositions();
						}
							
						fromSlot->connectTo(toSlot);
					}
				}
			}
			n = n.nextSiblingElement("Connector");
		}
			
	}


	// now create the black boxes
		
	foreach(VCalcItem* item, currentItems)
	{
		if(item->temp_id>=0) // Items being loaded
		{
			VOperatorItem* operatorItem=dynamic_cast<VOperatorItem*>(item);
			if(operatorItem!=NULL)
			{
				if(operatorItem->operation==QCALC_BLACKBOX)
				{
					// Get all the items defined by this black box
					QList<VCalcItem*> bb;
					bb.clear();
					foreach(VCalcItem* i, currentItems)
					{
						if(operatorItem->blackBoxIDList.contains(i->temp_id))
						{
							bb.push_back(i);
						}
					}
                    QString name=operatorItem->getName();
					operatorItem->setAsBlackBox(bb, name);
				}
			}
		}
	}
	
	foreach(VCalcItem* item, currentItems)
	{
		item->temp_id=-1; // Reset temp ids after connections are complete
	}
	
	emit currentPageChanged();
	updateVisibleItems(true);
	update();
}


void CalcScene::addMissingConnectors()
{
	QList<Connector*> existingConnectors;
	foreach(QGraphicsItem* item, items())
	{
		Connector* connector=qgraphicsitem_cast<Connector*>(item);
		if(connector!=NULL)
		{
			existingConnectors.push_back(connector);
		}
	}
	
	foreach(VCalcItem* item, currentItems)
	{
		foreach(VInSlot* inSlot, item->inputs)
		{
			if(inSlot->hasGUI())
			{
				VOutSlot* outSlot=inSlot->connectedSlot();
				if(outSlot!=NULL)
				{
					if(outSlot->hasGUI())
					{
						InSlot* inGui=inSlot->getGUI();
						OutSlot* outGui=outSlot->getGUI();
						
						// Look for connector from outGui to inGui
						bool found=false;
						for(QList<Connector*>::iterator iter=existingConnectors.begin(); iter!=existingConnectors.end() && !found; ++iter)
						{
							found=((*iter)->getFromSlot()==outGui && (*iter)->getToSlot()==inGui);
						}
						if(!found)
						{
							Connector* connector=new Connector(outGui, inGui);
							addItem(connector);
                            connector->page=item->page;

							outGui->updateConnectorsPositions();
						}
					}
				}
			}
		}
	}
}

