/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	BoundaryNormalCoords.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Boundary Normal Coordinates Plugin wrapper
*
* PURPOSE:      Calculate Rotation Matrix to transform into Boundary
*               Normal Coordinates lmn. n is normal to a model magnetopause
*               or bow shock, the l-n plane is parallel the Earth's dipole
*               axis, and lmn is a RH system.
* 
* DESCRIPTION:  .
*
* ARGUMENTS:
*    IN:      
*             pos_in      Constant vector or Vector TS
*             time_c      Crossing time
*             Vsw         Approx solar wind speed to calculate aberration
*             imodel      Model number from Ch10 of ISSI Book
*             frame_out   GSE->lmn (0) or GSM->lmn (1)
*
*   OUT:
*             frametolmn  Rotation Matrix R. 
*                          R dot vector(frame_out) = vector(lmn)		
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: built on qrtn and qbowshock modules
*
* HISTORY: July 2001 Steve Schwartz 
*          April 2015 Tony Allen (migration to DVOS)
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"

#include "qdutil.h"
#include "DvObj.h"
#include "qrtn.h"
#include "qbowshock.h"

const double Re = 6.3712e6; //Earth radius [m]  - a global	

using namespace QSAS;  

// local prototypes

extern "C" QplugReturnStatus DvBoundaryNormalCoords(vector <DvObject_var>&, vector <DvObject_var>&);

QplugReturnStatus QBNPosition(RtnVector *vec_gsep,
                              DvObject_var pos_in,
							  DvObject_var time_c);
QplugReturnStatus QBNConstPosition(RtnVector *vec_gsep,
								   DvObject_var pos_in, 
				                   DvObject_var time_c);
QplugReturnStatus QBNCentrePosition(RtnVector *vec_gsep,
                                    DvObject_var pos_in_seq, 
				                    DvObject_var time_c);
QplugReturnStatus QBNMakeGSERe (RtnVector *vecp);
void QrVector_diag_print(RtnVector *vecp);
  
//**********************************************************************

QplugReturnStatus DvBoundaryNormalCoords(vector <DvObject_var>& inputs, vector <DvObject_var>& outputs){

	if(inputs.size() < 5) {
    	QplugAppendTextDisplay ("5 inputs required\n");
    	return QPLUG_FAILURE; 
 	} 

	// define some C variables

	long  check_rtn;
	float Vsw, r_vec[3], n_gse[3];
	int   imodel;
	int   i, qbs_rtn, qar_status;

	DvObject_var modm;

	// input arguments
 
	DvObject_var        pos_in;        // 0
	DvObject_var		time_c;        // 1
	DvObject_var        Vsw_var;       // 2
	DvObject_var        imodel_var;    // 3
	DvObject_var        frame_out_var; // 4
	DvObject_var        frametolmn;    


	RtnVector  vec_gse, z, l, m, n;

	QplugReturnStatus ret_status;

	// ...first 2 items are input position vector/vector TS and time interval

	pos_in = inputs[0];
	if (pos_in.is_nil()){
  		QplugAppendTextDisplay("BNCoords>> Error getting pos_in object\n");
  		return QPLUG_FAILURE;
	}

	time_c = inputs[1];
	if( time_c.is_nil() || !time_c->is_time() ) {
  		QplugAppendTextDisplay ("BNCoords>> Time object invalid.\n");
  		return QPLUG_FAILURE;
	}

	vec_gse.IsPositionVector = 0;     // force so QrDoRotation will accept

	ret_status = QBNPosition(&vec_gse, pos_in, time_c);
	if (ret_status != QPLUG_SUCCESS){
  		QplugAppendTextDisplay("BNCoords>> QBNPosition failed\n");
  		return QPLUG_FAILURE;
	}

	// vec_gse now has a sequence with 1 element, the gse_xyz position
	// of the crossing in Re with one timetag. 

	Vsw_var = inputs[2];
	if (Vsw_var.is_nil() || Vsw_var->totalSize() > 1){
   		QplugAppendTextDisplay("BNCoords>> Input arg Vsw not Scalar\n");
   		return QPLUG_FAILURE;
	}
	
	imodel_var = inputs[3];
	if(imodel_var.is_nil() || imodel_var->totalSize() > 1){
		QplugAppendTextDisplay("BNCoords>> Input arg imodel not Scalar\n");
		return QPLUG_FAILURE;
	}

	// unpack inputs for QBOWSHOCK routines and find gse normal;
	Vsw = (float) Vsw_var->asDouble();
	imodel = (int) imodel_var->asInt();
	for (i=0; i<3; i++){
		r_vec[i] = (float) vec_gse.ptr->dbl(i);
	}

	qbs_rtn = QBSNormalAtSurfaceGSE(r_vec, Vsw, (QBSMP_models)imodel, n_gse);
	if (qbs_rtn != QBS_OK){
   		QplugAppendTextDisplay("BNCoords>> QBSNormal returned error\n");
   		return QPLUG_FAILURE;
	}

	// put normal in n, set z to (0,0,1)GSM

	n          = vec_gse;   // copy over n_recs = 1, timetag, ...
	n.in_frame = GSE;
	vector <size_t> dims;
	dims.push_back(3);
	n.ptr      = new DvObject(1.0, dims, 1);
	for (i=0; i<3; i++){n.ptr->dbl(i) = n_gse[i];}
	n.ptr->change_xref(FRAME, "vector>gse_xyz");

	z           = n;
	z.in_frame  = GSM;
	z.ptr       = new DvObject(0.0, dims, 1);
	z.ptr->dbl(2) = 1.0;   
	z.ptr->change_xref(FRAME, "vector>gsm_xyz");

	frame_out_var = inputs[4];
	if(frame_out_var.is_nil()){
		QplugAppendTextDisplay("BNCoords>> Frame_out not a scalar\n");
		return QPLUG_FAILURE;
	}

	dims.push_back(3); // now 3x3
	frametolmn = new DvObject(0.0, dims, 1);

	switch ( frame_out_var->asInt() ){
		case 1:      // gsm
			n.out_frame = GSM;
			z.out_frame = GSM;
			frametolmn->change_xref(FRAME, "rot>gsm_xyz");
			break;
		case 0: default:    // gse
			n.out_frame = GSE;
			z.out_frame = GSE;
			frametolmn->change_xref(FRAME, "rot>gse_xyz");
			break;
	}

	check_rtn = n.QrDoRotation();  // put n in correct frame
	if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
		QplugAppendTextDisplay( "BNCoords>> Failed to put n into out frame\n");
		return QPLUG_FAILURE;
	}
	
	n.in_frame = n.out_frame;      // make in_frame consistent with holding
	check_rtn = z.QrDoRotation();  // put z in correct frame
	if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
		QplugAppendTextDisplay( "BNCoords>> Failed to put z into out frame\n");
		return QPLUG_FAILURE;
	}
	z.in_frame = z.out_frame;      // make in_frame consistent with holding

	// lots of error checking; 

	m.ptr = n.ptr->vec(z.ptr);
	
	if (m.ptr.is_nil()) {
    	QplugAppendTextDisplay ("BNCoords>> n cross z failed\n");
    	return QPLUG_FAILURE;
	}

	modm = m.ptr->abs() ;
	if (modm.is_nil()){
    	QplugAppendTextDisplay ("BNCoords>> |m| returned invalid number\n");
    	return QPLUG_FAILURE;
	}

	if (m.ptr.is_nil()) {
    	QplugAppendTextDisplay ("BNCoords>> m/|m| failed\n");
    	return QPLUG_FAILURE;
	}
	*(m.ptr) /=  modm->dbl(0);

	l.ptr = m.ptr->vec(n.ptr);

	if (l.ptr.is_nil()){
    	QplugAppendTextDisplay ("BNCoords>> l = m x n failed\n");
    	return QPLUG_FAILURE;
	}

	for (i=0; i<3; i++){
    	frametolmn->dbl(0,0,i) = l.ptr->dbl(i);
    	frametolmn->dbl(0,1,i) = m.ptr->dbl(i);
    	frametolmn->dbl(0,2,i) = n.ptr->dbl(i);
	}

	// set the attributes for the output rotation matrix

	frametolmn->change_xref(FIELDNAM, "RMatrix_to_lmn");
	frametolmn->change_xref(SI_CONVERSION, "1>()");
	frametolmn->change_xref(UNITS, " ");
	frametolmn->change_xref("ToFrame", "vector>lmn_xyz");
	char model_name[200];
	sprintf(model_name,"Model %d: %s",imodel,BSMP_name[imodel]);
	frametolmn->change_xref("Boundary Model", model_name);

	DvObject_var crossing = vec_gse.ptr;
	char crossing_str[200];
	sprintf(crossing_str,"( %9.3g, %9.3g, %9.3g )GSE Re", crossing->dbl(0),crossing->dbl(1),crossing->dbl(2));
	frametolmn->change_xref("Crossing", crossing_str);
	char crossingt_str[200];
	sprintf(crossingt_str,"%s", vec_gse.t->time().getISOstring().c_str());
	frametolmn->change_xref("Crossing Time",crossingt_str);


	// put some results to plug-in text window
	char outstr[250];
	DvString FrameTxt = frametolmn->getXrefText("Frame");
	sprintf(outstr,"BNCoords>> Crossing %s at %s\n", crossing_str, crossingt_str);
	QplugAppendTextDisplay(outstr);
	sprintf(outstr,"BNCoords>> Normal is (%9.3g, %9.3g, %9.3g) Frame = %s\n", n.ptr->dbl(0), n.ptr->dbl(1), n.ptr->dbl(2), FrameTxt.c_str());
	QplugAppendTextDisplay(outstr);
	QplugAppendTextDisplay("BNCoords>> finished successfully\n");

	outputs.push_back(frametolmn);     

	return QPLUG_SUCCESS;

}

//**********************************************************************
QplugReturnStatus QBNPosition(	RtnVector *vec_gsep,
								DvObject_var pos_in,
								DvObject_var time_c){

	long  check_rtn;

	QplugReturnStatus ret_status;

	// sort out position of crossing. Input can either be a time series or
	// a constant vector, so this a bit complicated. Separate routines
	// return in vec_gse a sequence of 1 cartesian vector whose timetag
	// is the selected time, and whose frame is
	// that of pos_in. This is then rotated into gse. [NB qrtn routines at
	// present (July 2001) assume vectors are _xyz]

	if( pos_in->seqSize() > 1) {
		QplugAppendTextDisplay ("BNCoords>> Vector Sequence. Interpolating ...\n");
		ret_status = QBNCentrePosition(vec_gsep, pos_in, time_c);
		if (ret_status != QPLUG_SUCCESS){
			QplugAppendTextDisplay("BNCoords>> Failed to calculate crossing position\n");
			return QPLUG_FAILURE;
		}
	} 
	else {
		ret_status = QBNConstPosition(vec_gsep, pos_in, time_c);
		if (ret_status != QPLUG_SUCCESS){
			QplugAppendTextDisplay("BNCoords>> Failed to calc const. position\n");
			return QPLUG_FAILURE;
		}
	}

	// QrVector_diag_print(vec_gsep);
	vec_gsep->out_frame = GSE;
	check_rtn = vec_gsep->QrDoRotation();  // make sure vec_gse in gse
	if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
		QplugAppendTextDisplay( "BNCoords>> Failed to put pos into gse\n");
		return QPLUG_FAILURE;
	}
	// QrVector_diag_print(vec_gsep);
	vec_gsep->in_frame = GSE;        // advertise in GSE so don't double rotate

	return QPLUG_SUCCESS;
}

//**********************************************************************
QplugReturnStatus  QBNConstPosition(RtnVector *vec_gsep,
                                    DvObject_var pos_in, 
				    				DvObject_var tt){

	QplugReturnStatus ret_status;

	// grab constant vector and fill vec_gsep.

	if (!pos_in->isThreeVector()){
		QplugAppendTextDisplay("BNCoords>> Const pos_in not a 3Vector\n");
		return QPLUG_FAILURE;
	}

	vec_gsep->ptr = new DvObject(pos_in); // copy constructor includes xrefs
	vec_gsep->ptr->ensureVectorXYZ();
	
	vec_gsep->n_recs = 1;
	vec_gsep->t = tt;

	vec_gsep->in_frame = vec_gsep->QrGetFrameFromAttr(pos_in);
	if (vec_gsep->in_frame == UNKNOWN){
		QplugAppendTextDisplay("BNCoords>> Const pos_in unknown frame\n");
		return QPLUG_FAILURE;
	}

	ret_status = QBNMakeGSERe(vec_gsep);
	if (ret_status != QPLUG_SUCCESS){
		QplugAppendTextDisplay("BNCoords>> QBNMakeGSERe failed\n");
		return QPLUG_FAILURE;
	}

	return QPLUG_SUCCESS;
}


//**********************************************************************
QplugReturnStatus  QBNCentrePosition(RtnVector *vec_gsep,
                                     DvObject_var pos_in, 
				    				 DvObject_var tt){

	int i,nttin,ntrue,il,ih;
	QplugReturnStatus ret_status;
	double cl,ch, frac;

	if (!pos_in->isThreeVector()){
		QplugAppendTextDisplay("BNCoords>> Sequence pos_in not a 3Vector\n");
		return QPLUG_FAILURE;
	}

	// prepare vec_gsep 
	
	vec_gsep->t = tt;
	vec_gsep->n_recs = 1;
	vec_gsep->in_frame = vec_gsep->QrGetFrameFromAttr(pos_in);
	if (vec_gsep->in_frame == UNKNOWN){
		QplugAppendTextDisplay("BNCoords>> Const pos_in unknown frame\n");
		return QPLUG_FAILURE;
	}

	// linearly interpolate crossing position from bracketting time tags

	DvObject_var interpPos = pos_in->interpAt(tt->time(0));
	vec_gsep->ptr = interpPos;
	
	ret_status = QBNMakeGSERe(vec_gsep);
	if (ret_status != QPLUG_SUCCESS){
		QplugAppendTextDisplay("BNCoords>> QBNMakeGSERe failed\n");
		return QPLUG_FAILURE;
	}

	return QPLUG_SUCCESS;
}

//**********************************************************************
QplugReturnStatus QBNMakeGSERe (RtnVector *vecp){

	// rotate vecp to GSE and scale to Re

	int check_rtn;
	double factor, SIfactor;

	// Rotate

	vecp->out_frame = GSE;

	check_rtn = vecp->QrDoRotation();
	if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
		QplugAppendTextDisplay("BNCoords>> QBNMakeGSERe rotate to gse failed\n");
		printf("QBNMakeGSERe QrDoRotation returned %d\n",check_rtn);
		return QPLUG_FAILURE;
	}
	vecp->in_frame = GSE; // if don't do this, another call to QrDoRot will rotate again!
	
	// Scale
	vecp->ptr = vecp->ptr->changeUnitsTo("6.3712e+06>m" , "Re");

	return QPLUG_SUCCESS;
}


/**************************************************************************/
void QrVector_diag_print(RtnVector *vecp){
  printf("IsPositionVector = %d, in_frame = %d, out_frame = %d, n_recs = %d\n",
        vecp->IsPositionVector,vecp->in_frame,vecp->out_frame, vecp->n_recs);
  printf("First Vector: (%9.3g, %9.3g, %9.3g) at %s\n",
     (double)vecp->ptr->asDouble(0), (double)vecp->ptr->asDouble(1), (double)vecp->ptr->asDouble(2),
     vecp->t->time(0).iso_srep().c_str());
	 fflush(stdout);
  return;
}



