// calcBShock.cc Jeremy Mitchell 05/08/2009
// Uses qbowshock (sjs) routines 
// Fixed and migrated to DVOS Tony Allen 2015

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "qbowshock.h"

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include "Xrefs.h"

extern "C" QplugReturnStatus DvcalcBShock(vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;

QplugReturnStatus DvcalcBShock(vector <DvObject_var>& inputs, vector <DvObject_var>& outputs) {

	DvObject_var sw_vel_in = inputs[0];
	DvObject_var model_in = inputs[1];
	DvObject_var scale_in = inputs[2];
	DvObject_var pos_in = inputs[3];
	DvObject_var plane_in = inputs[4];
  
	if ( sw_vel_in.is_nil() || sw_vel_in->totalSize() != 1) {
    	QplugAppendTextDisplay("calcBShock Error: Unable to unpack SW vel\n");
    	return QPLUG_FAILURE;
	}

  	DvObject_var sw_vel_kps = sw_vel_in->changeUnitsTo("1.0e3>m s^-1", "km/s");
	
	float vel_sw = (float) sw_vel_kps->asDouble(0);
	if (vel_sw < 0) {
    	QplugAppendTextDisplay("calcBShock Error: SW vel < 0\n");
    	return QPLUG_FAILURE;
	}
  
	if (model_in.is_nil() || model_in->totalSize() != 1) {
    	QplugAppendTextDisplay("calcBShock Error: Unable to unpack model\n");
    	return QPLUG_FAILURE;
	}
	int model = (int) model_in->asInt(0);
	if (model < 0 || model > 13) {
    	QplugAppendTextDisplay("calcBShock Error: Must have 0 <= model <= 13\n");
    	return QPLUG_FAILURE;
	}
  
	if (scale_in.is_nil() || scale_in->totalSize() != 1) {
    	QplugAppendTextDisplay("calcBShock Error: Unable to unpack scale\n");
    	return QPLUG_FAILURE;
	}
	double scale = (double)scale_in->asDouble(0);
  
	bool useOrigin = false;
  
	if (pos_in.is_nil() ){
		QplugAppendTextDisplay("calcBShock Error: position missing\n");
		return QPLUG_FAILURE;
	}
	if( pos_in->arraySize() != 3) {
    	if ( pos_in->asInt(0) != 0 ) {
			QplugAppendTextDisplay("calcBShock Error: position must be either 0 or a vector\n");
			return QPLUG_FAILURE;
		}
		useOrigin = true;
	} 
	DvString frame = pos_in->getFrame();

	if( !frame.empty() && frame != "gse"){
		QplugAppendTextDisplay("calcBShock Error: position must be in GSE\n");
		return QPLUG_FAILURE;
	}
	
	DvObject_var pos_Re = pos_in->changeUnitsTo("6.3712e+06>m", "Re");
	
	if (plane_in.is_nil() || plane_in->totalSize() != 1) {
		QplugAppendTextDisplay("calcBShock Error: Unable to unpack plane\n");
    	return QPLUG_FAILURE;
	}
	float plane = (float) plane_in->asDouble(0);
	if (plane < 0. || plane > 360.) {
    	QplugAppendTextDisplay("calcBShock Error: Must have 0 <= plane <= 360\n");
    	return QPLUG_FAILURE;
	}
  
	//************************ Units ****************************
   
 	// convert plane to radians, unless already done
	if (plane_in->isDeg()) {
		plane /= 57.295779506;
		QplugAppendTextDisplay("Plane in degrees\n");
	}
	else if (plane_in->isRad()) QplugAppendTextDisplay("Plane in radians\n");
	else {
		plane /= 57.295779506;
		QplugAppendTextDisplay("Plane assumed to be in degrees\n");
    }
	//****************************************************
  
	int nmax = 20;
  
	vector <size_t> dims;
	dims.push_back(3);
	DvObject_var shock = new DvObject(0.0, dims, nmax);
	DvObject_var distance = new DvObject(0.0, nmax);
	DvObject_var angle = new DvObject(0.0, nmax);
  
	float min = 1.e30;
	float max = -1.e30;
  
	for (int i=0; i < nmax; i++) {
    	float alpha = ( 1.5 * 3.14159 * (i+1) )/ (nmax + 1) - 0.25 * 3.14159;
		float rin[3], pos0[3];
    	if (useOrigin) {
      		pos0[0] = 0.0;
      		pos0[1] = 0.0;
      		pos0[2] = 0.0;
    	} 
		else {
			pos0[0] = (float) pos_Re->asDouble(0);
			pos0[1] = (float) pos_Re->asDouble(1);
			pos0[2] = (float) pos_Re->asDouble(2);
		}
    
		rin[0] = -sin(alpha);
		rin[1] = -cos(alpha)*cos(plane);
		rin[2] = -cos(alpha)*sin(plane);
    
		QBSMP_models model_arg = (QBSMP_models) model;
    
		float dist, r_out[3];
    
		QBSPosDistanceAlongVec(pos0, vel_sw, model_arg, scale, rin, &dist, r_out);

		if (r_out[0] < min) min = r_out[0]; if (r_out[1] < min) min = r_out[1]; if (r_out[2] < min) min = r_out[2];
		if (r_out[0] > max) max = r_out[0]; if (r_out[1] > max) max = r_out[1]; if (r_out[2] > max) max = r_out[2];
    
		distance->dbl(i) = dist;
		angle->dbl(i) = alpha;
		shock->dbl(i, 0) = r_out[0];
    	shock->dbl(i, 1) = r_out[1];
    	shock->dbl(i, 2) = r_out[2];
	}
  
	//********************** Set Attributes *************************
  
	shock->change_xref(FRAME, "vector>gse_xyz");
	shock->change_xref("TENSOR_FRAME", "gse");
	shock->change_xref(SI_CONVERSION, "6.3712e6>m");
	shock->change_xref(UNITS, "Re");
	DvObject_var torder = new DvObject((int)1);
	shock->change_xref(TENSOR_ORDER, torder);
	if (model <= 5)  {
    	shock->change_xref(LABLAXIS, "Posn Bowshock");
    	shock->change_xref(FIELDNAM, "Position of Bowshock");
    	distance->change_xref(LABLAXIS, "Dist Bowshock");
    	distance->change_xref(FIELDNAM, "Distance to Bowshock");
  	} 
	else {
    	shock->change_xref(LABLAXIS, "Posn Magnetopause");
    	shock->change_xref(FIELDNAM, "Position of Magnetopause");
    	distance->change_xref(LABLAXIS, "Dist Magnetopause");
    	distance->change_xref(FIELDNAM, "Distance to Magnetopause");
	}
	DvObject_var rep = new DvObject(DvString(), dims );
	rep->str(0) = "x"; rep->str(1) = "y"; rep->str(2) = "z";
	shock->change_xref(REPRESENTATION_1, rep);
	shock->change_xref(LABEL_1, rep);
  
	distance->change_xref(SI_CONVERSION, "6.3712e6>m");
	distance->change_xref(UNITS, "Re");
	distance->change_xref(REPRESENTATION_1, rep);
	distance->change_xref(LABEL_1, rep);

	angle->change_xref(SI_CONVERSION, "1>rad");
	angle->change_xref(UNITS, "rad");
	angle->change_xref(FIELDNAM, "angle in plane from Y towards Z");
	angle->change_xref(LABLAXIS, "alpha");
    
	shock->change_xref(DEPEND_0, angle);
	distance->change_xref(DEPEND_0, angle);

	outputs.push_back( distance );
	outputs.push_back( shock );
  
	return QPLUG_SUCCESS;
  
}
