
/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	CSEQ2IES.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Plugin to rotate a vector from 67P CSEQ coordinates to RPC IES instrument frame
*
* PURPOSE:      As in FUNCTION
* 
* DESCRIPTION:  Calls NASA SPUICE routine to return rotation matrix at time(s) passed as input.
*               Then applies these to the input vector/vector time series and returns rotated
*               vector(s) to Working List.
*
*               Typical cases (think about how these should work):
*				tagsin, vecin are tags and values of B_vec time series
*				tagsin from some TS object; vecin = (1,0,0) for Sun direction
*
* ARGUMENTS:
*    IN:      
*             tagsin      Timetags, or object with timetags, or constant time
*             vecin       Input vector or vector time series. Warning issued if frame not 
*                         identifiable as CSEQ but will proceed anyway
* 
*   OUT:
*             vecout      vecin rotated from cseq to ies frame	
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: built on SPICE C routines
*
* Note: requires SPICE kernels valid for the times in tagsin and vecin
*
* HISTORY:   SJS 15 July 2015
*/

#include "qplug_if.h"
#include "DvObj.h"
//#include "SpiceUsr.h"

//#define DEBUG_IESSUNAZ    // for testing

extern "C" QplugReturnStatus DvCSEQ2IES(vector<DvObject_var>&, vector<DvObject_var>&);

//extern "C" int cseq2ies_mat(const char *time_string, SpiceDouble mat[3][3]);
extern "C" int cseq2ies_mat_dummy(const char *time_string, double mat[3][3]);
//extern "C" int printmat(SpiceDouble mat[3][3]);
extern "C" void clear_spice();

QplugReturnStatus DvCSEQ2IES(vector<DvObject_var>& inputs, vector<DvObject_var>& outputs)
{
	DvObject_var tagsin_ptr, vecin_ptr;
	DvObject_var out_ptr;
	
	DvObject_var mat_ptr;			// will hold rotation matrix
//	SpiceDouble spicemat[3][3];		// spice will fill this one tag at a time
	double spicemat[3][3];		// spice will fill this one tag at a time
   
    int ntags;

	DvString out_name;				// name of output object

 	if(inputs.size() < 2) {
    	QplugAppendTextDisplay ("insufficient inputs\n");
    	return QPLUG_FAILURE; 
 	} 
 
	//	Unpack and check the input data objects
	tagsin_ptr = inputs[0]->getTimeTags();
	if(tagsin_ptr.is_nil() ){
    	QplugAppendTextDisplay ( "Couldn't get TimeTags from 1st input\n");
    	return QPLUG_FAILURE; 
	}
	ntags = tagsin_ptr->seqSize();
 	
	vecin_ptr = inputs[1];
	if(vecin_ptr.is_nil() || !vecin_ptr->isThreeVector() ){
    	QplugAppendTextDisplay ( "Invalid input vector; nil or not vector \n");
    	return QPLUG_FAILURE; 
	}
	if(vecin_ptr->getFrame() != "CSEQ"){
		QplugAppendTextDisplay ( "!!! WARNING !!! Input vector not in CSEQ; proceed with caution \n");
	}

	// Get name of output object
 	out_name = GetSlotText(0);
 	if (out_name.empty()) {
    	out_name = "VectorInIESCoords";
    	SetSlotText(0, out_name.c_str());
	}

#ifdef DEBUG_IESSUNAZ	
	cout << "first time tag, last time tag " << tagsin_ptr->time(0).getISOstring() << " , " << tagsin_ptr->time(ntags-1).getISOstring() << endl;
#endif
	
// create DvObj TS of rotation matrices

   mat_ptr = new DvObject(0.0,DIM33,ntags);
   mat_ptr->change_xref(DEPEND_0,tagsin_ptr);
   
 // loop over timetags and fill rotation matrix DvObj
   for (int rec=0; rec<ntags; rec++){
		if (!cseq2ies_mat_dummy(tagsin_ptr->time(rec).getISOstring().c_str(),spicemat)){  // 0 is failure
			QplugAppendTextDisplay ( "cseq2ies_mat failed; exiting \n");
			QplugAppendTextDisplay ("Check console for SPICE message. Possible causes include\n");
			QplugAppendTextDisplay ("   Failure to load kernels (path problem?)\n");
			QplugAppendTextDisplay ("   Kernel out of date - doesn't cover requested time period\n");
			return QPLUG_FAILURE;
		}


 		for (int i=0; i<3; i++) {
			for (int j=0; j<3; j++) mat_ptr->dbl(rec,i,j) = spicemat[i][j];
		}
		/*
  		  cout << "mat_ptr matrix ("<<rec<<"):" << endl;
		  cout << mat_ptr->dbl(rec,0,0) << "  " << mat_ptr->dbl(rec,0,1) << "  " << mat_ptr->dbl(rec,0,2) << endl;
		  cout << mat_ptr->dbl(rec,1,0) << "  " << mat_ptr->dbl(rec,1,1) << "  " << mat_ptr->dbl(rec,1,2) << endl;
		  cout << mat_ptr->dbl(rec,2,0) << "  " << mat_ptr->dbl(rec,2,1) << "  " << mat_ptr->dbl(rec,2,2) << endl;
		  */
   }
   
   
   clear_spice();  // unload kernels, clear rest of spice, in case another plugin needs different config

// create output vector

	DvObject_var vecout_ptr = mat_ptr * vecin_ptr;   // How will it join?	

// copy xrefs, then set frame attribute to rpcies_xyz
//	vecout_ptr->copy_xrefs_from(vecin_ptr);  // not correct for Sun=(1,0,0) - rethink
	vecout_ptr->setFrameAttr("vector>rpcies_xyz");

// put back on WL
	outputs.push_back(vecout_ptr);	
	QplugAppendTextDisplay("CSEQ_2_IES>> Finished\n");
	
#ifdef DEBUG_IESSUNAZ
	cout << "End " << endl;
#endif
	
	return QPLUG_SUCCESS;
}


