#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"

using namespace QSAS;

extern "C" QplugReturnStatus DvCharacteristicFrequencies(vector <DvObject_var>&, vector <DvObject_var>&);

QplugReturnStatus DvCharacteristicFrequencies(vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{   

     bool hasB=true;
     bool hasNe=true;
     
     DvObject_var B = inputs[0];
	 DvObject_var B_Tesla;
	 
     if(B.is_nil() || B->arraySize() > 1)
     {
         QplugAppendTextDisplay("Warning: B is not a scalar sequence - considering B is unknown.");
		 hasB=false;
     }
     
     DvObject_var Ne = inputs[1];
     DvObject_var Ne_SI;   
	   
     if(Ne.is_nil())
     {
         QplugAppendTextDisplay("Warning: Ne is not a scalar sequence - considering Ne is unknown.");
		 hasNe=false;
     }

     
     if(!hasB && !hasNe)
     {
         QplugAppendTextDisplay("No data available: returning.");
         return QPLUG_FAILURE;
     }
     
     if(hasB && hasNe)
     {
	 	DvObject_var target = Ne->getDep0();
		B = B->nnJoin(*target);
		 
		if(B.is_nil())
		{
			QplugAppendTextDisplay("Failed to join input time series.");
			return QPLUG_FAILURE;
		}
     }
 
     //Check units
     
     if(hasB)
     {
		 B_Tesla = B->changeUnitsTo("1>T" , "Tesla");
                  
         if(B_Tesla.is_nil())
         {
             QplugAppendTextDisplay("Conversion to Tesla failed.");
             return QPLUG_FAILURE;
         }
     }
     
     if(hasNe)
     {
		 Ne_SI = Ne->changeUnitsTo("1 > m^-3" , "m^-3");
         
         if(Ne_SI.is_nil())
         {
             QplugAppendTextDisplay("Conversion to m^-3 failed.");
             return QPLUG_FAILURE;
         }
     }
     
     // Constants
     
     DvObject_var q_e; 
     DvObject_var m_e; double m_e_value;
     DvObject_var m_p; double m_p_value;
     DvObject_var eps_0; double eps_0_value;
     
     double pi_value = (double)  4.0 *atan(1.0);
     double pi_2 = pi_value * 2.;
     double pi_4 = pi_value * 4.;
	 
   // Electron Charge
     q_e = new DvObject(1.60210e-19);
	 
     q_e->change_xref(SI_CONVERSION, "1>C" );
     q_e->change_xref(FIELDNAM, "Electron charge");
     q_e->change_xref( FRAME, "scalar>na");
     q_e->change_xref(UNITS, "C");
     
  // electron mass
     m_e = new DvObject(9.1091e-31);
     m_e->change_xref(SI_CONVERSION, "1>kg");
     m_e->change_xref(FIELDNAM, "mass of electron");
     m_e->change_xref( FRAME, "scalar>na");
     m_e->change_xref(UNITS, "kg");
  
   //  proton mass
     m_p = new DvObject(1.6726e-27);
     m_p->change_xref(SI_CONVERSION, "1>kg");
     m_p->change_xref(FIELDNAM, "mass of proton");
     m_p->change_xref( FRAME, "scalar>na");
     m_p->change_xref(UNITS, "kg");
     
   //  epsilon 0
     eps_0 = new DvObject(8.854187817e-12);
     eps_0->change_xref(SI_CONVERSION, "1>F m^-1");
     eps_0->change_xref(FIELDNAM, "permittivity of vacuum");
     eps_0->change_xref( FRAME, "scalar>na");
     eps_0->change_xref(UNITS, "C^2 J^-1 m");
     
     // Outputs
 
     DvObject_var wce;
     DvObject_var wci;
     DvObject_var wpe;
     
     DvObject_var out_fce;
     DvObject_var out_fci;
     DvObject_var out_fpe;
     DvObject_var out_fuh;
     DvObject_var out_flh;
     DvObject_var out_fl;
     DvObject_var out_fr;
     
     int num=(hasB? B->seqSize() : Ne->seqSize());
     
	if(hasB)
	{
		wce = B_Tesla * q_e;
		wce /= m_e;
		wci = B_Tesla * q_e;
		wci /= m_p;
             
		out_fce = wce;
		*out_fce /= pi_2;
		out_fci = wci;
		*out_fci /= pi_2;
	}
         
	if(hasNe)
	{
		wpe = Ne_SI*q_e;
		wpe *= q_e;
		wpe /= m_e;
		wpe /= eps_0;
		wpe->sqrtThis();
		
		out_fpe = wpe;
		*out_fpe /= pi_2;
	}
         
	if(hasNe && hasB)
	{
		out_fuh = wpe*wpe;
		out_fuh += wce*wce;
		out_fuh->sqrtThis();
		*out_fuh /= pi_2;
		
		
		out_flh = wpe*wpe;
		out_flh += wce*wci;
		DvObject_var denom = wpe*wpe;
		             denom += wce*wce;
		out_flh /= denom;
		out_flh *= wce*wci;
		out_flh->sqrtThis();
		*out_flh /= pi_2;
		
		out_fl = wpe*wpe;
		*out_fl *= 4. ;
		out_fl += wce*wce;
		out_fl->sqrtThis();
		out_fl -= wce;
		*out_fl /= pi_4;
		
		out_fr = wpe*wpe;
		*out_fr *= 4.;
		out_fr += wce*wce;
		out_fr->sqrtThis();
		out_fr += wce;
		*out_fr /= pi_4;
	}
         
	
	 DvObject_var unitFill = new DvObject(-1.0);
     
     if(hasB)
     {
         out_fce->change_xref(SI_CONVERSION, "1>s^-1" );
		 DvString str_out = "f_ce (from ";
		 str_out += B->getXrefText(FIELDNAM);
		 str_out += ")";
         out_fce->change_xref(LABLAXIS, "f_ce");
         out_fce->change_xref(FILL_VALUE, unitFill );
         out_fce->change_xref(FIELDNAM, str_out);
         out_fce->change_xref(FRAME, "scalar>na");
         out_fce->change_xref(UNITS, "Hz");
     
		 out_fci->change_xref(SI_CONVERSION, "1>s^-1" );
         out_fci->change_xref(LABLAXIS, "f_ci");
         out_fci->change_xref(FILL_VALUE, unitFill );
         str_out.replace("f_ce", "f_ci");
         out_fci->change_xref(FIELDNAM, str_out);
         out_fci->change_xref(FRAME, "scalar>na");
         out_fci->change_xref(UNITS, "Hz");
     }
     
     if(hasNe)
     {
         out_fpe->change_xref(SI_CONVERSION, "1>s^-1" );
         DvString str_out = "f_pe (from ";
		 str_out += Ne->getXrefText(FIELDNAM);
		 str_out += ")";
         out_fpe->change_xref(LABLAXIS, "f_pe");
         out_fpe->change_xref(FILL_VALUE, unitFill );
         out_fpe->change_xref(FIELDNAM, str_out);
         out_fpe->change_xref(FRAME, "scalar>na");
         out_fpe->change_xref(UNITS, "Hz");
     }
 
     if(hasB && hasNe)
     {
         out_fuh->change_xref(SI_CONVERSION, "1>s^-1" );
         DvString str_out = "f_uh (from ";
		 str_out += Ne->getXrefText(FIELDNAM);
		 str_out += " and ";
		 str_out += B->getXrefText(FIELDNAM);
		 str_out += ")";
         out_fuh->change_xref(LABLAXIS, "f_uh");
         out_fuh->change_xref(FILL_VALUE, unitFill );
         out_fuh->change_xref(FIELDNAM, str_out);
         out_fuh->change_xref(FRAME, "scalar>na");
         out_fuh->change_xref(UNITS, "Hz");
     
         out_flh->change_xref(SI_CONVERSION, "1>s^-1" );
         str_out.replace("f_uh", "f_lh");
         out_flh->change_xref(LABLAXIS, "f_lh");
         out_flh->change_xref(FILL_VALUE, unitFill );
         out_flh->change_xref(FIELDNAM, str_out);
         out_flh->change_xref(FRAME, "scalar>na");
         out_flh->change_xref(UNITS, "Hz");
         
         out_fl->change_xref(SI_CONVERSION, "1>s^-1" );
         str_out.replace("f_lh", "f_l");
         out_fl->change_xref(LABLAXIS, "f_l");
         out_fl->change_xref(FILL_VALUE, unitFill );
         out_fl->change_xref(FIELDNAM, str_out);
         out_fl->change_xref(FRAME, "scalar>na");
         out_fl->change_xref(UNITS, "Hz");
     
         out_fr->change_xref(SI_CONVERSION, "1>s^-1" );
         str_out.replace("f_l", "f_r");
         out_fr->change_xref(LABLAXIS, "f_r");
         out_fr->change_xref(FILL_VALUE, unitFill );
		 out_fr->change_xref(FIELDNAM, str_out);
         out_fr->change_xref(FRAME, "scalar>na");
         out_fr->change_xref(UNITS, "Hz");
     }
     
 
     outputs.push_back(out_fce);
     outputs.push_back(out_fci);
     outputs.push_back(out_fpe);
     outputs.push_back(out_fuh);
     outputs.push_back(out_flh);
     outputs.push_back(out_fl);
     outputs.push_back(out_fr);
         
     return QPLUG_SUCCESS;

    
}
