/****************************************************************************
							LICENSE

  Copyright (C) 2012 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file eugen_gradient.cpp
 * For more information, please contact e.sorbalo (AT) jacobs-university DOT de
 *
 * \brief QSAS plug-in for the gradient estimation.
 *
 * This file is a transition between the EugClusterAnalysis class and QSAS.
 *
 */

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "ClusterCommons.h"

using namespace std;

const int VERBOSE = 1;

bool getFieldFromAllSatellites(vector <DvObject_var>& inputs, DvObject_var &tt_out, const bool is_Bfield, size_t nr_spacecrafts,
							 EugClusterAnalysis &nabla);

bool getPositionFromAllSatellites(vector <DvObject_var>& inputs, DvObject_var &tt_out, size_t nr_spacecrafts, EugClusterAnalysis &nabla);


extern "C" QplugReturnStatus  DvEugen_gradient  ( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{
	QplugReturnStatus status = QPLUG_SUCCESS;
	EugClusterAnalysis nabla;
	
	//specifies whether we are working with an array of scalars:
	bool scalars = !(inputs[0]->isThreeVector());  
	if (scalars)
	{
		Status::PrintMessage(Status::kPrintToUserAndConsole, "Input is a field of scalars\n");
	}
	else
	{
		Status::PrintMessage(Status::kPrintToUserAndConsole, "Input is a field of vectors\n");
	}
	
	Status::PrintMessage(Status::kPrintToUserAndConsole, "1. Extracting data\n");
	
	//INPUT 10: number of satellites
	size_t nr_spacecrafts = 0;  //Indicates whether we should apply 3- or 4-satellites technique
	
	nr_spacecrafts = (size_t) inputs[10]->asInt();	
		
	if (nr_spacecrafts !=3 && nr_spacecrafts !=4 )
	{
		Status::PrintMessage(Status::kPrintToUserAndConsole, "Currently %d-spacecraft methods is not supported yet", (int)nr_spacecrafts);
		return QPLUG_FAILURE;
	}
	else 	Status::PrintMessage(Status::kPrintToConsoleOnly, "Operating with %d satellites\n", (int)nr_spacecrafts);
	
	
	//INPUT 11: options to estimate the normal component for 3-satellites data
	if (nr_spacecrafts == 3) {
		unsigned short option_for_estimate = 4;
		//Option choice:
		option_for_estimate = inputs[11]->asInt();	
		
		if (option_for_estimate > 3) {
			Status::PrintMessage(Status::kPrintToUserAndConsole, "Undefined option %u for normal component\n", option_for_estimate);
			return QPLUG_FAILURE;
		}
		nabla.estimator.set_option(option_for_estimate);
		if (VERBOSE) {Status::PrintMessage(Status::kPrintToConsoleOnly, "Option for estimate: %u\n", option_for_estimate); }
	}
	
	short additional_join = 0;
	// ---- INPUT 12: the vector for the normal estimate for 3-sat method
	if (nabla.estimator.get_option() != nabla.estimator.kNone) {
		if (inputs[12].is_ok() && inputs[12]->arraySize() == 3) additional_join++;
	}
	if (VERBOSE) {Status::PrintMessage(Status::kPrintToConsoleOnly, "Additional join: %u\n", additional_join);}

	// ---- Get frame of vectors ------
	DvString frame_V = "";
	DvString frame_i = "";
	if (!scalars)
	{
		frame_V = inputs[0]->getFrameAttr();
		for (size_t i=1; i<nr_spacecrafts; i++){
			frame_i = inputs[i]->getFrameAttr();
			if (frame_i != frame_V){
				QplugAppendTextDisplay("Inputs not in same vector frame:");
				QplugAppendTextDisplay(frame_i.c_str());
				QplugAppendTextDisplay(" and ");
				QplugAppendTextDisplay(frame_V.c_str());
				return QPLUG_FAILURE;
			}
		}
		
		if ((nr_spacecrafts==3) && (nabla.estimator.get_option() != nabla.estimator.kNone) ){
			frame_i = inputs[12]->getFrameAttr();
			if (frame_i != frame_V) {
				QplugAppendTextDisplay("Vector for the normal component estimation is not in same vector frame:");
				QplugAppendTextDisplay(frame_i.c_str());
				QplugAppendTextDisplay(" and ");
				QplugAppendTextDisplay(frame_V.c_str());
				return QPLUG_FAILURE;
			}
		}
	}
	
	//Frame for satellites position vectors:
	DvString frame_P = inputs[4]->getFrameAttr();  //first position sequence
	{
		for (size_t i=1; i<nr_spacecrafts; i++){
			frame_i = inputs[i+4]->getFrameAttr();  //only position vectors
			if (frame_i != frame_P) {
				QplugAppendTextDisplay("Satellites' position not in same vector frame:");
				QplugAppendTextDisplay(frame_i.c_str());
				QplugAppendTextDisplay(" and ");
				QplugAppendTextDisplay(frame_P.c_str());
				return QPLUG_FAILURE;
			}
		}
	}

	bool is_Bfield = false;
	if(inputs[0]->sameBaseSI( "1.0e-9>T") ) is_Bfield = true;
	DvString SI_V = inputs[0]->getBaseSI();

	DvString SI_i = "";
	for (size_t i = 1; i < nr_spacecrafts; i++) {
		SI_i = inputs[i]->getBaseSI();
		if (SI_i.empty()) continue;
		
		if ( SI_i != SI_V ) {
			QplugAppendTextDisplay("Inputs not in same base units:");
			QplugAppendTextDisplay(SI_i.c_str());
			QplugAppendTextDisplay(" and ");
			QplugAppendTextDisplay(SI_V.c_str());
			return QPLUG_FAILURE;
		}
	}

	DvString SI_R = inputs[4]->getBaseSI();
	for (int i = 5; i < 8; i++) {
		SI_i = inputs[i]->getBaseSI();
		if (SI_i.empty()) continue;

		if (SI_i != SI_R) {
			QplugAppendTextDisplay("Inputs not in same base units:");
			QplugAppendTextDisplay(SI_i.c_str());
			QplugAppendTextDisplay(" and ");
			QplugAppendTextDisplay(SI_R.c_str());
			return QPLUG_FAILURE;
		}
	}

	//	..first, get input vectors from each SC and join ..
	DvObject_var ttJoined = inputs[2]->getTimeTags();    // tt of joined V and Pos from SC3 vec	
	
	// ---- INPUT 8: time interval
	DvEvent start_end = inputs[8]->getTimeRange();
	DvObject_var tt_out = ttJoined->subSequence(start_end);

	nabla.deltat_ = tt_out->get_spacing(); 


	Status::PrintMessage(Status::kPrintToConsoleOnly, "delta t = %f  ->  gap_fill if t > %f\n", nabla.deltat_, nabla.deltat_*1.5);
	
	
	// ---- INPUT 0-3 : get the measurements of the field as measured by all satellites -------------
	if( !getFieldFromAllSatellites(inputs, tt_out, is_Bfield, nr_spacecrafts, nabla) ) return QPLUG_FAILURE;

	// ---- INPUT 4-7 : get position for given time interval  ---------------------------
	if (!getPositionFromAllSatellites(inputs, tt_out,  nr_spacecrafts, nabla)) return QPLUG_FAILURE;



	// ---- INPUT 9: get scalefactor for Curl and check if this is a current (scale = mu_0 so that J = curl B / mu_0) ----
	DvString curlunit = "";
	DvString scaleunit = "";
	double scalefac = 1.0;
	bool iscurrent = false;
		
	scalefac = inputs[9]->asDouble();

	curlunit = inputs[2]->getXrefText(UNITS);
	curlunit += "/";
	curlunit += inputs[6]->getXrefText(UNITS);

	if (scalefac != 1.0) {
		scaleunit = inputs[9]->getXrefText(UNITS);
		if (!scaleunit.empty()){
			curlunit += "/";
			curlunit += scaleunit;
			printf("> Scalefactor supplied : %6.4g %s\n", scalefac, scaleunit.c_str());
			fflush(stdout);
			if (curlunit == "nT/km/H/m") iscurrent = true;
			//sprintf(curlunit, "A/m^2");
			//scalefac *= 1.0E12; // divide (10^-9 nT / 10^3 m) to get current in A/m2
		}
	}
	else {
		printf("> Curl unscaled\n");
		fflush(stdout);
	}
	
	// ---- INPUT 12: the vector for the normal estimate for 3-sat method
		
	if (nabla.estimator.get_option() != nabla.estimator.kNone) {
        double avg_width = 1.5 * nabla.deltat_;
	
		//Extract the vector for the estimate of the normal component:
		nabla.estimator.qvector = inputs[12];
		
		if(nabla.estimator.qvector.is_nil() || nabla.estimator.qvector->arraySize() != 3){
			QplugAppendTextDisplay("Optional vector is not a vector.\n");
			return QPLUG_FAILURE;
		}
		
		QplugAppendTextDisplay("Vector for estimating the normal component.\n");
		printf("Estimate vector[0]: %f  %f  %f\n", nabla.estimator.qvector->dbl(0),
							 nabla.estimator.qvector->dbl(1), nabla.estimator.qvector->dbl(2));
		fflush(stdout);
		
		if(nabla.estimator.qvector->seqSize() > 1) {
			nabla.estimator.qvector->change_xref(DV_GAP_WIDTH, avg_width);
			nabla.estimator.qvector->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
			nabla.estimator.qvector->change_xref(DV_GAP_OPTION, DV_LINEAR);	
		
			nabla.estimator.qvector = nabla.estimator.qvector->Join(tt_out, true); // includes xrefs
			if( nabla.estimator.qvector.is_nil() )
			{
				QplugAppendTextDisplay("Normal series has no overlap with target time interval");
				return QPLUG_FAILURE;
			}
		}
	}
	
	// ---------- now, calculate curl, div and grad for each time step ---------
	QplugAppendTextDisplay("2. Computing reciprocal vectors and gradient.\n");

	if (nabla.ComputeGradientEstimate(nr_spacecrafts) != Status::kOk) return QPLUG_FAILURE;

	Status::PrintMessage(false, "NablaEstimator: Computing curl and divergence.\n");

	if (nabla.nr_coordinates() == kVectorRank)
		if (nabla.ComputeCurlAndDiv(scalefac) != Status::kOk) return QPLUG_FAILURE;

	Status::PrintMessage(false, "NablaEstimator: Computing barycenter.\n");
	
	if (nabla.ComputeBarycenter() != Status::kOk) return QPLUG_FAILURE;
	
	if (nabla.nr_coordinates() == kVectorRank) {
		Status::PrintMessage(false, "NablaEstimator: Computing MVA.\n");
		if (nabla.Postprocess() != Status::kOk) return QPLUG_FAILURE;
	}
	

	// -------- send some outputs ---------------------------
	Status::PrintMessage(true, "NablaEstimator: Preparing the output.\n");

	DvString divunit = "";
	DvString reciprocunit = "";
	divunit =  inputs[2]->getXrefText(UNITS);
	divunit += "/";
	divunit += inputs[6]->getXrefText(UNITS);
	
	reciprocunit = "1/";
	reciprocunit += inputs[6]->getXrefText(UNITS);

	Status::PrintMessage(false, "NablaEstimator: find inverse of SI_conversion\n");
	// SI_conversion: find inverses first
	DvString attr_inv_ptr = inputs[4]->getSICInverse();
	DvObject_var km_inv_ptr = new DvObject(); // this is empty as we only need metadata
	if( !attr_inv_ptr.empty() ){
		km_inv_ptr->change_xref(SI_CONVERSION, attr_inv_ptr);
	}
		
	//In the following 6 lines we read the (char *) from functions and convert them to safe strings.
	DvString si_convers = inputs[0]->getSICProduct(*km_inv_ptr);  // T/m

	// ---- correct the SI conversion of the curl by considering the scale factor. --------------
	attr_inv_ptr = inputs[9]->getSICInverse();
	DvObject_var scalefactor_inv_ptr = new DvObject(); // this is empty as we only need metadata
	if( !attr_inv_ptr.empty() ){
		scalefactor_inv_ptr->change_xref(SI_CONVERSION, attr_inv_ptr);
	}
	DvObject_var curl_temp_obj_ptr = new DvObject(); // this is empty as we only need metadata
	curl_temp_obj_ptr->change_xref(SI_CONVERSION, si_convers);
	DvString si_convers_curl = curl_temp_obj_ptr->getSICProduct(*scalefactor_inv_ptr);
	curl_temp_obj_ptr->change_xref(SI_CONVERSION, si_convers_curl);
	
	//Status::PrintMessage(false, "NablaEstimator curl unit: %s\n unit: %s\n"
	//		, si_convers_curl.c_str(), curlunit);
	// ------ analyze if we have current --------------------------------------------------------
	//1e-12 > m^-2 A
	
	if(curl_temp_obj_ptr->sameBaseSI("m^-2 A")) iscurrent = true;

	// ------------------------------------------------------------------------------------------

	DvString frame = (scalars)? frame_P : frame_V;
	DvString curl_lablaxis = (iscurrent)? "J" : "Curl V";
				
	// ------------------------------------------------------------------------------------------
	Status::PrintMessage(false, "NablaEstimator: preparing output of gradient\n");
	if (nabla.grad_TS_.is_ok())
	{
		nabla.grad_TS_->change_xref(FRAME, frame);
		nabla.grad_TS_->change_xref(UNITS, divunit);
		nabla.grad_TS_->change_xref(SI_CONVERSION, si_convers);
		nabla.grad_TS_->change_xref("LABLAXIS","Grad V");
		nabla.grad_TS_->change_xref(DEPEND_0, tt_out);
	}

	Status::PrintMessage(false, "NablaEstimator: preparing output of gradient's magnitude\n");
	if (nabla.grad_magn_TS_.is_ok())
	{
		nabla.grad_magn_TS_->change_xref(FRAME, frame);
		nabla.grad_magn_TS_->change_xref(UNITS,divunit);
		nabla.grad_magn_TS_->change_xref(SI_CONVERSION, si_convers);
		nabla.grad_magn_TS_->change_xref("LABLAXIS","Grad magnitude V");
		nabla.grad_magn_TS_->change_xref(DEPEND_0, tt_out);
	}
		
	Status::PrintMessage(false, "NablaEstimator: preparing output of curl\n");
	if (nabla.curl_TS_.is_ok())
	{
		nabla.curl_TS_->change_xref(FRAME, frame);
		nabla.curl_TS_->change_xref(UNITS,curlunit);
		nabla.curl_TS_->change_xref(SI_CONVERSION, si_convers_curl);
		nabla.curl_TS_->change_xref("LABLAXIS", curl_lablaxis);
		nabla.curl_TS_->change_xref(DEPEND_0, tt_out);
	}
		
	Status::PrintMessage(false, "NablaEstimator: preparing output of curl in MVA\n");
	if (nabla.curl_TS_MVA_.is_ok())
	{
		nabla.curl_TS_MVA_->change_xref(FRAME, frame);
		nabla.curl_TS_MVA_->change_xref(UNITS,curlunit);
		nabla.curl_TS_MVA_->change_xref("LABLAXIS", curl_lablaxis);
		nabla.curl_TS_MVA_->change_xref(SI_CONVERSION, si_convers_curl);
		nabla.curl_TS_MVA_->change_xref(DEPEND_0, tt_out);
	}
		
	Status::PrintMessage(false, "NablaEstimator: preparing output of barycenter\n");
	if (nabla.bary_vector_TS_.is_ok())
	{
		nabla.bary_vector_TS_->copy_xref_from(UNITS, inputs[0]);
		nabla.bary_vector_TS_->copy_xref_from(SI_CONVERSION, inputs[0]);
		nabla.bary_vector_TS_->change_xref(LABLAXIS, "Field in barycenter");
		nabla.bary_vector_TS_->change_xref(FRAME, frame);
		nabla.bary_vector_TS_->change_xref(DEPEND_0, tt_out);
	}
	
	if (nabla.bary_scalar_TS_.is_ok())
	{
		nabla.bary_scalar_TS_->copy_xref_from(UNITS, inputs[0]);
		nabla.bary_scalar_TS_->copy_xref_from(SI_CONVERSION, inputs[0]);
		nabla.bary_scalar_TS_->change_xref(LABLAXIS, "Field in barycenter");
		nabla.bary_scalar_TS_->change_xref(DEPEND_0, tt_out);
	}

	Status::PrintMessage(false, "NablaEstimator: preparing output of barycenter position\n");
	if (nabla.bary_coordinates_TS_.is_ok())
	{
		nabla.bary_coordinates_TS_->copy_xref_from(UNITS, inputs[4]);
		nabla.bary_coordinates_TS_->copy_xref_from(SI_CONVERSION, inputs[4]);
		nabla.bary_coordinates_TS_->change_xref(LABLAXIS, "Barycenter position");
		nabla.bary_coordinates_TS_->change_xref(FRAME, frame_P);
		nabla.bary_coordinates_TS_->change_xref(DEPEND_0, tt_out);
	}
		
	Status::PrintMessage(false, "NablaEstimator: preparing output of divergence\n");
	if (nabla.div_TS_.is_ok())
	{
		nabla.div_TS_->change_xref(FRAME, "scalar>na");
		nabla.div_TS_->change_xref(UNITS,divunit);
		nabla.div_TS_->change_xref("LABLAXIS", "Div V");
		nabla.div_TS_->change_xref(SI_CONVERSION, si_convers);
		nabla.div_TS_->change_xref(DEPEND_0, tt_out);
	}
		
	Status::PrintMessage(false, "NablaEstimator: preparing output of volume/surface \n");
	if (nabla.vol_TS_.is_ok())
	{
		DvString strunit = inputs[6]->getUnitsProduct(*(inputs[6]));
		DvString str_si_conv = inputs[6]->getSICProduct(*(inputs[6]));
		if (nr_spacecrafts == 3){
			nabla.vol_TS_->change_xref(LABLAXIS,"TriangleSurface");
		}
		else{
			nabla.vol_TS_->change_xref(LABLAXIS,"TetraVolume");
			DvObject_var km2_ptr = new DvObject(); // this is empty as we only need metadata
			km2_ptr->change_xref(UNITS,strunit);
			km2_ptr->change_xref(SI_CONVERSION, str_si_conv);
			strunit = inputs[6]->getUnitsProduct(*km2_ptr);  //km^3
			str_si_conv = inputs[6]->getSICProduct(*km2_ptr);  //km^3
		}			
		
		nabla.vol_TS_->change_xref(FRAME, "scalar>na");
		nabla.vol_TS_->change_xref(UNITS,strunit);
		nabla.vol_TS_->change_xref(SI_CONVERSION, str_si_conv);
		nabla.vol_TS_->change_xref(DEPEND_0, tt_out);
		Status::PrintMessage(false, "NablaEstimator: cleaning after output of volume/surface \n");
		
		if(nabla.vol_TS_->nError() > 0) {
			Status::PrintMessage(true, nabla.vol_TS_->getError().c_str() );
			Status::PrintMessage(true, "------------------\n" );
		}

	}

	Status::PrintMessage(false, "NablaEstimator: preparing output of normals \n");
	if (nabla.normal_TS_.is_ok())
	{
		nabla.normal_TS_->copy_xref_from(UNITS, inputs[6]);
		nabla.normal_TS_->copy_xref_from(SI_CONVERSION, inputs[6]);
		nabla.normal_TS_->change_xref(LABLAXIS, "NormalToPlane");
		nabla.normal_TS_->change_xref(DEPEND_0, tt_out);
		nabla.normal_TS_->change_xref(FRAME, frame_P);
	}
		
	Status::PrintMessage(false, "NablaEstimator: preparing output of velocity in MVA \n");
	if (nabla.velo_TS_MVA_.is_ok())
	{
		///\todo Add SI_CONVERSION
		if(iscurrent) nabla.velo_TS_MVA_->change_xref(UNITS,"km/s" );
		nabla.velo_TS_MVA_->change_xref(LABLAXIS, "J");
		nabla.velo_TS_MVA_->change_xref(DEPEND_0, tt_out);
		nabla.velo_TS_MVA_->change_xref(FRAME, frame);
	}

	Status::PrintMessage(false, "NablaEstimator: output to QSAS.\n");
	outputs.push_back(nabla.grad_TS_);
	outputs.push_back(nabla.grad_magn_TS_);
	outputs.push_back(nabla.curl_TS_);
	outputs.push_back(nabla.curl_TS_MVA_);
	outputs.push_back(nabla.rotation_matrix_);
	if (nabla.bary_vector_TS_.is_ok())
		outputs.push_back(nabla.bary_vector_TS_);
	else
		outputs.push_back(nabla.bary_scalar_TS_);
	outputs.push_back(nabla.bary_coordinates_TS_);
	outputs.push_back(nabla.div_TS_);
	outputs.push_back(nabla.vol_TS_);
	outputs.push_back(nabla.normal_TS_);

	return status;
}

// -------------------------------------------------------------


bool getFieldFromAllSatellites(vector <DvObject_var>& inputs, DvObject_var &tt_out, const bool is_Bfield, size_t nr_spacecrafts,
							 EugClusterAnalysis &nabla)
{

	// ---- get Vector quantity for given time interval - store in nabla.field_TS_ ----
	int nestsize;
	double avg_width = 1.5 * nabla.deltat_;
	if(avg_width < 1.e-20){
		QplugAppendTextDisplay("Cannot estimate time series spacing, too small");
		return false;
	}

	for (size_t satellite=0; satellite<nr_spacecrafts; satellite++)
	{
		DvObject_var obj = inputs[satellite];
		obj->change_xref(DV_GAP_WIDTH, avg_width);
		obj->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
		obj->change_xref(DV_GAP_OPTION, DV_LINEAR);	

	
		obj = obj->Join(tt_out, true); // includes xrefs
		

		if( obj.is_nil() )
		{
			QplugAppendTextDisplay("Series has no overlap with target time interval");
			return false;
		}

		if(obj->arraySize() == 3) obj->ensureVectorXYZ();
		
		if(is_Bfield){
			obj = obj->changeUnitsTo("1.0e-9>T", "nT"); // ensure units are in Cluster units assumed
		}
		
		if (satellite==0)
		{
			nestsize = obj->seqSize();
			vector <size_t> dims;
			dims.push_back(nr_spacecrafts);
			dims.push_back(obj->arraySize());
				
			nabla.field_TS_ = new DvObject(0.0, dims, nestsize);
			nabla.field_TS_->copy_xrefs_from(obj);
		}

		for (int t=0; t<nestsize; t++) {
			for (int i=0; i<obj->arraySize(); i++) {
				nabla.field_TS_->dbl(t,satellite,i) = obj->dbl(t,i);
			}
		}
	}

	return true;
}


bool getPositionFromAllSatellites(vector <DvObject_var>& inputs, DvObject_var &tt_out, size_t nr_spacecrafts, EugClusterAnalysis &nabla)
{
	// ---- get position for given time interval - store in nabla.position_TS_ ----
	int nestsize;
	double avg_width = 1.5 * nabla.deltat_;
	
	for (unsigned short satellite=0; satellite<nr_spacecrafts; satellite++)
	{
		unsigned short index_cor = satellite+4;

		DvObject_var obj = inputs[index_cor];
		obj->change_xref(DV_GAP_WIDTH, avg_width);
		obj->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
		obj->change_xref(DV_GAP_OPTION, DV_LINEAR);	
		
		obj = obj->Join(tt_out, true); // includes xrefs
		
		if( obj.is_nil() )
		{
			QplugAppendTextDisplay("Position series has no overlap with target time interval");
			return false;
		}

		if(obj->arraySize() == 3) obj->ensureVectorXYZ();
		
		obj = obj->changeUnitsTo("1.0e3>m", "km"); // ensure units are in Cluster units assumed
		
		if (satellite==0)
		{
			nestsize = obj->seqSize();
			vector <size_t> dims;
			dims.push_back(nr_spacecrafts);
			dims.push_back(3);
				
			nabla.position_TS_ = new DvObject(0.0, dims, nestsize);
			nabla.position_TS_->copy_xrefs_from(obj);
		}

		for (int t=0; t<nestsize; t++) {
			for (int i=0; i<3; i++) {
				nabla.position_TS_->dbl(t,satellite,i) = obj->dbl(t,i);
			}
		}
	}

	return true;

}

