
// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for MER - minimum entropy residue
//
// SEH, Sep 7, 2003
//  
// Inputs :
//     B = magnetic field  (n x 3)     [nT]
//     V = plasma velocity (n x 3)     [km/s]
//     P = pressure tensor (n x 3 x 3) [eV/cm^3]
//     N = plasma density  (n)         [1/cm^3]  (proton mass used to get mass density)
//
// For simplification, all units are converted to SI units
//
// -----------------------------------------------------------------------------------------

#include "GenericResidue.h"
#include "DvObj.h"


// -----------------------------------------------------------------------------------------
//
// Returns the Q_matrix for MER (Minimum Energy Residue)
//
// SEH, 19 Sep, 2001
//      15 Oct, 2003, Doxygen DOC
//
// Doxygen info :
/*! 

  \brief Returns the covariance matrix for the Minimum Energy Residue 
  method.

  \param V - vector time series with plasma velocity in units of km/s
  \param B - vector time series with the magnetic field in units of nT
  \param P - matrix time series with the full 3x3 pressure tensor
  \param N - scalar time series with the plasma density in units of 1/cm3
  \param U - vector - holds the calculated velocity in units of km/s on successful return.
  \param nestsize - number of elements in V,B,P,and N.

  \warning If the density is constant, there will be a sigularity in the 
  U calculation. U is then replaced by <V> - the average plasma
  velocity.
*/ 
// -----------------------------------------------------------------------------------------
DvObject_var Q_MER_calc(DvObject_var &V, DvObject_var &B, DvObject_var &P,
			 DvObject_var &N, DvObject_var &U, int nestsize)
{
  int i,j,k,m;
  double mu0 = 12.566370614e-07;             // magnetic constant [N/A^2 == H/m]
  double m_p = 1.67e-27;                     // mass of proton [kg] (we assume ions are == protons)
  double normfac = (double) nestsize;
  double p,rho,ave_eta=0.00,d_eta2=0.00;
  bool PO = true;                           // PrintOut to screen ? 

  DvObject_var    V_k = new DvObject(0.0, DIM3);                    // holds the V(k) vector
  DvObject_var q_i = new DvObject(0.0,DIM3,nestsize);
  DvObject_var    ave_q_i = new DvObject(0.0, DIM3);
  DvObject_var eta = new DvObject(0.0, nestsize);
  DvObject_var    Q =  new DvObject(0.0, DIM33);


  for (k=0; k<nestsize; k++) {
    p = 0.0;
    for (i=0; i<3; i++) {
      V_k->dbl(i) = V->dbl(k,i) * 1E3;                     
      p += P->dbl(k,i,i)/3.0 * 1.609E-13;                    // scalar pressure [ kg/ms^2 ]
    }
    rho = m_p * N->dbl(k) * 1E6;                                        // 1E6 -> to get SI units kg/m^3 
    eta->dbl(k) = rho * log(p / pow(rho,(5.0/3.0)));
    ave_eta += eta->dbl(k);
    for (i=0; i<3; i++) {                                             // loop over three components
      q_i->dbl(k,i) = eta->dbl(k) *  V_k->dbl(i);             
      ave_q_i->dbl(i) += q_i->dbl(k,i);
    }
  }
  


  // ----- initialize U, get <q_i> and <eta>--------

  for (i=0; i<3; i++) {
    ave_q_i->dbl(i) /= normfac;
    U->dbl(i) = 0.0;
  }
  ave_eta /= normfac;

  if (PO) printf("Q_MER:\nq_i vector [kg/(m^2 s)] : \n");
  if (PO) printf("<q_j>     = [%E  %E  %E]; \n",ave_q_i->dbl(0), ave_q_i->dbl(1), ave_q_i->dbl(2));
  if (PO)
  {
	printf("\n<eta>      = %E [kg/m^3]\n", ave_eta);
	fflush(stdout);
  }

  // ---   get delta_q_i = q_i - <q_i> and delta_eta --------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) q_i->dbl(k,i) -= ave_q_i->dbl(i);
    eta->dbl(k) -= ave_eta;
    d_eta2 += (eta->dbl(k) * eta->dbl(k));
  }


  if (PO)
  {
	printf("\n<|deta|^2> = %E  [kg^2/(m^2s^4)]\n", d_eta2);
	fflush(stdout);
  }

  // ---- construct U* vector ---------------

  if (d_eta2 != 0.0) {
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) U->dbl(i) += eta->dbl(k) * q_i->dbl(k,i) / d_eta2; 
      }  
  } else {
     printf("MER : singularity - setting U* = <V_cis>\n");
	 fflush(stdout);
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) U->dbl(i) +=  V->dbl(k,i) * 1E3 / normfac;       
      }
  }

  if (PO)
  {
	printf("\nU*         = [%E  %E  %E];  [m/s] (eq 12 in manuscript)\n\n", U->dbl(0), U->dbl(1), U->dbl(2));
	fflush(stdout);
  }


  // --- construct Q_ij matrix --------------  

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          Q->dbl(0,i,j) += (q_i->dbl(k,i) - U->dbl(i)*eta->dbl(k)) * (q_i->dbl(k,j) - U->dbl(j)*eta->dbl(k));
      }
    }
  }
  

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      Q->dbl(0,i,j) /= normfac;
    }
  }

  if (PO)printf("Q_ij  [kg/(m^2 s)] (eq 15 in manuscript)\n");
  for (i=0; i<3; i++) {
    if (PO) printf("Q_i%d       = [%E  %E  %E];\n",i, Q->dbl(0,i,0), Q->dbl(0,i,1), Q->dbl(0,i,2));
	fflush(stdout);
  } 

  return (Q);
}

