
// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for MTER - minimum total energy residue
//
// SEH, Sep 7, 2003
// 
// Inputs :
//     B = magnetic field   (n x 3)     [nT]
//     V = plasma velocity  (n x 3)     [km/s]
//     P = pressure tensor  (n x 3 x 3) [eV/cm^3]
//     N = plasma density   (n)         [1/cm^3]  (proton mass used to get mass density)
//     H = heat flux vector (n x 3)  
//
// For simplification, all units are converted to SI units
//
// -----------------------------------------------------------------------------------------

#include "GenericResidue.h"
#include "DvObj.h"


// -----------------------------------------------------------------------------------------
//
// Returns the Q_matrix for MTER (Minimum Total Energy Residue)
//
// SEH, 07 Sep, 2001
//      19 Sep, 2001, bug fixes - U returned
//      15 Oct, 2003, Doxygen DOC
//      18 May, 2005, quality control through manual calc by BS
//
// Doxygen info :
/*! 
  \brief Returns the covariance matrix for the Minimum Total Energy Residue 
  (MTER) method.
.
  \param V - vector time series with plasma velocity in units of km/s
  \param B - vector time series with the magnetic field in units of nT
  \param P - matrix time series with the full 3x3 pressure tensor
  \param E - vector time series with the electric field in units of mV/m
  \param N - scalar time series with the plasma density in units of 1/cm3
  \param H - vector time series heat flux vector in units of erg/s cm3
  \param U - vector - holds the calculated velocity in units of km/s on successful return.
  \param nestsize - number of elements in V,B,P,E,N and H.

*/ 
// -----------------------------------------------------------------------------------------
DvObject_var Q_MTER_calc(DvObject_var &V, DvObject_var &B, DvObject_var &P,
			  DvObject_var &E, DvObject_var &N, DvObject_var &H,
			  DvObject_var &U ,int nestsize)

{
  int i,j,k;
  double m_p = 1.67e-27;                     // mass of proton [kg] (we assume ions == protons)
  double Mu0  = 12.566370614e-07;  
  double normfac = (double) nestsize;
  double p,p_ave = 0.0, rhov2=0.0, d_eta2=0.00, ave_eta=0.00;
  bool PO = false;                          // PrintOut to screen ?

  DvObject_var    E_k = new DvObject(0.0, DIM3);                    // holds the E^2(k) vector
  DvObject_var    V_k = new DvObject(0.0, DIM3);                    // holds the V^2(k) vector
  DvObject_var    B_k = new DvObject(0.0, DIM3);                    // holds the B^2(k) vector
  DvObject_var    v_P = new DvObject(0.0, DIM3);                    // holds the v_P(k) vector
  DvObject_var    H_k = new DvObject(0.0, DIM3);                    // holds the heatflux vector
  DvObject_var    S =   new DvObject(0.0, DIM3);                    // holds the Poynting flux vector
  DvObject_var eta = new DvObject(0.0, nestsize);
  DvObject_var    Q =   new DvObject(0.0, DIM33);
  DvObject_var q_i = new DvObject(0.0, DIM3, nestsize);
  DvObject_var    ave_q_i = new DvObject(0.0, DIM3);

  p_ave = aveTotalPressure(B, P, nestsize);
  if (PO)
  {
	printf("Q_MTER > p_ave = %10.4g\n",p_ave);
	fflush(stdout);
  }

  if (PO) {
     printf(" k           V_ * P_ii/2                       V_ * rho V^2/2                     V_ dot P__ ");
     printf("                      Heat flux                     S (Poynting flux)\n");
     for (i=0; i<169; i++) printf("=");
     printf("\n");
	 fflush(stdout);
  }

  for (k=0; k<nestsize; k++) {
    //   if (k > 9) m_p = 2.0e-27;
    p = 0.00;
    for (i=0; i<3; i++) {
      V_k->dbl(i) = V->dbl(k,i) * 1.0E3;                                                         // V vector in [m/s]
      E_k->dbl(i) = E->dbl(k,i) * 1.0E-3;                                                        // E field in  [V/m]
      B_k->dbl(i) = B->dbl(k,i) * 1.0E-9;                                                        // B field vec [T]
      H_k->dbl(i) = H->dbl(k,i) * 1.0E-3;                                                        // heat flux in [J/m^2 s] = [kg /s^3]
      v_P->dbl(i) = 0.00;                                                                       // initialize to 0.00
      for (j=0; j<3; j++) v_P->dbl(i) += V_k->dbl(j) * (P->dbl(k,i,j) * 1.60219E-13);           // 1.60219E-13 for eV/cm3 to J
// Experiment  
//      for (j=0; j<3; j++) (*v_P)[i] += (*V_k)[j] * ( vecsum2(B_k)/(2.0*Mu0) - p_ave);         // 1.60219E-13 for eV/cm3 to J
      p += (P->dbl(k,i,i))/3.0 * 1.60219E-13;                                         // scalar pressure = trace(P)/3 [ kg/ms^2 ]
    }
    S = B_k->vec(V_k);
    S = S->vec(B_k);                                            // S = E x B == (-V x B) x B = (B x V) x B [V/m * nT] = [kg/s^3];
// experiment :
//    p = p_ave - (vecsum2(B_k)/(2.0*Mu0));
    rhov2 = m_p * N->dbl(k) * 1E6 * vecsum2(V_k);                                     // 1E6 due to density [1/cm^3] -> [1/m^3]
    eta->dbl(k) = (3.0*p + rhov2 + vecsum2(B_k)/Mu0) / 2.0;
    ave_eta += eta->dbl(k);
    
    for (i=0; i<3; i++) {
      q_i->dbl(k,i) = V_k->dbl(i) * (3.0*p/2.0 + rhov2/2.0) + v_P->dbl(i) + H_k->dbl(i) + S->dbl(i)/Mu0;         // see eq. 37
      ave_q_i->dbl(i) += q_i->dbl(k,i);
    }
 
//    printf("%2d   %8.3e ", k+1,p);

    if (PO) {
     printf("%2d   %8.1e  %8.1e  %8.1e      ", k+1,3*p/2.0*V_k->dbl(0), 3*p/2.0*V_k->dbl(1), 3*p/2.0*V_k->dbl(2));
     printf("%8.1e  %8.1e  %8.1e      ",rhov2/2.0*V_k->dbl(0),rhov2/2.0*V_k->dbl(1), rhov2/2.0*V_k->dbl(2));
     printf("%8.1e  %8.1e  %8.1e      ",v_P->dbl(0), v_P->dbl(1), v_P->dbl(2));
     printf("%8.1e  %8.1e  %8.1e      ",H_k->dbl(0), H_k->dbl(1), H_k->dbl(2));
     printf("%8.1e  %8.1e  %8.1e\n",    S->dbl(0)/Mu0,S->dbl(1)/Mu0, S->dbl(2)/Mu0);
	 fflush(stdout);
    }
  }

  if (PO)
  {
	printf("\n\n");
	fflush(stdout);
  }

  if (PO) {
   printf(" k     eta                 q_i \n");
   for (i=0; i<44; i++) printf("=");
   printf("\n");
   for (k=0; k<nestsize; k++) {
     printf("%2d   %9.2e     %9.2e  %9.2e  %9.2e\n", k+1,eta->dbl(k), q_i->dbl(k,0), q_i->dbl(k,1),q_i->dbl(k,2));
    }
	fflush(stdout);
  }

  if (PO)
  {
	printf("\n\n");
	fflush(stdout);
  }


  // ----- initialize U, get <q_i> and <eta>--------

  for (i=0; i<3; i++) {
    ave_q_i->dbl(i) /= normfac;
    U->dbl(i) = 0.0;
  }
  ave_eta /= normfac;

  if (PO) printf("Q_MTER:\nq_i vector [kg/s^3] : \n");
  if (PO) printf("<q_j>     = [%E  %E  %E]; \n",ave_q_i->dbl(0), ave_q_i->dbl(1), ave_q_i->dbl(2));
  if (PO) printf("\n<eta>      = %E [kg/ms^2]\n", ave_eta);
  if (PO) fflush(stdout);

  // ---   get delta_q_i = q_i - <q_i> and delta_eta --------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) q_i->dbl(k,i) -= ave_q_i->dbl(i);
      eta->dbl(k) -= ave_eta;
      d_eta2 += eta->dbl(k) * eta->dbl(k);
  }


  if (PO)
  {
	printf("\n<|deta|^2> = %E  [kg^2/(m^2s^4)]\n", d_eta2);
	fflush(stdout);
  }

  // ---- construct U* vector ---------------

  if (d_eta2 != 0.0) {
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) U->dbl(i) += eta->dbl(k) * q_i->dbl(k,i) / d_eta2;
      }  
  } else {
     printf("MTER: singularity - setting U* = <V_cis>\n");
	 fflush(stdout);
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) U->dbl(i) += V->dbl(k,i) * 1E3 / normfac;
      }
  }

  if (PO)
  {
	printf("\nU*         = [%E  %E  %E];  [m/s] (eq 12 in manuscript)\n\n", U->dbl(0), U->dbl(1), U->dbl(2));
	fflush(stdout);
  }


  // --- construct Q_ij matrix --------------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          Q->dbl(0,i,j) += (q_i->dbl(k,i) - U->dbl(i)*eta->dbl(k)) * (q_i->dbl(k,j) - U->dbl(j)*eta->dbl(k));
      }
    }
  }
  

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      Q->dbl(0,i,j) /= normfac;
    }
  }

  if (PO)  printf("Q_ij  [kg/(s^3)] (eq 15 in manuscript)\n");
  if (PO)  for (i=0; i<3; i++) {
    printf("Q_i%d       = [%19.15E  %19.15E  %16.15E];\n",i, Q->dbl(0,i,0), Q->dbl(0,i,1), Q->dbl(0,i,2));
  }
 if(PO)  fflush(stdout);

 return(Q);
}
