// -----------------------------------------------------------------------------------------
//
// Performs a generic Analysis of a Single Nest size
//
// SEH, Sep 16, 2003
// 
// Inputs :
//     B = magnetic field   (n x 3) 
//     V = plasma velocity  (n x 3) 
//     E = electric field   (n x 3)
//     P = pressure tensor  (n x 3 x 3)
//     Q = heat flux vec    (n x 3)
//     N = density          (n)
//     nestsize = number of points in data sets
//     constrain = 1 for Bn=0, 2 for (<V> - U)n etc - see below.
//     alpha = slope of walen correlation coeff (used for Alfvenic constrain)
//
// SEH, 12 Sep 2003
//      06 Nov 2003, added SCID and constrain type on output
//      14 May 2005, added alpha - the Walenslope for constraining
//      31 May 2006, added MVA dB/dt
//      08 Aug 2008, disable Spreadsheet out
//      20 Jan 2010, catch errors if nestsize < 10, warning if no SI_conversion
// -----------------------------------------------------------------------------------------
#include "GenericMinVar.h"
#include "GenericResidue.h"

int SingleAnalysis(DvObject_var &V, DvObject_var &B, DvObject_var &dB, DvObject_var &P, 
		   DvObject_var &E, DvObject_var &N, DvObject_var &H, 
		   DvObject_var &ttJoined,  DvObject_var &ExWeights,
                   int constrained, double alpha, DvString SCc_ID, bool doType[10],int nestsize,
                   vector <DvObject_var>& outputs)
{

  printf("SingleGeneric > \n");
  bool GenerateSpreadSheet = false;
  bool GeneratePolarOutput = false;
  FILE *f1;
  int i,j,k;
  DvObject_var ProjMx  = new DvObject(0.0, DIM33);
  DvObject_var Evec = new DvObject(0.0, DIM33);  
  DvObject_var EvecMVAB;      // save MVAB eigenvecs for later use in constrain
  DvObject_var Evalues;
  DvObject_var Dummy =  new DvObject(0.0, DIM3);  
  DvObject_var U = new DvObject(0.0, DIM3);        
  char buf[255],typebuf[16], constrbuf[4]="";
  double un=0.0, vn,van,bn,ucis=0.0,deltat,N_ave=0.0;
 

  DvTime t_start = ttJoined->time(0);
  DvTime t_stop = ttJoined->time(nestsize-1);
  DvTime t_center = t_start + (t_stop - t_start) / 2.;
 

  DvObject_var Valf = new DvObject(0.0, DIM3, nestsize);  
  ttJoined->isRegular(deltat);

  DvObject_var B_singleave = new DvObject(0.0, DIM3);
  DvObject_var V_singleave = new DvObject(0.0, DIM3);
  DvObject_var Va_singleave = new DvObject(0.0, DIM3);
  
  // ----- SEH 18 Sep 2007 - to plot intrinsic E-field -----
  DvObject_var Erperp =  new DvObject(0.0, DIM3, nestsize); 
  Erperp->change_xref(DEPEND_0, ttJoined);
  Erperp->setFrameAttr("vector>gse_xyz");
  Erperp->change_xref("UNITS", "mV/m" );
  Erperp->change_xref("SI_conversion","1.0e-3 > V m^-1" );
  
  DvObject_var Erpar  =  new DvObject(0.0, DIM3, nestsize); 
  Erpar->setFrameAttr("vector>gse_xyz");
  Erpar->change_xref("UNITS", "mV/m" );
  Erpar->change_xref("SI_conversion","1.0e-3 > V m^-1" );
  Erpar->change_xref(DEPEND_0, ttJoined);
  // ------ eo SEH 18 Sep 2007 --------------------


  // --- send some info (time, nestsize etc to output ----------------

  sprintf(buf,"# t = %s <-- %s --> %s (%d samples)\n",t_start.getISOstring().c_str(),t_center.getISOstring().c_str(),t_stop.getISOstring().c_str(),nestsize);
  printf("%s",buf);
  QplugAppendTextDisplay(buf);
  
  sprintf(buf,"# Nestsize N = %d\n",nestsize);
  printf("%s",buf);
  QplugAppendTextDisplay(buf);

  sprintf(buf,"# Given Walenslope = %5.3f\n",alpha);
  printf("%s",buf);
  QplugAppendTextDisplay(buf);

  
 //  try{  
//     printf("\n");
//     printf("  k                 B                              V                      N           Pxx      Pyy      Pzz      Pxy      Pxz      Pyz\n");
//     printf("                  [nT]                          [km/s]               [1/cm^3]                            [keV/cm^3]  \n");
//     printf("--------------------------------------------------------------------------------------------------------------------------------------\n");
//     for (k=0; k<nestsize; k++) {
//       printf("%3d  %8.3f %8.3f %8.3f       ",k+1,(float)(*B)[k][0], (float)(*B)[k][1], (float)(*B)[k][2]);
//       printf("%8.3f %8.3f %8.3f      ",(float)(*V)[k][0], (float)(*V)[k][1], (float)(*V)[k][2]);
//       printf("%7.4f      ",(float)(*N)[k]);
//       for (i=0; i<3; i++) printf("%7.2f  ",(double)(*P)[k][i][i]);
//       printf("%7.2f  %7.2f  %7.2f\n",(float)(*P)[k][0][1], (float)(*P)[k][0][2], (float)(*P)[k][1][2]);
//     }
//   }
//  catch(Exception& e) {
//    printf("exception in SingleGeneric printout..\n");
//  }
  

  // ------ out to spreadsheet  -----------------------

const char *filename="/tmp/outGenericSpread.csv";
  
  if (GenerateSpreadSheet) {
    if (!fopen(filename,"r")) {
      f1 = fopen(filename,"w");
      fprintf(f1,"ID, tstart, tstop, nestsize, ");
      if (doType[0]) fprintf(f1,"b_ave_x,  b_ave_y,  b_ave_z,  n_mva_x,  n_mva_y,  n_mva_z,  Bn,  dBn,  [B], "); 
      if (doType[2]) fprintf(f1,"v_cis_x,  v_cis_y,  v_cis_z,  v_cis_n,  N,        [N],  ");
      if (doType[2]) fprintf(f1,"v_ht_x,   v_ht_y,   v_ht_z,   v_ht_n,   vht'n,    VAn,      wal_sl,    wal_cc,   ");
      if (doType[3]) fprintf(f1,"n_mfr_x,  n_mfr_y,  n_mfr_z,  u_mfr_x,  u_mfr_y,  u_mfr_z,  u_mfr_n,   du_mfr_n,   vmfr'n,   dvmfr'n,  VAn,  Bn,  dBn,");
      if (doType[4]) fprintf(f1,"n_mmr_x,  n_mmr_y,  n_mmr_z,  u_mmr_x,  u_mmr_y,  u_mmr_z,  u_mmr_n,   du_mmr_n,   vmmr'n,   dvmmr'n,  VAn,  Bn,  dBn,");
      if (doType[5]) fprintf(f1,"n_mlmr_x, n_mlmr_y, n_nlmr_z, u_mlmr_x, u_mlmr_y, u_mlmr_z, u_mlmr_n, ");
      if (doType[7]) fprintf(f1,"n_mter_x, n_mter_y, n_mter_z, u_mter_x, u_mter_y, u_mter_z, u_mter_n, ");
      if (doType[6]) fprintf(f1,"n_mer_x,  n_mer_y,  n_mer_z,  u_mer_x,  u_mer_y,  u_mer_z,  u_mer_n, ");
      if (doType[8]) fprintf(f1,"n_mvav_x, n_mvav_y, n_mvav_z, u_mvav_x, u_mvav_y, u_mvav_z, u_mvav_n,  du_mvav_n,  vmvav'n,   dvmvav'n,  VAn,  Bn,  dBn,");
                     fprintf(f1,"n_com_x,  n_com_y,  n_com_z,  u_com_x,  u_com_y,  u_com_z,  u_com_n,   du_com_n,   vcom'n,    dvcom'n,   VAn,  Bn,  dBn,");
		     fprintf(f1,"VAx,      VAy,      VAz\n");
      fclose(f1);
    }
    f1 = fopen(filename,"a+");
    fprintf(f1,"IDNO, %s , %s , %d , ",t_start.getISOstring().c_str(),t_stop.getISOstring().c_str(),nestsize);
    fclose(f1);
  }
  
  
  // ---- eo SEH 15 Nov 2008 ---
  
  DvObject_var Q_COM = new DvObject(0.0, DIM33); 
  DvObject_var U_COM = new DvObject(0.0, DIM3); 
  DvObject_var dYij  = new DvObject(0.0, DIM33);       // matrix to hold the three angular errors from eq 8.23 for MVA

  DvObject_var Normal = new DvObject(0.0, DIM3);       
  DvObject_var V_res = new DvObject(0.0, DIM3);       
  DvObject_var E_ave = new DvObject(0.0, DIM3);  
  DvObject_var E_res = new DvObject(0.0, DIM3);  
  DvObject_var E_tang= new DvObject(0.0, DIM3);  


  double mu_0 = 1.25664e-06;     // modified mu_0 (due to B in nT and dens in 1/ccm)
  double rho,m_p= 1.67e-27;      // rho = mass dens with proton_mass m_p
  double dBx3, sigma_Un, sigma_Vcis;
  double wl, cc, weight_k, weight_sum;
  DvObject_var Q_MVAB;


  // ------ start calculating the Q-matrices - establish EvecMVAB for constraining --------------
  
    
  if (doType[0]) {
    printf("trying MVAB to establish constraints\n");
    B->change_xref(DEPEND_0, ttJoined);
    Q_MVAB = Q_MVAB_calc(B,V,U,Valf,N, &wl, &cc, nestsize, 0);
    
    GenericMinVar minvar(0, Q_MVAB, ProjMx);  
    EvecMVAB = minvar.eigenvectors();
    Evalues = minvar.eigenvalues(); 

  
    Report(Evalues, EvecMVAB, true, "MVAB", 0, alpha, SCc_ID.c_str());                 // unconstrained
    if (constrained > 0) sprintf(constrbuf,"C%1d",constrained);
  }  


  // ------------- 0 - MVAB -------------------------------

  bool doV = V.is_ok();
  bool doN = N.is_ok();
  bool doB = B.is_ok();
  
  if (doType[0]) {
    printf("trying MVAB\n");
    sprintf(typebuf,"%s (%d)",SCc_ID.c_str(),nestsize);
    for (k=0; k<nestsize; k++) {
        for (i=0; i<3; i++) {      
            B_singleave->dbl(i) += B->dbl(k,i);
            if (doV) V_singleave->dbl(i) += V->dbl(k,i);
        }
        if (doN) N_ave += N->dbl(k);
    }
    *B_singleave /= (double) nestsize;
   
   
    //  sprintf(typebuf,"MVAB%s",constrbuf);
    sprintf(typebuf,"%s (%d)",SCc_ID.c_str(),nestsize);
    Q_MVAB = Q_MVAB_calc(B, V, U, Valf, N, &wl, &cc, nestsize,doType[2]);
    
    ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained, N_ave);

    GenericMinVar minvar(constrained, Q_MVAB, ProjMx);  
    Evec = minvar.eigenvectors();
    Evalues = minvar.eigenvalues(); 
   
   
    if (doV) {
        *V_singleave /= (double) nestsize;
        if (doN) {
            N_ave /= (double) nestsize;
            Va_singleave = new DvObject(B_singleave);
            *Va_singleave *= alpha * 1.0e-12 / sqrt(mu_0 * N_ave *1.e6 * m_p);
        }
    }

    weight_k = ExWeights->dbl(0) * weight_k_calc(Evalues);
    weight_sum = weight_k;
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          Q_COM->dbl(0,i,j) += weight_k * Q_MVAB->dbl(0,i,j) / ( Q_MVAB->dbl(0,0,0) + Q_MVAB->dbl(0,1,1) + Q_MVAB->dbl(0,2,2));
      }
      U_COM->dbl(i) += weight_k * U->dbl(i);
    }

    
    dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVAB, 0.0, nestsize, 0); 
    if (doV) sigma_Un = sigma(V,Evec,nestsize);  // calculate standarddeviation in <V> dot n_COM
    //  double sigma_Vcis = sigma(V, Evec, nestsize);  // calculate standarddeviation in <V> dot n
    if (doV) sigma_Vcis = err_824(V_singleave, Evec, Evalues, dYij, Q_MVAB,  0.0, nestsize, 1);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
    
    un =  bn = vn = van = 0.0;

    for (i=0; i<3; i++) {
      un += U->dbl(i) * Evec->dbl(0,0,i) / 1.e3;
      bn += B_singleave->dbl(i) * Evec->dbl(0,0,i);
      if (doV) vn += V_singleave->dbl(i) * Evec->dbl(0,0,i);
      if (doV && doN) van += Va_singleave->dbl(i) * Evec->dbl(0,0,i);
    }
    
    Report(Evalues, Evec, true, "MVAB",constrained, alpha, typebuf);
    
//     printf("Details about covariance matrix :\n");
//     Q_MVAB->display();
//     printf("weight_MVAB = %10.4f\n\n",weight_k);
//     printf("Accumulated Q_COM : (after MVAB)\n- - - - - - - - - - - - - - - - - - - - - - \n"); 
//     Q_COM->display(); 
    
    sprintf(buf,"! Given weight, calculated weight =  %5.3f   %5.3f\n",ExWeights->dbl(0),weight_k);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    
    sprintf(buf,"! V_HT   = [%10.2f  %10.2f  %10.2f] (%10.2f)\n",U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3, un);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    
    sprintf(buf,"! <Vcis> = [%10.2f  %10.2f  %10.2f] (V'n =%10.2f +/- %8.4f)\n",
	    V_singleave->dbl(0),V_singleave->dbl(1),V_singleave->dbl(2), vn-un, 0.00);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    
    sprintf(buf,"! <w_Va> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2), van, sigma_Un);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    
    
    sprintf(buf,"! <Bvec> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2), bn, dBx3);
    printf("%s\n",buf);
    QplugAppendTextDisplay(buf);

    printf("Calc jumps B...\n");
    
    double bjump = jumpvalue(B);
    printf("Calc jumps N...\n");
    double njump = jumpvalue(N);
    printf("Spreadsheet.....\n");
    if (GenerateSpreadSheet) {
      FILE *f1 = fopen(filename,"a+");
      fprintf(f1,"%g , %g , %g , ",B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2)); 
      fprintf(f1,"%g , %g , %g , ",Evec->dbl(0,0,0), Evec->dbl(0,0,1), Evec->dbl(0,0,2));
      fprintf(f1,"%g , %g , %g , ", bn, dBx3, bjump);
      fprintf(f1,"%g , %g , %g , ",V_singleave->dbl(0),V_singleave->dbl(1),V_singleave->dbl(2));
      fprintf(f1,"%g , %g , %g , ",vn, N_ave, njump);
      fprintf(f1,"%g , %g , %g , ",U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3);
      fprintf(f1,"%g , %g , %g , ",un, vn-un, van);
      fprintf(f1,"%g , %g , ",wl, cc);
      fclose(f1);
    }
  }


 // --- addition by SEH, 15 Nov 2008 - export B for calc of stddev ---

//   char *exportfile = "/tmp/bfield_IDNO.csv";
//   FILE *f9 =  fopen(exportfile,"w");
//   fprintf(f9,"KnetterID, t_start, t_center, t_stop\n");
//   fprintf(f9,"IDNO, %s, %s, %s\n",t_start.iso_srep().c_str()+11,t_center.iso_srep().c_str()+11, t_stop.iso_srep().c_str()+11);
//   fprintf(f9,"B_ave = ,%10.4f, %10.4f, %10.4f\n\n",(double)(*B_singleave)[0],(double)(*B_singleave)[1],(double)(*B_singleave)[2]);   
//   for (int i=0; i<nestsize; i++) {
//     fprintf(f9,"B_%2d,  %10.4f, %10.4f, %10.4f\n",i, (double)(*B)[i][0], (double)(*B)[i][1] ,(double)(*B)[i][2]);
//   }
//   switch(nestsize) {
//   case 28: 
//     fprintf(f9,",,,,\n,,,,\n,,,,\n,,,,\n");
//     break;
//   case 29: 
//     fprintf(f9,",,,,\n,,,,\n,,,,\n");
//     break;
//   case 30:
//     fprintf(f9,",,,,\n,,,,\n");
//     break;
//   case 31:
//     fprintf(f9,",,,,\n");
//     break;
//   default:
//     break;    
//   }
//   fclose(f9);
 

  // ------------- 1 - MVAdB/dt -------------------------------

  printf("Do MVAdBdt..\n");
  if (doType[1]) {
    printf("trying MVAdB/dt\n");
    sprintf(typebuf,"%s (%d)",SCc_ID.c_str(),nestsize);
    DvObject_var Udummy = new DvObject(0.0, DIM3);        
    DvObject_var Valfdummy = new DvObject(0.0, DIM3, nestsize);  
    DvObject_var Q_MVADB = Q_MVAB_calc(dB, V, Udummy, Valfdummy, N, &wl, &cc, nestsize,doType[2]);
    ProjMx = Projection(dB, V, Udummy, Valf, EvecMVAB, alpha, nestsize, constrained, N_ave);
      
    GenericMinVar minvar(constrained, Q_MVADB, ProjMx);  
    Evec = minvar.eigenvectors();
    Evalues = minvar.eigenvalues(); 
   
    
    for (i=0; i<3; i++) {
      Va_singleave->dbl(i) = alpha * 1.0e-3 * B_singleave->dbl(i) * 1e-9 / sqrt(mu_0 * N_ave *1e06 * m_p);
    }
    
    weight_k = ExWeights->dbl(1) * weight_k_calc(Evalues);
    weight_sum = weight_k;
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          Q_COM->dbl(0,i,j) += weight_k * Q_MVADB->dbl(0,i,j) / ( Q_MVADB->dbl(0,0,0) + Q_MVADB->dbl(0,1,1) + Q_MVADB->dbl(0,2,2));
      }
      U_COM->dbl(i) += weight_k * U->dbl(i);
    }
    
    dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVADB, 0.0, nestsize, 0); 
    if (doV) sigma_Un = sigma(V,Evec,nestsize);  // calculate standarddeviation in <V> dot n_COM
    if (doV) sigma_Vcis = err_824(V_singleave, Evec, Evalues, dYij, Q_MVADB,  0.0, nestsize, 1);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
    
    
    un =  bn = vn = van = 0.0;

    for (i=0; i<3; i++) {
      un += U->dbl(i) * Evec->dbl(0,0,i) / 1e3;
      bn += B_singleave->dbl(i) * Evec->dbl(0,0,i);
      if (doV) vn += V_singleave->dbl(i) * Evec->dbl(0,0,i);
      if (doV && doN) van += Va_singleave->dbl(i) *Evec->dbl(0,0,i);
    }
    
    
    Report(Evalues, Evec, true, "MVAdBdT",constrained, alpha, typebuf);
    
    printf("Details about covariance matrix :\n");
      for(size_t i=0; i<3; i++){
          for(size_t j=0; i<3; i++){
              cout << Q_MVADB->dbl(0,i,j);
              if (j != 2) cout << ", ";
          }
          cout << endl;
      }
    printf("weight_MVAdB/dt = %10.4f\n\n",weight_k);
    printf("Accumulated Q_COM : (after MVAdB/dT)\n- - - - - - - - - - - - - - - - - - - - - - - \n"); 
      for(size_t i=0; i<3; i++){
          for(size_t j=0; i<3; i++){
              cout << Q_COM->dbl(0,i,j);
              if (j != 2) cout << ", ";
          }
          cout << endl;
      }
    
    
    sprintf(buf,"! Given weight, calculated weight =  %5.3f   %5.3f\n",ExWeights->dbl(0),weight_k);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    
    sprintf(buf,"! V_HT   = [%10.2f  %10.2f  %10.2f] (%10.2f)\n",U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3, un);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    
    sprintf(buf,"! <Vcis> = [%10.2f  %10.2f  %10.2f] (V'n =%10.2f +/- %8.4f)\n",
	    V_singleave->dbl(0),V_singleave->dbl(1),V_singleave->dbl(2), vn-un, 0.00);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    
    sprintf(buf,"! <w_Va> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2), van, sigma_Un);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    
    
    sprintf(buf,"! <Bvec> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2), bn, dBx3);
    printf("%s\n",buf);
    QplugAppendTextDisplay(buf);
    
    
    if (GenerateSpreadSheet) {
      FILE *f1 = fopen(filename,"a");
      fprintf(f1,"%g , %g , %g , ",Evec->dbl(0,0,0), Evec->dbl(0,0,1), Evec->dbl(0,0,2));
      fprintf(f1,"%g , %g , %g , ",U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3);
      fprintf(f1,"%g , %g , %g , ",un, wl, cc);
      fclose(f1);
    }
  }




  // ------------- 3 - MFR -------------------------------

  if (doType[3]) {
    printf("trying MFR\n");
    sprintf(typebuf,"%s (%d)",SCc_ID.c_str(),nestsize);
    DvObject_var Q_MFR = Q_MFR_calc(B,V,U,nestsize);
    ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize,constrained, N_ave);
    
    GenericMinVar minvar(constrained, Q_MFR, ProjMx);  
    Evec = minvar.eigenvectors();
    Evalues = minvar.eigenvalues(); 
      
    weight_k = ExWeights->dbl(2)  * weight_k_calc(Evalues);
    weight_sum += weight_k;
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          Q_COM->dbl(0,i,j) += weight_k * Q_MFR->dbl(0,i,j) / ( Q_MFR->dbl(0,0,0) + Q_MFR->dbl(0,1,1) + Q_MFR->dbl(0,2,2));
      }
      U_COM->dbl(i) += weight_k * U->dbl(i);
      Normal->dbl(i) = Evec->dbl(0,0,i); 
      V_res->dbl(i) = U->dbl(i)/1e3 - V_singleave->dbl(i);
    }


    dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVAB, 0.0, nestsize, 1); 
    sigma_Un = err_824(U, Evec, Evalues, dYij, Q_MFR,  0.0, nestsize, 3);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
    sigma_Vcis = err_824(V_singleave, Evec, Evalues, dYij, Q_MFR,  0.0, nestsize, 1);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
    //sigma_Vcis = sigma(V, Evec, nestsize);  // calculate standarddeviation in <V> dot n

    un = 0.0;   for (i=0; i<3; i++) un += U->dbl(i) * Evec->dbl(0,0,i) / 1e3;
    bn = 0.0;   for (i=0; i<3; i++) bn += B_singleave->dbl(i) * Evec->dbl(0,0,i);
    vn = 0.0;   for (i=0; i<3; i++) vn += V_singleave->dbl(i) * Evec->dbl(0,0,i);
    van = 0.0;  for (i=0; i<3; i++) van += Va_singleave->dbl(i) * Evec->dbl(0,0,i);

    Report(Evalues, Evec, true, "MFR",constrained, alpha, typebuf);

    sprintf(buf,"! Given weight, calculated weight =  %5.3f   %5.3f\n",ExWeights->dbl(1),weight_k);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! V_MFR  = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    U->dbl(0)/1e3,U->dbl(1)/1e3, U->dbl(2)/1e3, un, sigma_Un/1e3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Vcis> = [%10.2f  %10.2f  %10.2f] (V'n =%10.2f +/- %8.4f)\n",
	    V_singleave->dbl(0),V_singleave->dbl(1),V_singleave->dbl(2), vn-un, sigma_Vcis);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <w_Va> = [%10.2f  %10.2f  %10.2f] (%10.2f)\n",Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2), van);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Bvec> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	   B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2), bn, dBx3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
 
    // --------- SEH 12.09.2007 ----------------

    DvObject_var E_vecn_ave = new DvObject(0.0, DIM3);
    DvObject_var E_tan2_ave =  new DvObject(0.0, DIM3);

    DvObject_var Vtemp = V->changeUnitsTo("1>m s^-1", "m/s");
    DvObject_var Btemp = B->changeUnitsTo("1>T", "T");
    DvString frame = V->getFrameAttr();
    U->change_xref(FRAME, frame);
    Normal->change_xref(FRAME, frame);
        
    DvObject_var  E_vecn = U - Vtemp;
    E_vecn = E_vecn->vec(Btemp);
      
    DvObject_var E_tan2 = E_vecn->dot(Normal);
    E_tan2 = E_tan2 * Normal;


    for (i=0; i<nestsize; i++) {
          
        for (int j=0; j<3; j++) {
            Erperp->dbl(i,j) = 1e3 * (E_vecn->dbl(i,j) - E_tan2->dbl(i,j));
            Erpar->dbl(i,j)  = 1e3 * E_vecn->dbl(i,j);
            E_tan2_ave->dbl(j) += Erperp->dbl(i,j) / (double)nestsize;
            E_vecn_ave->dbl(j) += Erpar->dbl(i,j) / (double)nestsize;
        }
       // printf("E_vecn[%2d] =  [%10.6f  %10.6f  %10.6f] ",i, 1e3*E_vecn->dbl(0), 1e3*E_vecn->dbl(1), 1e3*E_vecn->dbl(2));
       // printf("E_tan =   [%10.6f  %10.6f  %10.6f] mV/m\n", Erperp->dbl(i,0), Erperp->dbl(i,1), Erperp->dbl(i,2));
    }
      
    printf("<E_tan>    =  [%10.6f  %10.6f  %10.6f] mV/m\n", E_tan2_ave->dbl(0), E_tan2_ave->dbl(1), E_tan2_ave->dbl(2));
    printf("<E_vecn>   =  [%10.6f  %10.6f  %10.6f] mV/m\n", E_vecn_ave->dbl(0), E_vecn_ave->dbl(1), E_vecn_ave->dbl(2));
      
    DvObject_var Q_ETAN = Q_GEN_calc(Erperp);
    GenericMinVar minvar1(0, Q_ETAN, ProjMx);
    Evec = minvar1.eigenvectors();
    Evalues = minvar1.eigenvalues();
     
    for(size_t i=0; i<3; i++){
        for(size_t j=0; i<3; i++){
            cout << Q_ETAN->dbl(0,i,j);
            if (j != 2) cout << ", ";
        }
        cout << endl;
    }
        
    int order[3]={0,1,2};
    for (int i=0; i<3; i++){
        sprintf(buf,"%12.6g  [%10.6g, %10.6g, %10.6g]\n", Evalues->dbl(order[i]),
        Evec->dbl(0,order[i],0), Evec->dbl(0,order[i],1), Evec->dbl(0,order[i],2));
        QplugAppendTextDisplay(buf);
        printf("%s",buf);
    }
    
    // --------- e.o. SEH 12.09.2007 ----------------
    
    if (GenerateSpreadSheet) {
      FILE *f1 = fopen(filename,"a");
      fprintf(f1,"%g , %g , %g , ",Evec->dbl(0,0,0), Evec->dbl(0,0,1), Evec->dbl(0,0,2));
      fprintf(f1,"%g , %g , %g , ",U->dbl(0)/1e3, U->dbl(1)/1e3, U->dbl(2)/1e3);
      fprintf(f1,"%g , %g , %g , %g ,",un, sigma_Un/1e3,  vn-un, sigma_Vcis);
      fprintf(f1,"%g , %g , %g ,", van, bn, dBx3);
      fclose(f1);
    }
  }


  // ------------- 4 - MMR -------------------------------

  if (doType[4]) {
    printf("trying MMR\n");
    DvObject_var Q_MMR = Q_MMR_calc(V,N,U,nestsize);
    ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained, N_ave);
   
    GenericMinVar minvar(constrained, Q_MMR, ProjMx);  
    Evec = minvar.eigenvectors();
    Evalues = minvar.eigenvalues(); 
    
  
    weight_k = ExWeights->dbl(3)  * weight_k_calc(Evalues);
    weight_sum += weight_k;
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
        Q_COM->dbl(0,i,j) += weight_k * Q_MMR->dbl(0,i,j) / ( Q_MMR->dbl(0,0,0) + Q_MMR->dbl(0,1,1) + Q_MMR->dbl(0,2,2));
      }
      U_COM->dbl(i) += weight_k * U->dbl(i);
    }
  
    N_ave /= (double) nestsize;
    for (i=0; i<3; i++) {
      if (doB) Va_singleave->dbl(i) = alpha * 1.0e-3 * B_singleave->dbl(i) * 1e-9 / sqrt(mu_0 * N_ave *1e06 * m_p);
    }

    if (doB) dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVAB, 0.0, nestsize, 1); 
    sigma_Un = err_824(U, Evec, Evalues, dYij, Q_MMR,  0.0, nestsize, 3);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
    if (doB) sigma_Vcis = err_824(V_singleave, Evec, Evalues, dYij, Q_MMR,  0.0, nestsize, 1);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]

    un =  bn = vn = van = 0.0;

    for (i=0; i<3; i++) {
      un += U->dbl(i) * Evec->dbl(0,0,i) / 1e3;
      if (doB) bn += B_singleave->dbl(i) * Evec->dbl(0,0,i);
      vn += V_singleave->dbl(i) * Evec->dbl(0,0,i);
      if (doB) van += Va_singleave->dbl(i) * Evec->dbl(0,0,i);
      Normal->dbl(i) = Evec->dbl(0,0,i); 
      V_res->dbl(i) = V_singleave->dbl(i) - U->dbl(i)/1e3;
    }

    Report(Evalues, Evec, true, "MMR",constrained, alpha, typebuf);

    sprintf(buf,"! Given weight, calculated weight =  %5.3f   %5.3f\n",ExWeights->dbl(2),weight_k);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! V_MMR  = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    U->dbl(0)/1e3, U->dbl(1)/1e3, U->dbl(2)/1e3, un, sigma_Un/1e3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Vcis> = [%10.2f  %10.2f  %10.2f] (V'n =%10.2f +/- %8.4f)\n",
	    V_singleave->dbl(0), V_singleave->dbl(1), V_singleave->dbl(2), vn-un, sigma_Vcis);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <w_Va> = [%10.2f  %10.2f  %10.2f] (%10.2f)\n",Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2),van);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Bvec> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2), bn, dBx3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    // --------- SEH 12.09.2007 ----------------
    
    if (1 == 0) {
      DvObject_var E_tan2 = new DvObject(0.0, DIM3);   
      DvObject_var E_vecn = new DvObject(0.0, DIM3);   
      DvObject_var E_vecn_ave = new DvObject(0.0, DIM3);
      DvObject_var E_tan2_ave =  new DvObject(0.0, DIM3);
      DvObject_var Vtemp = new DvObject(0.0, DIM3);        
      DvObject_var Btemp = new DvObject(0.0, DIM3); 
      
      for (i=0; i<nestsize; i++) {
        for (int j=0; j<3; j++) {
	      Vtemp->dbl(j) = V->dbl(i,j)*1e3;   // to get mV/m
	      Btemp->dbl(j) = B->dbl(i,j)*1e-9;
	    }
	
	    E_vecn = U-Vtemp;
        E_vecn = E_vecn->vec(Btemp);
	    E_tan2 = Normal * Normal->dot(E_vecn);
	
	    for (int j=0; j<3; j++) {
            Erperp->dbl(i,j) = 1e3 * (E_vecn->dbl(j) - E_tan2->dbl(j));
            Erpar->dbl(i,j)  = 1e3 * E_vecn->dbl(j);
            E_tan2_ave->dbl(j) += Erperp->dbl(i,j) / (double)nestsize;
	        E_vecn_ave->dbl(j) += Erpar->dbl(i,j) / (double)nestsize;
	    }

      }
      
      printf("<E_tan>    =  [%10.6f  %10.6f  %10.6f] mV/m\n", E_tan2_ave->dbl(0), E_tan2_ave->dbl(1), E_tan2_ave->dbl(2));
      printf("<E_vecn>   =  [%10.6f  %10.6f  %10.6f] mV/m\n", E_vecn_ave->dbl(0), E_vecn_ave->dbl(1), E_vecn_ave->dbl(2));
      
      DvObject_var Q_ETAN = Q_GEN_calc(Erperp);

      GenericMinVar minvar(0, Q_ETAN, ProjMx);  
      Evec = minvar.eigenvectors();
      Evalues = minvar.eigenvalues(); 
      
      
    // display();
     for(size_t i=0; i<3; i++){
            for(size_t j=0; i<3; i++){
                cout << Q_ETAN->dbl(0,i,j);
                if (j != 2) cout << ", ";
            }
            cout << endl;
      }
  
      int order[3]={0,1,2};
      for (int i=0; i<3; i++){
	    sprintf(buf,"%12.6g  [%10.6g, %10.6g, %10.6g]\n", 
		Evalues->dbl(order[i]),
		Evec->dbl(0,order[i],0),
		Evec->dbl(0,order[i],1), 
		Evec->dbl(0,order[i],2) );
	    QplugAppendTextDisplay(buf);
	    printf("%s",buf);
      }
    }  

  // --------- e.o. SEH 12.09.2007 ----------------

    
    if (GenerateSpreadSheet) {
      FILE *f1 = fopen(filename,"a");
      fprintf(f1,"%g , %g , %g , ",Evec->dbl(0,0,0), Evec->dbl(0,0,1), Evec->dbl(0,0,2));
      fprintf(f1,"%g , %g , %g , ",U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3);
      fprintf(f1,"%g , %g , %g , %g ,",un, sigma_Un/1e3,  vn-un, sigma_Vcis);
      fprintf(f1,"%g , %g , %g , ", van, bn, dBx3);
      fclose(f1);
    }
  }
  



  // ------------- 5 -  MLMR -------------------------------
  
  if (doType[5]) {
    printf("trying MLMR\n");
    DvObject_var Q_MLMR = Q_MLMR_calc(V,B,P,N,U,nestsize);
    ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained, N_ave);

    GenericMinVar minvar(constrained, Q_MLMR, ProjMx);
    Evec = minvar.eigenvectors();
    Evalues = minvar.eigenvalues(); 
   
    
    weight_k = ExWeights->dbl(4) * weight_k_calc(Evalues);
    weight_sum += weight_k;
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
	   Q_COM->dbl(0,i,j) += weight_k * Q_MLMR->dbl(0,i,j) / ( Q_MLMR->dbl(0,0,0) + Q_MLMR->dbl(0,1,1) + Q_MLMR->dbl(0,2,2));
      }
      U_COM->dbl(i) += weight_k * U->dbl(i);
      Normal->dbl(i) = Evec->dbl(0,0,i); 
      V_res->dbl(i) = V_singleave->dbl(i) - U->dbl(i)/1e3;
    }
  
    dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVAB, 0.0, nestsize, 1); 
    sigma_Un = err_824(U, Evec, Evalues, dYij, Q_MLMR,  0.0, nestsize, 3);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
    //  sigma_Vcis = sigma(V, Evec, nestsize);  // calculate standarddeviation in <V> dot n
    sigma_Vcis = err_824(V_singleave, Evec, Evalues, dYij, Q_MLMR,  0.0, nestsize, 1);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]

  
    un = 0.0;   for (i=0; i<3; i++) un += U->dbl(i) * Evec->dbl(0,0,i) / 1e3;
    bn = 0.0;   for (i=0; i<3; i++) bn += B_singleave->dbl(i) * Evec->dbl(0,0,i);
    vn = 0.0;   for (i=0; i<3; i++) vn += V_singleave->dbl(i) * Evec->dbl(0,0,i);
    van = 0.0;  for (i=0; i<3; i++) van += Va_singleave->dbl(i) * Evec->dbl(0,0,i);

    Report(Evalues, Evec, true, "MLMR",constrained, alpha, typebuf);
     sprintf(buf,"! Given weight, calculated weight =  %5.3f   %5.3f\n",ExWeights->dbl(3),weight_k);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! V_MLMR = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3, un, sigma_Un/1e3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Vcis> = [%10.2f  %10.2f  %10.2f] (V'n =%10.2f +/- %8.4f)\n",
	    V_singleave->dbl(0),V_singleave->dbl(1),V_singleave->dbl(2), vn-un, sigma_Vcis);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <w_Va> = [%10.2f  %10.2f  %10.2f] (%10.2f)\n",Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2), van);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Bvec> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2), bn, dBx3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf); 

    // --------- SEH 12.09.2007 ----------------
    E_res = B_singleave->vec(V_res);
    *E_res *= 1e-3;
    E_tang = E_res;
    E_tang -= Normal * E_res->dot(Normal);
    *E_tang *= 1e-3;
    sprintf(buf,"! <E_res>= [%10.3f  %10.3f  %10.3f] \n",
	    E_res->dbl(0),E_res->dbl(1),E_res->dbl(2));
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
 
    sprintf(buf,"! <E_tan>= [%10.3f  %10.3f  %10.3f] \n",
	    E_tang->dbl(0),E_tang->dbl(1),E_tang->dbl(2));
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
 
   // --------- e.o. SEH 12.09.2007 ----------------
    if (GenerateSpreadSheet) ToSpreadsheet(filename, Evec, U, 0, false);
  }
  

 // ------------- 6 - MER -------------------------------
  
  if (doType[6]) {
    printf("trying MER\n");
    DvObject_var Q_MER = Q_MER_calc(V,B,P,N,U,nestsize);
    ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize,constrained, N_ave);

    GenericMinVar minvar(constrained, Q_MER, ProjMx);  
    Evec = minvar.eigenvectors();
    Evalues = minvar.eigenvalues(); 
    
    weight_k = ExWeights->dbl(6) * weight_k_calc(Evalues);
    weight_sum += weight_k;
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
        Q_COM->dbl(0,i,j) += weight_k * Q_MER->dbl(0,i,j) / ( Q_MER->dbl(0,0,0) + Q_MER->dbl(0,1,1) + Q_MER->dbl(0,2,2));
      }
      U_COM->dbl(i) += weight_k * U->dbl(i);
    } 

    dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVAB, 0.0, nestsize, 1); 
    sigma_Un = err_824(U, Evec, Evalues, dYij, Q_MER,  0.0, nestsize, 3);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
    //  sigma_Vcis = sigma(V, Evec, nestsize);  // calculate standarddeviation in <V> dot n
    sigma_Vcis = err_824(V_singleave, Evec, Evalues, dYij, Q_MER,  0.0, nestsize, 1);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]

    un = 0.0;   for (i=0; i<3; i++) un += U->dbl(i) * Evec->dbl(0,0,i) / 1e3;
    bn = 0.0;   for (i=0; i<3; i++) bn += B_singleave->dbl(i) * Evec->dbl(0,0,i);
    vn = 0.0;   for (i=0; i<3; i++) vn += V_singleave->dbl(i) * Evec->dbl(0,0,i);
    van = 0.0;  for (i=0; i<3; i++) van += Va_singleave->dbl(i) * Evec->dbl(0,0,i);

    Report(Evalues, Evec, true, "MER",constrained, alpha, typebuf);
//     printf("Details about covariance matrix :\n");
//     Q_MER->display();
//     printf("weight_MER = %10.4f\n",weight_k);
//     printf("Accumulated Q_COM : (after MER)\n- - - - - - - - - - - - - - - - - - - - - - \n"); Q_COM->display(); 

    sprintf(buf,"! Given weight, calculated weight =  %5.3f   %5.3f\n",ExWeights->dbl(5),weight_k);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! V_MER  = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	   U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3, un, sigma_Un/1e3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Vcis> = [%10.2f  %10.2f  %10.2f] (V'n =%10.2f +/- %8.4f)\n",
	    V_singleave->dbl(0),V_singleave->dbl(1),V_singleave->dbl(2), vn-un, sigma_Vcis);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <w_Va> = [%10.2f  %10.2f  %10.2f] (%10.2f)\n",Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2), van);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Bvec> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2), bn, dBx3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    if (GenerateSpreadSheet) ToSpreadsheet(filename, Evec, U, 0, false);
  }
  

  // ------------- 7 - MTER -------------------------------
 
  if (doType[7]) {  
    printf("trying MTER\n");
    DvObject_var Q_MTER = Q_MTER_calc(V,B,P,E,N,H,U,nestsize);
    ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize,constrained, N_ave);
    GenericMinVar minvar(constrained, Q_MTER, ProjMx);  
    Evec = minvar.eigenvectors();
    Evalues = minvar.eigenvalues(); 
    
  
    weight_k = ExWeights->dbl(5)  * weight_k_calc(Evalues);
    weight_sum += weight_k;
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
	    Q_COM->dbl(0,i,j) += weight_k * Q_MTER->dbl(0,i,j) / ( Q_MTER->dbl(0,0,0) + Q_MTER->dbl(0,1,1) + Q_MTER->dbl(0,2,2));
      }
      U_COM->dbl(i) += weight_k * U->dbl(i);
    }
  
    dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVAB, 0.0, nestsize, 1); 
    sigma_Un = err_824(U, Evec, Evalues, dYij, Q_MTER,  0.0, nestsize, 3);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
    //  sigma_Vcis = sigma(V, Evec, nestsize);  // calculate standarddeviation in <V> dot n
    sigma_Vcis = err_824(V_singleave, Evec, Evalues, dYij, Q_MTER,  0.0, nestsize, 1);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]

    un = 0.0;   for (i=0; i<3; i++) un += U->dbl(i) * Evec->dbl(0,0,i) / 1e3;
    bn = 0.0;   for (i=0; i<3; i++) bn += B_singleave->dbl(i) * Evec->dbl(0,0,i);
    vn = 0.0;   for (i=0; i<3; i++) vn += V_singleave->dbl(i) * Evec->dbl(0,0,i);
    van = 0.0;  for (i=0; i<3; i++) van += Va_singleave->dbl(i) * Evec->dbl(0,0,i);

    Report(Evalues, Evec, true, "MTER",constrained, alpha, typebuf);
    //   printf("Details about covariance matrix :\n");
    //   Q_MTER->display();
    //   printf("weight_MTER = %10.4f\n",weight_k);
    //   printf("Accumulated Q_COM : (after MTER)\n- - - - - - - - - - - - - - - - - - - - - - \n"); Q_COM->display(); 

    sprintf(buf,"! Given weight, calculated weight =  %5.3f   %5.3f\n",ExWeights->dbl(4),weight_k);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! V_MTER = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3, un, sigma_Un/1e3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Vcis> = [%10.2f  %10.2f  %10.2f] (V'n =%10.2f +/- %8.4f)\n",
	    V_singleave->dbl(0),V_singleave->dbl(1),V_singleave->dbl(2), vn-un, sigma_Vcis);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <w_Va> = [%10.2f  %10.2f  %10.2f] (%10.2f)\n",
	    Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2), van);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Bvec> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2), bn, dBx3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    if (GenerateSpreadSheet) ToSpreadsheet(filename, Evec,U, 0, false);
  }
  
  
  

  // ------------- 8 - MVAV -------------------------------
  
  if (doType[8]) {
    printf("trying MVAV\n");
    sprintf(typebuf,"%s (%d)",SCc_ID.c_str(),nestsize);
    DvObject_var Q_MVAV = Q_MVAV_calc(V,U,nestsize);
    ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained, N_ave);
    GenericMinVar minvar(constrained, Q_MVAV, ProjMx);  
    Evec = minvar.eigenvectors();
    Evalues = minvar.eigenvalues(); 
    

    for (k=0; k<nestsize; k++) {
	  for (i=0; i<3; i++) {      
	    V_singleave->dbl(i) += V->dbl(k,i);
	  }
    }
    *V_singleave /= (float) nestsize;            
    

    weight_k = ExWeights->dbl(7) * weight_k_calc(Evalues);
    weight_sum += weight_k;
    printf("MVAV Weighs %f  %f\n",weight_sum, weight_k);
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
	    Q_COM->dbl(0,i,j) += weight_k * Q_MVAV->dbl(0,i,j) / ( Q_MVAV->dbl(0,0,0) +  Q_MVAV->dbl(0,1,1) + Q_MVAV->dbl(0,2,2));
      }
      U_COM->dbl(i) += weight_k * U->dbl(i);
    }

    printf("calc avg../n");
    un = bn = vn = van = 0.0;

    for (i=0; i<3; i++) {
      un += U->dbl(i) * Evec->dbl(0,0,i) / 1e3;
      if (doType[0]) bn += B_singleave->dbl(i) * Evec->dbl(0,0,i);
      vn += V_singleave->dbl(i) * Evec->dbl(0,0,i);
      van += Va_singleave->dbl(i) * Evec->dbl(0,0,i);
    }

    dBx3 = 0.0;
    if (doType[0]) dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVAB, 0.0, nestsize, 1); 
    sigma_Un = err_824(U, Evec, Evalues, dYij, Q_MVAV,  0.0, nestsize, 3);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
    //  sigma_Vcis = sigma(V, Evec, nestsize);  // calculate standarddeviation in <V> dot n
    sigma_Vcis = err_824(V_singleave, Evec, Evalues, dYij, Q_MVAV,  0.0, nestsize, 1);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]

    Report(Evalues, Evec, true, "MVAV",constrained, alpha, typebuf);
//     printf("Details about covariance matrix :\n");
//     Q_MVAV->display();
//     printf("weight_MVAV = %10.4f\n",weight_k);
//     printf("Accumulated Q_COM : (after MVAV)\n- - - - - - - - - - - - - - - - - - - - - - \n"); Q_COM->display(); 

    sprintf(buf,"! Given weight, calculated weight =  %5.3f   %5.3f\n",ExWeights->dbl(6),weight_k);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! V_MVAV = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	    U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3, un, sigma_Un/1e3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Vcis> = [%10.2f  %10.2f  %10.2f] (V'n =%10.2f +/- %8.4f)\n",
	    V_singleave->dbl(0),V_singleave->dbl(1),V_singleave->dbl(2), vn-un, sigma_Vcis);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <w_Va> = [%10.2f  %10.2f  %10.2f] (%10.2f)\n",Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2), van);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);

    sprintf(buf,"! <Bvec> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	   B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2), bn, dBx3);
    printf("%s",buf);
    QplugAppendTextDisplay(buf);
    if (GenerateSpreadSheet) {
      FILE *f1 = fopen(filename,"a");
      fprintf(f1,"%g , %g , %g , ",Evec->dbl(0,0,0), Evec->dbl(0,0,1), Evec->dbl(0,0,2));
      fprintf(f1,"%g , %g , %g , ",U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3);
      fprintf(f1,"%g , %g , %g , %g ,",un, sigma_Un/1e3,  vn-un, sigma_Vcis);
      fprintf(f1,"%g , %g , %g , ", van, bn, dBx3);
      fclose(f1);
    }
  }
  
  // ------------------ COM -------------------------------

  printf("COM calculations...\n");
 
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      Q_COM->dbl(0,i,j) /= weight_sum;
    }
    U_COM->dbl(i) /=  weight_sum;
  }

  ProjMx = Projection(B, V, U_COM, Valf, EvecMVAB, alpha, nestsize,constrained, N_ave);
  GenericMinVar minvar(constrained, Q_COM, ProjMx);  
  Evec = minvar.eigenvectors();
  Evalues = minvar.eigenvalues(); 
 

  un = bn = vn = van = 0.00;
  
  for (i=0; i<3; i++) {
    un += U_COM->dbl(i)/1e3 * Evec->dbl(0,0,i);
    if (doType[0]) bn += B_singleave->dbl(i) * Evec->dbl(0,0,i);
    if (doType[8]) vn += V_singleave->dbl(i) * Evec->dbl(0,0,i);
    if (doType[8]) van += Va_singleave->dbl(i) * Evec->dbl(0,0,i);
  }

  Report(Evalues, Evec, true, "COM",constrained, alpha, typebuf); 

  dBx3 = 0.0;
  if (doType[0]) {
    dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVAB, 0.0, nestsize, 0); 
    dBx3 = err_824(B_singleave, Evec, Evalues, dYij, Q_MVAB, 0.0, nestsize, 1); 
  }
  sigma_Un = err_824(U, Evec, Evalues, dYij, Q_COM,  0.0, nestsize, 3);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]
  //  sigma_Vcis = sigma(V, Evec, nestsize);  // calculate standarddeviation in <V> dot n
  sigma_Vcis = err_824(V_singleave, Evec, Evalues, dYij, Q_COM,  0.0, nestsize, 1);         // calculater |dUn| error after eq 8.23/8.24 - first term from [ni Q_MVAij nj]

  sprintf(buf,"! V_COM  = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	  U_COM->dbl(0)/1e3,U_COM->dbl(1)/1e3,U_COM->dbl(2)/1e3, un, sigma_Un/1e3);
  printf("%s",buf);
  QplugAppendTextDisplay(buf);

  sprintf(buf,"! <Vcis> = [%10.2f  %10.2f  %10.2f] (V'n = %10.2f +/- %8.4f)\n",
	  V_singleave->dbl(0),V_singleave->dbl(1),V_singleave->dbl(2), vn-un, sigma_Vcis);
  printf("%s",buf);
  QplugAppendTextDisplay(buf);

  sprintf(buf,"! <w_Va> = [%10.2f  %10.2f  %10.2f] (%10.2f)\n",Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2), van);
  printf("%s",buf);
  QplugAppendTextDisplay(buf);

  sprintf(buf,"! <Bvec> = [%10.2f  %10.2f  %10.2f] (%10.2f +/- %8.4f)\n",
	 B_singleave->dbl(0),B_singleave->dbl(1),B_singleave->dbl(2), bn, dBx3);
  printf("%s",buf);
  QplugAppendTextDisplay(buf);
  if (GenerateSpreadSheet) {
    FILE *f1 = fopen(filename,"a");
    fprintf(f1,"%g , %g , %g , ",Evec->dbl(0,0,0), Evec->dbl(0,0,1), Evec->dbl(0,0,2));
    fprintf(f1,"%g , %g , %g , ",U_COM->dbl(0)/1e3,U_COM->dbl(1)/1e3,U_COM->dbl(2)/1e3);
    fprintf(f1,"%g , %g , %g , %g ,",un, sigma_Un/1e3,  vn-un, sigma_Vcis);
    fprintf(f1,"%g , %g , %g ,", van, bn, dBx3);
    fprintf(f1,"%g , %g , %g\n",Va_singleave->dbl(0),Va_singleave->dbl(1),Va_singleave->dbl(2));
    fclose(f1);
  }
  

  QplugAppendTextDisplay("\n");

  // for (int i=0; i<10; i++) printf("[%10.3f  %10.3f %10.3f]\n",Erperp->dbl(i,0), Erperp->dbl(i,0), Erperp->dbl(i,0));

  outputs.push_back(Erperp);
  outputs.push_back(Erpar);
 
  return (0);
} 




