
/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	MMS_Array4_2_3vector.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Plugin to decompose MMS 1x4 arrays into 3-vector + magnitude
*
* PURPOSE:      As in FUNCTION
* 
* DESCRIPTION:  MMS variables include 1x4 arrays corresponding to a vector quantity, B, say
*				consisting of (Bx, By, Bz, |B|). This plugin extracts the 3vector and
*				optionally also the magnitude, also adding the appropriate metadata.
*
* ARGUMENTS:
*    IN:      
*             inarray     1x4 array
*             coordsys    Coordinate system of the vector; blank to parse from var name
* 
*   OUT:
*             vecout      output vector
*			  magout      4th array element which holds the magnitude of the vector; 
*                         blank to skip	
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
*
* HISTORY:   SJS 18 Aug 2015; first working version 27 Aug 2015
*            Tweak 31 Aug to remove space in labl_ptr as this fouls qsas labelling
*            1 Sep : removed LABLAXIS from vec; add SC_ID
*/

#include "qplug_if.h"
#include "DvObj.h"

//#define DEBUG_ARRAY4    // for testing

extern "C" QplugReturnStatus DvMMS_Array4_2_3vector(vector<DvObject_var>&, vector<DvObject_var>&);

QplugReturnStatus DvMMS_Array4_2_3vector(vector<DvObject_var>& inputs, vector<DvObject_var>& outputs)
{
	DvObject_var inarray_ptr, coordsys_ptr;
	DvObject_var vecout_ptr, magout_ptr;
		
	DvString out_name,magout_name;				// name of output object(s)
	DvString coordsys, objname, commonStr, lablaxisStr;

 	if(inputs.size() < 2) {
    	QplugAppendTextDisplay ("insufficient inputs\n");
    	return QPLUG_FAILURE; 
 	} 
 
	//	Unpack and check the input data objects
	inarray_ptr = inputs[0];
	if(inarray_ptr->nDims() != 1 && inarray_ptr->arraySize() !=4){
    	QplugAppendTextDisplay ( "Input array is not 1x4\n");
    	return QPLUG_FAILURE; 
	}

	// if coord sys slot left blank, try to parse it from object name
	objname = inarray_ptr->getXrefText("ObjectName");
	coordsys_ptr = inputs[1];
	coordsys = coordsys_ptr->str().trimmed();
	if(coordsys.blank()){
		if (objname.contains("gsm")){
			coordsys = "gsm";
		} else if (objname.contains("gse")){
			coordsys = "gse";
		} else if (objname.contains("bcs")){
			coordsys = "bcs";
		} else if (objname.contains("omb")){
			coordsys = "omb";
		} else if (objname.contains("dmpa")){
			coordsys = "dmpa";
		} else {
			coordsys = "unk";
		}
	}
	QplugAppendTextDisplay (" Coordinate system set to "+coordsys);

	// Get name of output object
 	out_name = GetSlotText(0);
 	if (out_name.empty()) {
    	out_name = objname+"_vec";
    	SetSlotText(0, out_name.c_str());
	}

	vecout_ptr = inarray_ptr->subsetDim(0,0,2);  // not sure this copies xrefs
	DvString frame="vector>";
	frame +=coordsys;
	frame += "_xyz";
	vecout_ptr->setFrameAttr(frame);
	vecout_ptr->change_xref("COORDINATE_SYSTEM",coordsys);
	DvString blank = "";
	DvObject_var rep_ptr = new DvObject(blank,DIM3,1);
	rep_ptr->str(0) = "x";
	rep_ptr->str(1) = "y";
	rep_ptr->str(2) = "z";
	vecout_ptr->change_xref("REPRESENTATION_1",rep_ptr);
	
	// find common string for base of label_ptr and lablaxis
	if(vecout_ptr->xref_exists("LABL_PTR_1")) {
		DvObject_var label_ptr = vecout_ptr->get_xref("LABL_PTR_1");
		if(label_ptr.is_ok() && label_ptr->is_str()) {
			commonStr = label_ptr->str(0).common(label_ptr->str(2));
		} else if (vecout_ptr->xref_exists("LABLAXIS")){
			DvObject_var labelaxisStr = vecout_ptr->get_xref("LABL_PTR_1");
			if (labelaxisStr->is_str()){
					commonStr = labelaxisStr->str();
			} else {
				commonStr = "Vector";
			}
		}
	}
	// vecout will have labl_ptr_1 = "Bx gse" for example, and LABLAXIS = B_GSE
	// Remove LABLAXIS as not required (or correct in addition to LABL_PTR?)
	//DvString labeltext = commonStr;
	//	labeltext +="_"+coordsys;
	//	vecout_ptr->change_xref("LABLAXIS",labeltext); 
	DvObject_var new_label_ptr = new DvObject(commonStr,DIM3,1);
	new_label_ptr->str(0) +="x_"; new_label_ptr->str(0)+=coordsys;
	new_label_ptr->str(1) +="y_"; new_label_ptr->str(1)+=coordsys;
	new_label_ptr->str(2) +="z_"; new_label_ptr->str(2)+=coordsys;

	vecout_ptr->change_xref("LABL_PTR_1",new_label_ptr);
	vecout_ptr->change_xref("TENSOR_ORDER",1);
	
	// add SC_ID
	DvString scid = objname.str()[3];  // objname begins mms1, etc.
	vecout_ptr->change_xref("SC_ID",scid);

	// put back on WL
	outputs.push_back(vecout_ptr);	
	QplugAppendTextDisplay("Vector created\n");

	// if magnitude required (slot not blank) do it now	
	magout_name = GetSlotText(1);
	if (!magout_name.empty()){
		if (magout_name == "auto"){
			magout_name = objname+"_mag";
		}
		SetSlotText(1,magout_name.c_str());
		magout_ptr = inarray_ptr->sliceDim(0,3);
		if (magout_ptr->xref_exists("LABL_PTR_1")){
			magout_ptr->delete_xref("LABL_PTR_1");
		}
		DvString labeltextm = "|";
		labeltextm +=commonStr;
		labeltextm += "|";
		magout_ptr->change_xref("LABLAXIS",labeltextm);
		magout_ptr->change_xref("SC_ID",scid);
		outputs.push_back(magout_ptr);
		QplugAppendTextDisplay("Magnitude created\n");
	}
	
	return QPLUG_SUCCESS;
}


