
/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	MMS_CDF_moments.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Plugin to check/fix metadata, make vectors, 
*                  tensors, etc for MMS FPI cdf files
*
* PURPOSE:      As in FUNCTION
* 
* DESCRIPTION:  Check/add SC_ID, SI_CONVERSION, LABLAXIS, merge components
*                 into vectors, tensors. Also centre timetags.
*
* ARGUMENTS:  (ALL inputs are individually optional; as are outputs!)
*    IN:      
*             density     number density
*             vx, vy, vz  components of bulk velocity
*             tpara,tperp scalar temperature (any order)
*			  qx, qy, qz  components of heat flux
* 
*   OUT:
*             n_out       density (blank to fix in place on WL)
*			  v_out_vec   bulk velocity as qsas vector with Frame (gse)
*			  tpara,tperp temperatures (blank to fix in place on WL)
*             q           heat flux as qsas vector with Frame (gse)
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
*
* HISTORY:   SJS 7 May 2016; 
*/

#include "qplug_if.h"
#include "DvObj.h"

#define DEBUG_FPI_MOMS    // for testing

extern "C" QplugReturnStatus DvMMS_FPI_CDF_moments(vector<DvObject_var>&, vector<DvObject_var>&);

QplugReturnStatus CentreObjTimetags (DvObject_var dobj_ptr);

QplugReturnStatus DvMMS_FPI_CDF_moments(vector<DvObject_var>& inputs, vector<DvObject_var>& outputs)
{
	// Outputs
	DvObject_var n_out_ptr, v_out_vec_ptr, t1_out_ptr, t2_out_ptr, q_out_vec_ptr;  
	DvObject_var nil_ptr;   //use to push back nothing to some slots.
		

#ifdef DEBUG_FPI_MOMS
	printf("starting\n");
#endif
	
 	if(inputs.size() < 9) {
    	QplugAppendTextDisplay ("insufficient inputs\n");
    	return QPLUG_FAILURE; 
 	} 
 
	//	Unpack and process the input data objects
	// ***********************
	// DENSITY
	// ***********************
	DvObject_var density_ptr = inputs[0];
	DvString n_objname = density_ptr->getXrefText("ObjectName");

	if(!density_ptr->is_nil() 
	  && (n_objname.contains("des") || n_objname.contains("dis"))
	  && ( n_objname.contains("mms1") || n_objname.contains("mms2") 
	     || n_objname.contains("mms3") || n_objname.contains("mms4"))
	  && n_objname.contains("density")
	  ){
    	QplugAppendTextDisplay ( "Density being processed");
		// set up output object
		DvString n_out_name = GetSlotText(0);
		if (n_out_name.blank() || n_out_name.empty()){
			n_out_ptr = density_ptr;  // work on WL object
		} else {
			n_out_ptr = new DvObject(density_ptr); //work on copy
		}
		// SI_CONVERSION
		n_out_ptr->change_xref("SI_CONVERSION", "1.e6>m^-3");	

		// SC_ID
		DvString n_objname = n_out_ptr->getXrefText("ObjectName");
		DvString n_scid_str = n_objname.str()[3];//object names start mmsn
		n_out_ptr->change_xref("SC_ID",n_scid_str);
		
		// LABLAXIS
		if (n_objname.contains("dis")){
			n_out_ptr->change_xref("LABLAXIS", "Ni");
		} else {
			n_out_ptr->change_xref("LABLAXIS", "Ne");
		}

		// CENTRE TIMETAGS
		if(CentreObjTimetags(n_out_ptr)==QPLUG_FAILURE){
			QplugAppendTextDisplay("Failed to centre timetags for "+n_out_ptr->getXrefText("ObjectName"));
			return QPLUG_FAILURE;

		}
		
		// push back to outputs if not done on WL object
		if (!n_out_name.blank() || !n_out_name.empty()){
			outputs.push_back(n_out_ptr);
		} else {
			outputs.push_back(nil_ptr);
		}
		
    } else {
		QplugAppendTextDisplay ( "No density input or name does not contain mmsN, density, des or dis, so skipped");
		outputs.push_back(nil_ptr);
	}

	// *************************
	// VELOCITY
	// *************************
	DvObject_var vx_ptr = inputs[1];
	DvObject_var vy_ptr = inputs[2];
	DvObject_var vz_ptr = inputs[3];
	DvString vx_name = vx_ptr->getXrefText("ObjectName");
	DvString vy_name = vy_ptr->getXrefText("ObjectName");
	DvString vz_name = vz_ptr->getXrefText("ObjectName");
	
// initial checks for inputs exist and same seqSize and des or dis
	if(  !vx_ptr->is_nil() 
	  && !vy_ptr->is_nil() 
	  && !vz_ptr->is_nil()
	  && (
	     (vx_name.contains("des") && vy_name.contains("des") && vz_name.contains("des"))
	   ||(vx_name.contains("dis") && vy_name.contains("dis") && vz_name.contains("dis"))
	  )
	  && vx_name.contains("bulkx")
	  && vy_name.contains("bulky")
	  && vz_name.contains("bulkz")
	  ){
	  if (vx_ptr->seqSize()==vy_ptr->seqSize() &&
	     vx_ptr->seqSize()==vz_ptr->seqSize()){
		   // now check s/c number and coord sys
		   DvString vx_sc   = vx_name.str()[3];
		   DvString vy_sc   = vy_name.str()[3];
		   DvString vz_sc   = vz_name.str()[3];
		   DvString vx_coor;
		   if (vx_name.contains("gse")) {vx_coor = "gse";
		     }else {vx_coor="dbcs";}
		   DvString vy_coor;
		   if (vy_name.contains("gse")) {vy_coor = "gse";
		     }else {vy_coor="dbcs";}
		   DvString vz_coor;
		   if (vz_name.contains("gse")) {vz_coor = "gse";
		   } else {vz_coor="dbcs";}
		   if ((vx_sc=="1" || vx_sc=="2" || vx_sc=="3" || vx_sc=="4") &&
		       vx_sc  == vy_sc  && vx_sc   == vz_sc && 
		       vx_coor== vy_coor&& vx_coor == vz_coor){
			 // enough checking [should check brst, etc - sigh ]
		
				QplugAppendTextDisplay ( "Velocity being processed");

				// stolen from makevector		
				DvObject_var x_unit = new DvObject(0.0, DIM3, 1);
				x_unit->dbl(0) = 1.0;
				DvObject_var y_unit = new DvObject(0.0, DIM3, 1);
				y_unit->dbl(1) = 1.0;
				DvObject_var z_unit = new DvObject(0.0, DIM3, 1);
				z_unit->dbl(2) = 1.0;
				
				// convert to vectors
				vx_ptr = vx_ptr * x_unit;
				vy_ptr = vy_ptr * y_unit;
				vz_ptr = vz_ptr * z_unit;
				
				// add SI_CONVERSION here for safety as Dvos checks before
				// performing arithmetic
				vx_ptr->change_xref("SI_CONVERSION","1.e3>m s^-1");
				vy_ptr->change_xref("SI_CONVERSION","1.e3>m s^-1");
				vz_ptr->change_xref("SI_CONVERSION","1.e3>m s^-1");

				// add together
				v_out_vec_ptr = vx_ptr + vy_ptr;
				v_out_vec_ptr += vz_ptr;

				// set up output object
				DvString v_out_name = GetSlotText(1);
				if (v_out_name.blank() || v_out_name.empty()){
					v_out_name = vx_name.replace("x","V");
					SetSlotText(1, v_out_name.c_str());
				}
				// add object name here for centre timetag parsing if needed
				// qplugin interface will rename using sanitised output slot text.
				v_out_vec_ptr->change_xref("ObjectName",vx_name);

				// CENTRE TIMETAGS
				if(CentreObjTimetags(v_out_vec_ptr)==QPLUG_FAILURE){
					QplugAppendTextDisplay("Failed to centre timetags for "+v_out_name);
					return QPLUG_FAILURE;
				}
		
				// SI_CONVERSION
				v_out_vec_ptr->change_xref("SI_CONVERSION", "1.e3>m s^-1");	

				// SC_ID
				v_out_vec_ptr->change_xref("SC_ID",vx_sc);
				// FRAME
				DvString frame = "vector>gse_xyz";    // force gse even for dbcs
				v_out_vec_ptr->change_xref("FRAME",frame);
				if (vx_coor != "gse"){
					v_out_vec_ptr->change_xref("DBCS2GSE","Input in DBCS relabelled as near GSE and given Frame=gse_xyz for qsas convenience");
					QplugAppendTextDisplay("Caution: dbcs frame relabelled as gse for internal QSAS convenience.");
				}
			
				// LABLAXIS & FIELDNAM
				DvString fieldnam_str = "mms";
				fieldnam_str +=vx_sc;
				if (vx_name.contains("dis")){
					v_out_vec_ptr->change_xref("LABLAXIS", "Vi");
					fieldnam_str +="_dis_bulkV";
					v_out_vec_ptr->change_xref("FIELDNAM", fieldnam_str);
				} else {
					fieldnam_str +="_des_bulkV";
					v_out_vec_ptr->change_xref("FIELDNAM", fieldnam_str);
					v_out_vec_ptr->change_xref("LABLAXIS", "Ve");
				}
								
				// push back to outputs
				outputs.push_back(v_out_vec_ptr);
		
			 } else {
			   QplugAppendTextDisplay("Velocity component sc# or coord mismatch; skipped");
			   outputs.push_back(nil_ptr);
			 }

		  } else {
			QplugAppendTextDisplay ("Velocity series lengths; velocity skipped");
			  outputs.push_back(nil_ptr);
		  }


		} else {
			QplugAppendTextDisplay ( "One or more velocity inputs not given, incorrect component or do not contain mmsN, bulk, des, or dis, so skipped");
			outputs.push_back(nil_ptr);

		}
	
	
	// ***********************
	// SCALAR TEMPERATURES
	// ***********************
	// T1
	DvObject_var t1_ptr = inputs[4];  // order immaterial	
	DvString t1_objname = t1_ptr->getXrefText("ObjectName");

	if(!t1_ptr->is_nil() 
	   && (t1_objname.contains("des") || t1_objname.contains("dis"))
	   && (t1_objname.contains("mms1") || t1_objname.contains("mms2") 
	      || t1_objname.contains("mms3") || t1_objname.contains("mms4"))
	   && t1_objname.contains("temp")
	){
    	QplugAppendTextDisplay ( "First temperature being processed");
		// set up output object
		DvString t1_out_name = GetSlotText(2);
		if (t1_out_name.blank() || t1_out_name.empty()){
			t1_out_ptr = t1_ptr;  // work on WL object
		} else {
			t1_out_ptr = new DvObject(t1_ptr);
		}
		// SI_CONVERSION
		t1_out_ptr->change_xref("SI_CONVERSION", "11604.50520>K");	

		// SC_ID
		DvString t1_scid_str = t1_objname.str()[3];//object names start mmsn
		t1_out_ptr->change_xref("SC_ID",t1_scid_str);
		
		// LABLAXIS
		DvString perppara_str, t1_label;
		if (t1_objname.contains("perp")){
			perppara_str="perp";
		} else {
			perppara_str = "para";
		}
		if (t1_objname.contains("dis")){
			t1_label = "Ti_";
			t1_label +=perppara_str;
			t1_out_ptr->change_xref("LABLAXIS", t1_label);
		} else {
			t1_label = "Te_";
			t1_label +=perppara_str;
			t1_out_ptr->change_xref("LABLAXIS", t1_label);
		}

		// CENTRE TIMETAGS
		if(CentreObjTimetags(t1_out_ptr)==QPLUG_FAILURE){
			QplugAppendTextDisplay("Failed to centre timetags for "+t1_out_ptr->getXrefText("ObjectName"));
		return QPLUG_FAILURE;

		}
		
		// push back to outputs if not done on WL object
		if (!t1_out_name.blank() || !t1_out_name.empty()){
			outputs.push_back(t1_out_ptr);
		} else {
			outputs.push_back(nil_ptr);
		}
		
    } else {
		QplugAppendTextDisplay ( "No T1 input, or name does not include mmsN, temp, des, or dis so skipped");
		outputs.push_back(nil_ptr);
	}
	
	// T2
	DvObject_var t2_ptr = inputs[5];  // order immaterial
	DvString t2_objname = t2_ptr->getXrefText("ObjectName");
	
	if(!t2_ptr->is_nil() 
	   && (t2_objname.contains("des") || t2_objname.contains("dis"))
	   && (t2_objname.contains("mms1") || t2_objname.contains("mms2") 
	      || t2_objname.contains("mms3") || t2_objname.contains("mms4"))
	   && t2_objname.contains("temp")
	){
    	QplugAppendTextDisplay ( "Second temperature being processed");
		// set up output object
		DvString t2_out_name = GetSlotText(3);
		if (t2_out_name.blank() || t2_out_name.empty()){
			t2_out_ptr = t2_ptr;  // work on WL object
		} else {
			t2_out_ptr = new DvObject(t2_ptr);
		}
		// SI_CONVERSION
		t2_out_ptr->change_xref("SI_CONVERSION", "11604.50520>K");	

		// SC_ID
		DvString t2_scid_str = t2_objname.str()[3];//object names start mmsn
		t2_out_ptr->change_xref("SC_ID",t2_scid_str);
		
		// LABLAXIS
		DvString perppara_str, t2_label;
		if (t2_objname.contains("perp")){
			perppara_str="perp";
		} else {
			perppara_str = "para";
		}
		if (t2_objname.contains("dis")){
			t2_label = "Ti_";
			t2_label +=perppara_str;
			t2_out_ptr->change_xref("LABLAXIS", t2_label);
		} else {
			t2_label = "Te_";
			t2_label +=perppara_str;
			t2_out_ptr->change_xref("LABLAXIS", t2_label);
		}

		// CENTRE TIMETAGS
		if(CentreObjTimetags(t2_out_ptr)==QPLUG_FAILURE){
			QplugAppendTextDisplay("Failed to centre timetags for "+t2_out_ptr->getXrefText("ObjectName"));
			return QPLUG_FAILURE;
		}
		
		// push back to outputs if not done on WL object
		if (!t2_out_name.blank() || !t2_out_name.empty()){
			outputs.push_back(t2_out_ptr);
		} else {
			outputs.push_back(nil_ptr);
		}
		
    } else {
		QplugAppendTextDisplay ( "No T2 input, or name does not include mmsN, temp, des or dis, so skipped");
		outputs.push_back(nil_ptr);
	}
	
	// ************************
	// HEATFLUX
	// ************************

	DvObject_var qx_ptr = inputs[6];
	DvObject_var qy_ptr = inputs[7];
	DvObject_var qz_ptr = inputs[8];
	DvString qx_name = qx_ptr->getXrefText("ObjectName");
	DvString qy_name = qy_ptr->getXrefText("ObjectName");
	DvString qz_name = qz_ptr->getXrefText("ObjectName");
	
// initial checks for inputs exist and same seqSize and des or dis

	if(  !qx_ptr->is_nil() 
	  && !qy_ptr->is_nil() 
	  && !qz_ptr->is_nil()
	  && (
	     (qx_name.contains("des") && qy_name.contains("des") && qz_name.contains("des"))
	   ||(qx_name.contains("dis") && qy_name.contains("dis") && qz_name.contains("dis"))
	  )	  
	  && qx_name.contains("heatx")
	  && qy_name.contains("heaty")
	  && qz_name.contains("heatz")
	  ){
	  if (qx_ptr->seqSize()==qy_ptr->seqSize() ||
	     qx_ptr->seqSize()==qz_ptr->seqSize()){
		   // now check s/c number and coord sys
		   DvString qx_sc   = qx_name.str()[3];
		   DvString qy_sc   = qy_name.str()[3];
		   DvString qz_sc   = qz_name.str()[3];
		   DvString qx_coor;
		   if (qx_name.contains("gse")) {qx_coor = "gse";
		     }else {qx_coor="dbcs";}
		   DvString qy_coor;
		   if (qy_name.contains("gse")) {qy_coor = "gse";
		     }else {qy_coor="dbcs";}
		   DvString qz_coor;
		   if (qz_name.contains("gse")) {qz_coor = "gse";
		   } else {qz_coor="dbcs";}
		   if ((qx_sc=="1" || qx_sc=="2" || qx_sc=="3" || qx_sc=="4") &&
		       qx_sc  == qy_sc  && qx_sc   == qz_sc && 
		       qx_coor== qy_coor&& qx_coor == qz_coor){
			 // enough checking [should check brst, etc - sigh ]
		
				QplugAppendTextDisplay ( "Heatflux being processed");

				// stolen from makevector		
				DvObject_var x_unit = new DvObject(0.0, DIM3, 1);
				x_unit->dbl(0) = 1.0;
				DvObject_var y_unit = new DvObject(0.0, DIM3, 1);
				y_unit->dbl(1) = 1.0;
				DvObject_var z_unit = new DvObject(0.0, DIM3, 1);
				z_unit->dbl(2) = 1.0;

				// convert to vectors
				qx_ptr = qx_ptr * x_unit;
				qy_ptr = qy_ptr * y_unit;
				qz_ptr = qz_ptr * z_unit;

				// add SI_CONVERSION here for safety as Dvos checks before
				// performing arithmetic
				qx_ptr->change_xref("SI_CONVERSION","1.e-3>W m^-2");
				qy_ptr->change_xref("SI_CONVERSION","1.e-3>W m^-2");
				qz_ptr->change_xref("SI_CONVERSION","1.e-3>W m^-2");

				// add together
				q_out_vec_ptr = qx_ptr + qy_ptr;
				q_out_vec_ptr += qz_ptr;

				// set up output object
				DvString q_out_name = GetSlotText(4);
				if (q_out_name.blank() || q_out_name.empty()){
					q_out_name = qx_name.replace("x","vec");
					SetSlotText(4, q_out_name.c_str());
				}
				// add object name here for centre timetag message
				// qplugin interface will rename using sanitised output slot text.
				q_out_vec_ptr->change_xref("ObjectName",qx_name);

				// CENTRE TIMETAGS
				if(CentreObjTimetags(q_out_vec_ptr)==QPLUG_FAILURE){
					QplugAppendTextDisplay("Failed to centre timetags for "+q_out_name);
					return QPLUG_FAILURE;
				}
		
				// SI_CONVERSION
				q_out_vec_ptr->change_xref("SI_CONVERSION", "1.e-3>W m^-2");	

				// SC_ID
				q_out_vec_ptr->change_xref("SC_ID",qx_sc);
				// FRAME
				DvString frame = "vector>gse_xyz";    // force gse even for dbcs
				q_out_vec_ptr->change_xref("FRAME",frame);
				if (qx_coor != "gse"){
					q_out_vec_ptr->change_xref("DBCS2GSE","Input in DBCS relabelled as near GSE and given Frame=gse_xyz for qsas convenience");
					QplugAppendTextDisplay("Caution: dbcs frame relabelled as gse for internal QSAS convenience.");
				}
		
				// LABLAXIS & FIELDNAM
				DvString fieldnam_str = "mms";
				fieldnam_str +=qx_sc;
				if (qx_name.contains("dis")){
					q_out_vec_ptr->change_xref("LABLAXIS", "qi");
					fieldnam_str +="_dis_heatflux";
					q_out_vec_ptr->change_xref("FIELDNAM",fieldnam_str);
				} else {
					q_out_vec_ptr->change_xref("LABLAXIS", "qe");
					fieldnam_str +="_des_heatflux";
					q_out_vec_ptr->change_xref("FIELDNAM",fieldnam_str);
				}
								
				// push back to outputs
				outputs.push_back(q_out_vec_ptr);
		
			 } else {
			   QplugAppendTextDisplay("Heatflux component sc# or coord mismatch; skipped");
			   outputs.push_back(nil_ptr);
			 }

		  } else {
			QplugAppendTextDisplay ("Heatflux series lengths don't match; heatflux skipped");
			  outputs.push_back(nil_ptr);
		  }


		} else {
			QplugAppendTextDisplay ( "One or more heatflux inputs not given, incorrect component or do not contain mmsN, heat, des, or dis, so skipped");
			outputs.push_back(nil_ptr);

		}
	
	return QPLUG_SUCCESS;
    }

