// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Constrained MVA
// 
// Performs MVA with or without a Ts * minvec = 0 constraint. Based on
// The MVA PlugIn by S.Schwartz
//
// Inputs :
// 	vts_in;		- vector time series or matrix
//      start_end;      - time interval
//      constraints     - 1 for constrain Bn=0, 0 else
//      pgdev_do        - output device # for PGPLOT
//      filename        - filename for output to *.ps file
//
// Outputs
//      ts_out          - ts_in rotated into the new c'sys
//      rotmx           - transformation matrix form ts_in -> ts_out
//
// SEH, 14 May, 2002
//      25 Aug, 2002, covariance matrix as output to WL
//      25 Aug, 2000, <TS> * minvec= 0 where <TS> may be a separate input
//      26 Feb, 2003, timerange, <TS> and <TS> * minvec as outputs 
//      07 Jan, 2004, QSAS 2.0 version
//      15 Jan, 2004, allow for a matrix as input
//      02 May, 2005, output projection matrix
//      08 Aug, 2005, allow selection of covariance matrix type
//      19 Aug, 2005, corrected minor bug in printout of Bn for Weimer type matrix
//      10 Oct, 2005, removed (redundant) link with lapack
//      03 Mar, 2009, removed PGPLOT calls, 
//      07 Mar, 2016, AJA migrate to DVOS native
//      
//
// ------------------------------------------------------------------

#ifndef __cplusplus
#define __cplusplus
#endif

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "MinVar.h"


// needed for QSAS 2.0
extern "C" QplugReturnStatus  DvMinVarB0  ( vector <DvObject_var>&, vector <DvObject_var>&);

DvObject_var Projection(DvObject_var &A);

using namespace QSAS;


QplugReturnStatus  DvMinVarB0 (vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{

  DvObject_var     vts_in;
  DvObject_var     vts_in2;
  DvObject_var     vts_out;
  DvObject_var     vts_out2;
  DvObject_var	     Rmatx;
  DvObject_var	     Proj_matx;
  DvObject_var        Covar_matx;
  DvObject_var	     Eigenvalues;
  DvEvent   start_end;

  const char * filename;
  char attrtxt[150];
  QString str;

  QplugReturnStatus     MV_return;  // used to return from sub-modules

  int    constrained,nseq=2,triage_cnt=0,Qsel=0;
  double deltat=1.0,avg_width=0.0;
  char   evalsevecs[80];
  char   SC__id[127] = "";

  QTime t;
  t.start();                     // start clock
 
  //	 get the input data objects

  vts_in = inputs[0];
  strncpy(SC__id, vts_in->getXrefText("SC_id").c_str(), 1);		//get SC ID

  //  check whether the input is a covarience matrix - if yes - process here.

  if( vts_in.is_nil() ) {
    QplugAppendTextDisplay("no input TS\n");
    return QPLUG_FAILURE;
  }
  
  if(vts_in->seqSize() == 1 && vts_in->arraySize() == 9){
    DvObject_var covEigvals;       // Matrix Dim(3) of eigenvalues
    DvObject_var covEigvecs;       // Matrix Dim(3,3) of eigenvectors (in rows)
    printf("..Covariance matrix supplied...\n");
    QplugAppendTextDisplay("Covariance matrix supplied..\n");
    DvObject_var covMatrix = new DvObject(0.0, DIM33); // covariance matrix
    DvObject_var Rmatx = new DvObject(0.0, DIM33);  
    DvObject_var Covar_matx = new DvObject(0.0, DIM33);  
    DvObject_var Empty_TS = new DvObject(0.0, DIM3, 1);  
    for (int i=0; i<3; i++) for (int j=0; j<3; j++) covMatrix->dbl(0,i,j) = vts_in->dbl(0,i,j);
    for (int i=0; i<3; i++) for (int j=0; j<3; j++) covMatrix->dbl(0,i,j) = covMatrix->dbl(0,j,i); // ensure symmetry
    
    DvMatrixEig eig(covMatrix);
    covEigvecs=eig.eigenvectors();
    covEigvals=eig.eigenvalues();
    for (int i=0; i<3; i++){
      for (int j=0; j<3; j++) {
        Rmatx->dbl(0,i,j) = covEigvecs->dbl(0,i,j);
        Covar_matx->dbl(0,i,j) = vts_in->dbl(0,i,j);
      }
      sprintf(evalsevecs,"%10.5f  [%10.5g, %10.5g, %10.5g ]\n", 
	      covEigvals->dbl(i), covEigvecs->dbl(0,i,0), 
	      covEigvecs->dbl(0,i,1), covEigvecs->dbl(0,i,2));
      str += evalsevecs;
    }  
    QplugAppendTextDisplay (str.toLatin1().data());

    outputs.push_back(Empty_TS);     
    outputs.push_back(Rmatx);   
    outputs.push_back(Covar_matx);   
    outputs.push_back(Covar_matx);   
    return QPLUG_FAILURE;
  }

 
  // ---- check if input_TS is a tensor 

  bool ts_is_tensor = (vts_in->nDims() == 2);
  printf("Tensor input : %ld  %d\n", vts_in->nDims(), ts_is_tensor);

  // check input is cartesian
  DvString frame_ptr = vts_in->getFrameAttr();
  if(  !ts_is_tensor ){
    printf("Checking frame is cartesian...\n");
    if( !frame_ptr.contains("xyz") ){
		QplugAppendTextDisplay ("MinVar Failed >> vector is not cartesian\n");
		return QPLUG_FAILURE;
    }
  }
 

  // ---- get time interval ----------------

  printf("Time interval...\n");
  start_end = inputs[1]->getTimeRange();
  if( start_end.empty() ) return QPLUG_FAILURE;


  // ---- Constrained MVA; calculate so that ts2 * n = 0 + select Covariance matrix type --------

  printf("Checking for constrained...\n");
  constrained = 0;
  Qsel = 0;
  vts_in2  = inputs[2];
  if (vts_in2->totalSize() == 1 ) {
    constrained = vts_in2->asInt();
    Qsel = constrained;
    vts_in2 = vts_in;                      // if not present, use input ts
    if (constrained == 0) printf("Unconstrained...\n");
    if (constrained == 1) printf("Constrained...\n");
  }
  else {    
	if( !vts_in2->isVectorXYZ() ){
		QplugAppendTextDisplay ("MinVar Failed >> constraint is not cartesian vector\n");
		return QPLUG_FAILURE;
	}
    
    printf("Constrained...\n");
    constrained = 1;
  }
  constrained = (constrained == 1) ? 1 : 0; // clean up bad values
 
  QString txtbuf("# Requested constrain and MVA type :\n");
  txtbuf += QString("# Constrain = %1\n").arg(constrained);
  txtbuf += QString("# Covariance matrix type = %1 ").arg(Qsel);
  txtbuf += QString(" (Classic / (Constrained PQP / (Weimer type / (Siscoe type / (extreme Weimer /").section('/',Qsel,Qsel);
  txtbuf += QString(")\n#");
  QplugAppendTextDisplay (txtbuf.toLatin1().data());



  // ---- join TS and TS2 --------------

  DvObject_var masked_TS = vts_in->subSequence(start_end);
  if(masked_TS->seqSize() < 4){
    QplugAppendTextDisplay ("MinVar Failed >> too few data points in time range\n");
    return QPLUG_FAILURE;
  }
  
  DvObject_var masked_TS2 = vts_in2->Join(masked_TS);
    
  int nestsize = masked_TS->seqSize();

  if (nestsize <= 0) {
    QplugAppendTextDisplay("No Data in Interval\n");
    return QPLUG_FAILURE;
  }

  //    ..some output info
  DvObject_var tt = masked_TS->getTimeTags();
  sprintf(attrtxt,"# Time interval = %s - %s (%d samples)",tt->time(0).getISOstring().c_str(),tt->time(nestsize-1).getISOstring().c_str(), nestsize);
  QplugAppendTextDisplay(attrtxt);

#ifdef VERBOSE_MV
  printf(">>MinVar.cc Subsetting done successful I\n");
#endif

  // --- special treatment for tensor_TS as input (covar matrix == tesnor_TS[i]) -------

  if (ts_is_tensor) {
    printf("TS_is tensor \n");
    DvObject_var covEigvals;       // Matrix Dim(3) of eigenvalues
    DvObject_var covEigvecs;       // Matrix Dim(3,3) of eigenvectors (in rows)
    DvObject_var covMatrix = new DvObject(0.0, DIM33);        // covariance matrix
    DvObject_var A  = new DvObject(0.0, DIM3);
    DvObject_var P;
    DvObject_var Q  = new DvObject(0.0, DIM33);
    DvObject_var T;
    DvObject_var C  = new DvObject(0.0, DIM33);
    DvObject_var V  = new DvObject(0.0, DIM33);
    DvObject_var OutMatx_TS = new DvObject(0.0, DIM33, nestsize);
    DvObject_var RotMatx_TS = new DvObject(0.0, DIM33, nestsize);
    DvObject_var CovarMatx_TS = new DvObject(0.0, DIM33, nestsize);
    DvObject_var ProjMatx_TS = new DvObject(0.0, DIM33, nestsize);
 
    for (int k=0; k<nestsize; k++) {
      for (int i=0; i<3; i++) {
        A->dbl(i) = masked_TS2->dbl(k,i);
        for (int j=0; j<3; j++) C->dbl(0,i,j) = masked_TS->dbl(k,i,j);
      }
      C->dbl(0,1,0) = C->dbl(0,0,1);     // force symmetry
      C->dbl(0,2,0) = C->dbl(0,0,2);    
      C->dbl(0,2,1) = C->dbl(0,1,2);    

      // ------- constrain by projection C = P#C#P 
      P = Projection(A);
      T = C * P;
      C = P * T;

      for (int i=0; i<3; i++) {
        for (int j=0; j<3; j++) {
            covMatrix->dbl(0,i,j)   = C->dbl(0,i,j); 
            ProjMatx_TS->dbl(k,i,j) = P->dbl(0,i,j);
        }
      }

      for (int i=0; i<3; i++) {
        cout <<  C->dbl(0,i,0) << " " << C->dbl(0,i,1) << " " << C->dbl(0,i,2) << " " << endl;
      }
      
      covMatrix->dbl(0,1,0) = covMatrix->dbl(0,0,1);
      covMatrix->dbl(0,2,0) = covMatrix->dbl(0,0,2);
      covMatrix->dbl(0,2,1) = covMatrix->dbl(0,1,2);
    
    
#ifdef VERBOSE_MV
  printf(">>MinVar.cc Subsetting done successful Ib\n");
#endif

      // --- find eigenvals, eigenvecs
      DvMatrixEig eig(covMatrix);
      covEigvecs=eig.eigenvectors();
      covEigvals=eig.eigenvalues();
      for (int i=0; i<3; i++){
        for (int j=0; j<3; j++) {
            RotMatx_TS->dbl(k,i,j) = covEigvecs->dbl(0,i,j);                   // save copy for WL output object
            CovarMatx_TS->dbl(k,i,j) = covMatrix->dbl(0,i,j);                           // projected covariance matrix for WL output
            V->dbl(0,i,j) = covEigvecs->dbl(0,i,j);                               // V contain eigenvecs = rotation matrix
            P->dbl(0,i,j) = covEigvecs->dbl(0,j,i);                               // P is the transposed of V
            Q->dbl(0,i,j) = masked_TS->dbl(k,i,j);                          // Q is copy of input matrix
        }
      } 
    }
    
    RotMatx_TS->change_xref(DEPEND_0, tt);
    CovarMatx_TS->change_xref(DEPEND_0, tt);

    OutMatx_TS->copy_xrefs_from(masked_TS);
    ProjMatx_TS->copy_xrefs_from(masked_TS2);
 
    OutMatx_TS->change_xref(FRAME, "vector>mv_xyz");
    ProjMatx_TS->change_xref(FRAME, "vector>mv_xyz");
    
    RotMatx_TS->change_xref("CATDESC", "Time series with rotation matrices");
    CovarMatx_TS->change_xref("CATDESC", "Time series with covariance matrices");
    OutMatx_TS->change_xref("CATDESC", "Time series with the rotated input tensor");
    ProjMatx_TS->change_xref("CATDESC", "Time series with projection matrices");

    outputs.push_back(OutMatx_TS);     
    outputs.push_back(RotMatx_TS);   
    outputs.push_back(CovarMatx_TS);   
    outputs.push_back(ProjMatx_TS);   
    return QPLUG_SUCCESS; 
  }

    
    DvObject_var nilObj;
  //       Do the vector minimum variance analysis
  if (constrained == 1) MV_return = MV_analyseB0(masked_TS, masked_TS2, Eigenvalues, Rmatx, Proj_matx, Covar_matx, constrained, nestsize, SC__id, Qsel);
  else MV_return = MV_analyseB0(masked_TS, nilObj, Eigenvalues, Rmatx, Proj_matx, Covar_matx, constrained, nestsize, SC__id, Qsel);

  printf("%d\n",MV_return);
#ifdef VERBOSE_MV
  // ---- print some output info --------------

  printf(">>MinVar.cc Minimum Variance Analysis done \n");
  printf("Eigenvalues, eigenvectors are:\n");
  for (i=0; i<3; i++){
    sprintf(evalsevecs,"%10.5g  [%10.5g, %10.5g, %10.5g ]\n", 
	    Eigenvalues->dbl(i), Rmatx->dbl(0,i,0), Rmatx->dbl(0,i,1), Rmatx->dbl(0,i,2));
    printf ("%s", evalsevecs);
  }   // end for


  // --- 19.6.2006 - normalized covariance matrix --------


  printf("\nNormalized covariance matrix:\n");  
  double l[4] = {0.0, Eigenvalues->dbl(2),Eigenvalues->dbl(1),Eigenvalues->dbl(0)};
  double w = ((l[1] - l[3])*(l[2] - l[3]))  / (l[3] * sqrt(l[2]*l[1]));
  printf("l = %10.4f  %10.4f  %10.4f, w = %10.4f\n", l[1], l[2], l[3], w);
  for (i=0; i<3; i++){
    sprintf(evalsevecs,"[%10.5g, %10.5g, %10.5g ]\n", 
	    Covar_matx->dbl(0,i,0)/w,
	    Covar_matx->dbl(0,i,1)/w, 
	    Covar_matx->dbl(0,i,2)/w);
    printf ("%s", evalsevecs);
  }   // end for

#endif

  double TS_ave_x=0.0;
  double TS_ave_y=0.0;
  double TS_ave_z=0.0;
  for (int j=0; j<nestsize; j++) {
    TS_ave_x += masked_TS->dbl(j,0);
    TS_ave_y += masked_TS->dbl(j,1);
    TS_ave_z += masked_TS->dbl(j,2); 
  }

  TS_ave_x /= (double) nestsize;    
  TS_ave_y /= (double) nestsize;      
  TS_ave_z /= (double) nestsize; 



  int compsel = Qsel < 1 ? 0 : 1;
  float avebt = sqrt(TS_ave_x*TS_ave_x + TS_ave_y*TS_ave_y + TS_ave_z*TS_ave_z);
  float b_dot_n = TS_ave_x * Rmatx->dbl(0,compsel,0) + TS_ave_y * Rmatx->dbl(0,compsel,1) + TS_ave_z * Rmatx->dbl(0,compsel,2);
  sprintf(attrtxt,"# <Ts in>  = [%6.3f %6.3f %6.3f],  <|Ts in|> = %7.3f,   <Ts in> * n = %10.4f\n\n",TS_ave_x,TS_ave_y, TS_ave_z, avebt, b_dot_n );
  QplugAppendTextDisplay (attrtxt);

  if (MV_return != QPLUG_SUCCESS) return MV_return; 

  // rotate vector TS into MV frame

    DvObject_var rotated_TS =  Rmatx * masked_TS;

    rotated_TS->copy_xrefs_from(masked_TS); // lost as taken from Rmatx
    
  // set attributes

  rotated_TS->change_xref("Eigenvalues", Eigenvalues);
  rotated_TS->change_xref("Eigenvectors", Rmatx);

  DvObject_var name_xref = vts_in->get_xref(OBJECT_NAME);
  Rmatx->change_xref("InputSeriesName", name_xref);
  rotated_TS->change_xref("InputSeriesName", name_xref);
  rotated_TS->change_xref(FRAME, "vector>mv_xyz");
 
  sprintf(attrtxt, "%d", (int)(rotated_TS->seqSize()) );
  Rmatx->change_xref("NumberOfDataValues", attrtxt);
  Covar_matx->change_xref("NumberOfDataValues", attrtxt);
  Eigenvalues->change_xref("NumberOfDataValues", attrtxt);

  Rmatx->change_xref("FIELDNAM", "Rotation matrix");
  Covar_matx->change_xref("FIELDNAM", "Covariance matrix");
  Eigenvalues->change_xref("FIELDNAM", "Eigenvalues");
  Rmatx->change_xref("CATDESC", "Matrix of Eigenvectors as rows");
  Eigenvalues->change_xref("CATDESC", "Eigenvectors are corresponding rows in Rmatx");

  DvObject_var tiObj = new DvObject(start_end);
  Rmatx->change_xref("TimeInterval", tiObj);
  Covar_matx->change_xref("TimeInterval", tiObj);
  Eigenvalues->change_xref("TimeInterval", tiObj);
  Proj_matx->change_xref("TimeInterval", tiObj);

  Rmatx->change_xref("Eigenvalues", Eigenvalues);
  Proj_matx->change_xref("Eigenvalues", Eigenvalues);
  Proj_matx->change_xref(FRAME, "vector>mv_xyz");

    
    //      ...next do the plotting
    mvplot(rotated_TS, Eigenvalues, Rmatx, constrained);

  // put new objects on working list

  //return  QPLUG_FAILURE;
  outputs.push_back(rotated_TS);
  outputs.push_back(Rmatx);   
  outputs.push_back(Covar_matx);   
  outputs.push_back(Proj_matx);   

 
  //	..clean up and exit

  QplugAppendTextDisplay ("MinVar >> - exiting successfully..\n");

  QString ts="Time elapsed = "+QString::number(t.elapsed()/1000.0)+"\n";
  QplugAppendTextDisplay (ts.toLatin1().data());
  return  QPLUG_SUCCESS;
 
}  // end MinVar




// ---------------------------------------------------------
//
//                   MVA analyse 
//
//  Last modified
//  SEH, 26 Aug 2002, - Covariance as output
// ---------------------------------------------------------


QplugReturnStatus  MV_analyseB0(DvObject_var &ts_out,
				DvObject_var &ts_out2,
				DvObject_var &Eigenvalues,
				DvObject_var &Rmatx,
				DvObject_var &Proj_matx,
				DvObject_var &Covar_matx,
				int constrained,int nestsize,
				const char *SC__ID, const int Qsel){

  QString str("# Eigenvalues and eigenvectors are :\n");

  char evalsevecs[80];
  char projmxstr[80];
  char typebuf[80];

  if (constrained) {
    sprintf(typebuf,"MVABC1_SC%s (%d)",SC__ID,nestsize);
  } else {
    sprintf(typebuf,"MVAB_SC%s (%d)",SC__ID,nestsize);
  }   
  str+=typebuf;

    DvMinVariance minvar(ts_out, ts_out2, Qsel);   // dvos minvariance call
    double dot=minvar.orthogonality_quality();
    if(dot>0.001){
      QplugAppendTextDisplay(
			     "MinVar >> Eigenvectors failed simple orthonormal test at level 0.001\n");
    }
    
    Rmatx = minvar.eigenvectors();
    Eigenvalues = minvar.eigenvalues(); 
    Covar_matx = minvar.covariance();
    Proj_matx =  minvar.projection();

  str += "\n";
  for (int i=0; i<3; i++){
    sprintf(evalsevecs,"%10.5f  [%10.5f, %10.5f, %10.5f ]\n", 
	    Eigenvalues->dbl(i), Rmatx->dbl(0,i,0),
	    Rmatx->dbl(0,i,1), Rmatx->dbl(0,i,2));
    str += evalsevecs;
  }   // end for
  QplugAppendTextDisplay (str.toLatin1().data());

  return QPLUG_SUCCESS;
}




// -------------------------------------------------------------------------
//
//  Returns a projection matrix jfr blue book
//
//  SEH, 27 Jul 2005
//
// -------------------------------------------------------------------------
DvObject_var Projection(DvObject_var &A) 
{
  DvObject_var C  = new DvObject(0.0, DIM33);
 
  DvObject_var A2 = A->dot(A);
  double sz2 = A2->dbl();
  
  C->dbl(0,0,0) = 1.0 - (A->dbl(0)*A->dbl(0)) / sz2;
  C->dbl(0,0,1) = - (A->dbl(0)*A->dbl(1)) / sz2;
  C->dbl(0,0,2) = - (A->dbl(0)*A->dbl(2)) / sz2;

  C->dbl(0,1,0) = - (A->dbl(1)*A->dbl(0)) / sz2;
  C->dbl(0,1,1) = 1.0 - (A->dbl(1)*A->dbl(1)) / sz2;
  C->dbl(0,1,2) = - (A->dbl(1)*A->dbl(2)) / sz2;
  
  C->dbl(0,2,0) = - (A->dbl(2)*A->dbl(0)) / sz2;
  C->dbl(0,2,1) = - (A->dbl(2)*A->dbl(1)) / sz2;
  C->dbl(0,2,2) = 1.0 - (A->dbl(2)*A->dbl(2)) / sz2;
  
  return(C);
}

