/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

#ifndef __cplusplus
#define __cplusplus
#endif


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "MinVar.h"

//#define VERBOSE_MV 1



using namespace QSAS;

extern "C" QplugReturnStatus  DvMinVar (vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{

DvObject_var	 vts_in;
DvObject_var	 vts_out;
DvObject_var	 Rmatx;
DvObject_var	 Eigenvalues;
DvEvent   start_end;

DvString attrtxt;

QplugReturnStatus     MV_return;  // used to return from sub-modules

int    i;
char   evalsevecs[80];


//	 get the input data objects

vts_in = inputs[0];

if( vts_in.is_nil() ) return QPLUG_FAILURE;

// check input is cartesian

if( !vts_in->isVectorXYZ() ){
    QplugAppendTextDisplay ("MinVar Failed >> vector is not cartesian\n");
    return QPLUG_FAILURE;
}

start_end = inputs[1]->getTimeRange();

//	...next, subset the time range and data array

 vts_out = vts_in->subSequence(start_end);
 
 if(vts_out->seqSize() < 4){
    QplugAppendTextDisplay ("MinVar Failed >> too few data points in time range\n");
    return QPLUG_FAILURE;

 }

int align = inputs[2]->asInt();

#ifdef VERBOSE_MV
  printf(">>MinVar.cc Subsetting done successful\n");
  fflush(stdout);
#endif

//      ...next do the minimum variance analysis

  MV_return = MV_analyse(vts_out, Eigenvalues, Rmatx, align);
	
#ifdef VERBOSE_MV
  printf(">>MinVar.cc Minimum Variance Analysis done \n");
  printf("Eigenvalues, eigenvectors are:\n");
  for (i=0; i<3; i++){
    sprintf(evalsevecs,"%10.4g  (%10.4g, %10.4g, %10.4g )\n", 
      Eigenvalues->dbl(i),
      Rmatx->dbl(0,i,0),
      Rmatx->dbl(0,i,1), 
      Rmatx->dbl(0,i,2));
    printf ("%s",evalsevecs);
  }   // end for
  fflush(stdout);
#endif

if (MV_return != QPLUG_SUCCESS) return MV_return; 

// rotate vector TS into MV frame
// set Frame info first for DVOS to use

    DvString inFrame = vts_in->getFrameAttr().replace("vector", "rot");
    Rmatx->change_xref(FRAME, inFrame);
    DvString ToFrame = "vector>mv_xyz";
    Rmatx->change_xref("ToFrame", ToFrame);
    
    DvString fieldnamXref = vts_out->getXrefText(FIELDNAM);    
    vts_out =  Rmatx * vts_out;
    vts_out->change_xref(FIELDNAM, fieldnamXref); // removed by multiply
   
//      ...next do the plotting

mvplot(vts_out, Eigenvalues, Rmatx);
    
    // do this after plot to avoid it being used in title
    fieldnamXref += "_MV";
    vts_out->change_xref(FIELDNAM, fieldnamXref); 

#ifdef VERBOSE_MV
  printf(">>MinVar.cc Plotting done successfully\n");
  fflush(stdout);
#endif

 // set attributes

  vts_out->change_xref("Eigenvalues", Eigenvalues);
  vts_out->change_xref("Eigenvectors", Rmatx);
  vts_out->change_xref(FRAME, "vector>mv_xyz");

  DvObject_var name_xref = vts_in->get_xref(OBJECT_NAME);
  Rmatx->change_xref("InputSeriesName", name_xref);
  vts_out->change_xref("InputSeriesName", name_xref);
 
  attrtxt = (int) vts_out->seqSize();
  Rmatx->change_xref("NumberOfDataValues", attrtxt);
  Eigenvalues->change_xref("NumberOfDataValues", attrtxt);

  Rmatx->change_xref("FIELDNAM", "Rotation matrix");
  Rmatx->change_xref("CATDESC", "Matrix of Eigenvectors as rows");
  Eigenvalues->change_xref("FIELDNAM", "Eigenvalues");
  Eigenvalues->change_xref("CATDESC", "Eigenvectors are corresponding rows in Rmatx, rotates vector to MV frame");

    
  DvObject_var tiObj = new DvObject(start_end);
  Rmatx->change_xref("TimeInterval", tiObj);
  Eigenvalues->change_xref("TimeInterval", tiObj);

  Rmatx->change_xref("Eigenvalues", Eigenvalues);

  // put new objects on working list

  outputs.push_back(vts_out);     
  outputs.push_back(Rmatx);   

  QplugAppendTextDisplay ("MinVar >> - exiting successfully\n");
  return  QPLUG_SUCCESS;
 
}  // end MinVar




/***********************************************************************
************************************************************* MV_analyse
***********************************************************************/

QplugReturnStatus  MV_analyse(DvObject_var &ts_out,
                              DvObject_var &Eigenvalues,
                              DvObject_var &Rmatx,
                              int align){

  char evalsevecs[80];
  
  DvObject_var nilObj; // empty constraint
  DvMinVariance minvar(ts_out, nilObj, 0, align);   // dvos minvariance call
  double dot=minvar.orthogonality_quality();

  if(dot>0.001){
     QplugAppendTextDisplay(
    "MinVar >> Eigenvectors failed simple orthonormal test at level 0.001\n");
  }
  Rmatx = minvar.eigenvectors(); 
  Eigenvalues = minvar.eigenvalues(); 


QplugAppendTextDisplay("Eigenvalues, eigenvectors are:\n");
for (int i=0; i<3; i++){
  sprintf(evalsevecs,"%10.4g  (%10.4g, %10.4g, %10.4g )\n",  Eigenvalues->dbl(i), Rmatx->dbl(0,i,0),
	  Rmatx->dbl(0,i,1), Rmatx->dbl(0,i,2));
  QplugAppendTextDisplay (evalsevecs);
}   // end for

return QPLUG_SUCCESS;
}







