/****************************************************************************
							LICENSE

  sf_Pomme -  plug-in for the QSAS software.
  Copyright (C) 2010 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file sf_pomme.cc
 * 
 * \brief Computes the magnetic field model using POMME-7. Based on the code of Stefan Maus.
 * 
 * 
 */

#include <stdio.h>
//the next line is needed for visual studio to find M_PI and others.
#define _USE_MATH_DEFINES
#include <math.h>
#include <string.h>

#include "qdutil.h"
#include "DvObj.h"
#include "qplug_if.h"

#include "./pomme.h"
#include "./POMME7_ext.h"
#include "qrtn.h"  //coordinates rotation

using namespace QSAS;
using namespace std;

/**
 * The class Status has the purpose to print messages either to user or to the Console.
 */
class Status
{
public:
	static const bool kPrintToUserAndConsole = true;
	static const bool kPrintToConsoleOnly = false;
	static void PrintMessage(bool to_user, const char fmt[], ...);
public:
	enum ReturnStatus
	{
		kOk = 0, kUnknownError, kWrongInput, kNoInput, kVectorFieldRequired
	} code;
	std::string Message() const;
	inline void set_message(const std::string message)
	{
		message_ = message;
	}
	;
	Status();
	virtual ~Status();
private:
	std::string message_;
};


// ---------------------------------------------------------------------

const int kCallListIndexOfPosition  = 0;
const int kCallListIndexOfTimeRange = 1;
const int kCallListIndexOfIMFBy		= 2;
const int kCallListIndexOfEm		= 3;
const int kCallListIndexOfEst 		= 4;
const int kCallListIndexOfIst 		= 5;
const int kCallListIndexOff107 		= 6;
const int kCallListIndexOfMaxOrder	= 7;

struct TExternalParameters
{
	//printf("IMF-By 35 minutes earlier (enter '0' if unknown)? ");
	double IMF_By;
	//printf("Merging electric field, 60 minutes earlier (enter '0.5' if unknown)? ");
	double Em;
	//printf("E_st (enter '0' if unknown)? ");
	double E_st;
	//printf("I_st (enter '0' if unknown)? ");
	double I_st;
	//printf("F10.7, 81-day average, 20 months earlier (enter '120' if unknown)? ");
	double f107;
	unsigned int model_number;
	int oder_internal_field;
};

QplugReturnStatus GetPositions(vector<DvObject_var>& inputs, int call_list_index
		, DvEvent &start_end, DvObject_var &V);

QplugReturnStatus ComputePommeModel(DvObject_var &pos, TExternalParameters &parameters, DvObject_var &B_field);

QplugReturnStatus GetExternalParameters(vector<DvObject_var>& inputs, TExternalParameters &parameters);

QplugReturnStatus  ConvertToDesiredFrame (DvObject_var &ts_in, int frame, DvObject_var &ts_out);

// ---------------------------------------------------------------------
extern "C" QplugReturnStatus Dvsf_pomme(vector<DvObject_var>& inputs, vector<DvObject_var>& outputs )
{
//	   INPUT  VECTOR_TIME_SERIES Pos_sat,
//	   INPUT  TIME_INTERVAL      start_end,
//	   INPUT  FLOAT              IMF_By,
//	   INPUT  FLOAT              Em,
//	   INPUT  FLOAT              E_st,
//	   INPUT  FLOAT              I_st,
//	   INPUT  FLOAT              f107,
//	   INPUT  INTEGER            model_no,
//	   OUTPUT VECTOR_TIME_SERIES B_field)
	QplugReturnStatus status = QPLUG_SUCCESS;

	// 1. Importing data into memory --------------------------------------------------------------
	Status::PrintMessage(Status::kPrintToConsoleOnly, "1. Extracting data from the working list\n");

	DvEvent start_end = inputs[kCallListIndexOfTimeRange]->getTimeRange();
	
	DvObject_var P_in;
    

    DvObject_var P_in_orig_frame;
    
    status = GetPositions(inputs, kCallListIndexOfPosition, start_end, P_in_orig_frame);

    if (status == QPLUG_FAILURE || P_in_orig_frame.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: could not read data. Contact the developer.\n");
        return QPLUG_FAILURE;
    }
    
    RtnVector vecEmpty;
    int init_frame = vecEmpty.QrGetFrameFromAttr(P_in_orig_frame);

    status = ConvertToDesiredFrame(P_in_orig_frame, GEOC, P_in);
    

	// Extract external parameters
	TExternalParameters external_params;
    
    status = GetExternalParameters(inputs, external_params);

	// 2. Computing model -------------------------------------------------------------------------
	Status::PrintMessage(Status::kPrintToConsoleOnly, "2. Computing the POMME model.\n");
	
    DvObject_var B_field_geoc;
    
    ComputePommeModel(P_in, external_params, B_field_geoc);
		
	// 3. Preparing output ----------------------------------------------------
	if (status == QPLUG_SUCCESS && B_field_geoc.is_ok())
	{
		Status::PrintMessage(Status::kPrintToConsoleOnly, "3. Placing output into the working list\n");
        
        //Units and SI conversion:
        B_field_geoc->change_xref(FRAME, "vector>geoc_xyz");
        B_field_geoc->change_xref(UNITS, "nT");
        B_field_geoc->change_xref(SI_CONVERSION, "1.0E-9>T");
        B_field_geoc->change_xref(LABLAXIS, "POMME6 B field");
        B_field_geoc->change_xref(FIELDNAM, "POMME6 B field");		
	}

	// 4. Coordinate transformation to the frame of the position vector ---------------------------

	DvObject_var B_field;
    
	if (status == QPLUG_SUCCESS && B_field_geoc.is_ok())
	{
		Status::PrintMessage(Status::kPrintToConsoleOnly, "4. Rotate to the desire frame.\n");
		status = ConvertToDesiredFrame(B_field_geoc, init_frame, B_field);
	}

	// 5. Placing output into the working list ----------------------------------------------------
	if (status == QPLUG_SUCCESS) {
        Status::PrintMessage(Status::kPrintToConsoleOnly, "5. Push back\n");
        outputs.push_back(B_field);
		
	}
	return status;
}

QplugReturnStatus GetPositions(vector<DvObject_var>& inputs, int call_list_index
		, DvEvent &start_end, DvObject_var &position)
{

	DvObject_var P_full = inputs[call_list_index];

	if (P_full.is_nil()) {
		Status::PrintMessage(Status::kPrintToUserAndConsole, "error: Input is not a sequence of vectors.\n");
        return QPLUG_FAILURE;
	}
    
    position = P_full->subSequence(start_end);

    position = position->changeUnitsTo( "1.0e3>m", "km");
    if(position.is_nil()){
        Status::PrintMessage(true, "Position is not in units of distance.\n");
        return QPLUG_FAILURE;
    }
	
	return QPLUG_SUCCESS;
}

QplugReturnStatus GetExternalParameters(vector<DvObject_var>& inputs, TExternalParameters &parameters)
{
	//IMF By
	DvObject_var temp_scalar  = inputs[kCallListIndexOfIMFBy];
	if (temp_scalar.is_nil()) {
		Status::PrintMessage(Status::kPrintToUserAndConsole,
				"error: Could not get IMF By. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
	}
    
    parameters.IMF_By = temp_scalar->asDouble();
	
	//Em
    temp_scalar = inputs[kCallListIndexOfEm];
    
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Em. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
    } 
    
    parameters.Em = temp_scalar->asDouble();
		
	
	//Est
    temp_scalar = inputs[kCallListIndexOfEst];
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Est. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
    } 
    
    parameters.E_st = temp_scalar->asDouble();
		
	
	//Ist
    temp_scalar = inputs[kCallListIndexOfIst];
    
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Ist. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
    } 
    
    parameters.I_st = temp_scalar->asDouble();
		
	
	//F107
    temp_scalar = inputs[kCallListIndexOff107];
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get F107. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
    } 
    
    parameters.f107 = temp_scalar->asDouble();
		
	
	//Main field order.
    temp_scalar = inputs[kCallListIndexOfMaxOrder];
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get the max order of the main field. Check that it is an integer.\n");
        return QPLUG_FAILURE;
    } 
    
    parameters.oder_internal_field = static_cast<unsigned int> (temp_scalar->asInt());
		
	
	return QPLUG_SUCCESS;
}

/**
 * Converts a vector object to the desired new frame.
 * \param[in] ts_in is a vector TS object, which has a well-defined frame
 * \param[in] frame is an integer. Acronyms are set to integer values by #define in qrtn.h
 *        0  GSE,  Geocentric Solar Ecliptic
 *        1  GSM,  Geocentric Solar Magnetic
 *        2  GEOC, Geographic (geocentric)
 *        3  SM,   Solar Magnetic
 *        4  GEI,  Geocentric Inertial
 *        5  MAGD,  Geomagnetic (dipole)
 *        6  HAE,  Heliospheric Aries Ecliptic
 *        7  HEE,  Heliospheric Earth Ecliptic
 *        8  HEEQ, Heliospheric Earth Equatorial
 *        9  GSEQ, Geocentric Solar Equatorial
 * \param[out] ts_out is the converted (rotated) TS object.
 */
QplugReturnStatus ConvertToDesiredFrame (DvObject_var &ts_in, int frame, DvObject_var &ts_out)
{
    
	RtnVector vec;
	vec.n_recs = ts_in->seqSize();
	vec.in_frame = vec.QrGetFrameFromAttr(ts_in);
    vec.out_frame = frame;
    
    vec.IsPositionVector = vec.QrIsPositionVector(ts_in);
	if(vec.IsPositionVector == 1) {
		vec.IsPositionVector = 0; /* issue warning & continue, delta r rotates */
		if(vec.QrOriginsDiffer(vec.in_frame, vec.out_frame)) {
			QplugAppendTextDisplay
			("WARNING: CoordRotation plug-in provides frame rotation only.\n");
			QplugAppendTextDisplay
			("Transformation is between geocentric and heliocentric frames.\n");
			QplugAppendTextDisplay
			("Position vector coordinate transformations will be WRONG.\n");
			QplugAppendTextDisplay
			("Note: Spacecraft separations will rotate OK.\n");
		}
	}

	ts_out = new DvObject(ts_in);
    
	vec.ptr = ts_out;

	vec.t = ts_in->getTimeTags();
	if( (vec.t).is_nil() ) {
		QplugAppendTextDisplay ("Cannot find time tags, some frames are time dependent\n");
		return QPLUG_FAILURE;
	}

	// call C function to do rotation
	long check = vec.QrDoRotation();

	if (check == QRTN_OK || check == SAME_FRAME) {

		ts_out = vec.ptr;
	}
	else{
		ts_out = new DvObject();
		QplugAppendTextDisplay (vec.QrErrorMessage(check));
		return QPLUG_FAILURE;
	} // end if on check

	return  QPLUG_SUCCESS;
}

// ---------  Model functions  -------------------------------

extern void pomme_geocentric (int control[MAXCONTROL],
		double fday, double r, double lat, double lon,
		double est, double ist, double imf_by, double f107, double Em,  /* time lagged but not peaked or subtracted mean */
		double *bx, double *by, double *bz);

/**
 *\param[in] x, y, z : cartesian coordinate of the position with respect to geo-center.
 *\param[out] lon, lat, r : position in spherical coordinates. lon and lat are given in RADIANS!
 */
void ConvertGeoXYZ2LongLatR(double x, double y, double z, double &lon, double &lat, double &r)
{
	double temp = x*x + y*y;
	r = sqrt(temp + z*z);
	lon = atan2(y, x);
	lat = atan2(z, sqrt(temp));
}

/**
 * \param[in] B_nec is an array with 3 elements
 * \param[in] lon, lat (radian) define geographic position of the origin of the vector. These must be in radians..
 * \param[out] B_geo is an array with 3 elements
 */
void ConvertNEC2Geo(double *B_nec, double lon, double lat, double *B_geo)
{
	double Bsph[3];  //frame defined as: [in colatitude direction; radially; eastward]
	Bsph[0] = -B_nec[2];
	Bsph[1] = -B_nec[0];
	Bsph[2] =  B_nec[1];
	double colat = M_PI/2 - lat;
	double sin_col = sin(colat), cos_col = cos(colat), sin_lon = sin(lon), cos_lon = cos(lon);
	B_geo[0] = Bsph[0] * sin_col * cos_lon  +  Bsph[1] * cos_col * cos_lon  +  Bsph[2] * (-sin_lon);
	B_geo[1] = Bsph[0] * sin_col * sin_lon  +  Bsph[1] * cos_col * sin_lon  +  Bsph[2] * ( cos_lon);
	B_geo[2] = Bsph[0] * cos_col            +  Bsph[1] * (-sin_col);
}

QplugReturnStatus ComputePommeModel(DvObject_var &pos, TExternalParameters &parameters, DvObject_var &B_field)
{
	if (pos.is_nil())
	{
		return QPLUG_FAILURE;
	}
	else
	{
		size_t nr_measurements = pos->seqSize();
		B_field = new DvObject(0.0, DIM3, nr_measurements);

		//Initialize model's variables:
		tcontrol control;        /* array of user-control parameters */

		/* mainfield_open_file_mp initializes the control parameters. You can now modify these: */
		control[POS_0] =  parameters.oder_internal_field;  //133;   /* default: model.deg_int, 0=off */
		control[POS_1] =         15;    /* default: model.deg_1st, 0=off */

		control[POS_SM] =        DEG_SM;       /* degree of SM field, default = (*model).deg_sm, 0=off */
		control[POS_SM_IND] =    DEG_SM;       /* degree of field induced by SM field, default = (*model).deg_sm, 0=off */
		control[POS_F107] =      DEG_F107;     /* degree of SM field, default = (*model).deg_sm, 0=off */
		control[POS_F107_IND] =  DEG_F107;     /* degree of field induced by SM field, default = (*model).deg_sm, 0=off */
		control[POS_EST] =       1;                  /* Est dependent field: 1=on, 0=off */
		control[POS_IST] =       1;                  /* Ist dependent field: 1=on, 0=off */
		control[POS_EST_FAC] =   1;                  /* 1 = use model Est factor, 0 = use factor 1.0 */

		control[POS_GSM] =       DEG_GSM;     /* degree of GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_GSM_IND] =   DEG_GSM;     /* degree of field induced by GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_IMFBY] =     1;                 /* IMF-By dependent field: 1=on, 0=off */
		control[POS_IMFBY_IND] = 1;                 /* Induction by IMF-By dependent field: 1=on, 0=off */
		control[POS_EM] =        DEG_EM;      /* degree of GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_EM_IND] =    DEG_EM;      /* degree of field induced by GSM field, default = (*model).deg_gsm, 0=off */


		// Computation of the model for each position
		DvTime day_Jan2000(2000, 1, 1, 0, 0, 0.0);
		DvObject_var tt = pos->getTimeTags();
		
        double lon = 0, lat = 0, r = 0;
		double B_nec[3], B_geo[3];  /* Resulting field */
		for (size_t t = 0; t < nr_measurements; ++t) {
			ConvertGeoXYZ2LongLatR(pos->dbl(t,0), pos->dbl(t,1), pos->dbl(t,2), lon, lat, r);
			// Universal time in floating point days since 00:00:00 on 1/1/2000
			double fday = (tt->time(t) - day_Jan2000) / 86400.0;

			pomme_geocentric (control, fday, r, lat * 180.0 / M_PI, lon * 180.0 / M_PI, parameters.E_st
					, parameters.I_st, parameters.IMF_By, parameters.f107
					, parameters.Em  /* time lagged but not peaked or subtracted mean */
					, &(B_nec[0]), &(B_nec[1]), &(B_nec[2]) );
			ConvertNEC2Geo(B_nec, lon, lat, B_geo);
			B_field->dbl(t,0) = B_geo[0];
			B_field->dbl(t,1) = B_geo[1];
			B_field->dbl(t,2) = B_geo[2];
		}
        B_field->change_xref(DEPEND_0, tt);
	}
	return QPLUG_SUCCESS;
}

// --------------  class Status  -------------------------

Status::Status()
{
	message_ = "";
}

Status::~Status()
{
}

string Status::Message() const
{
	string result = "sf_Pomme: ";
	result += message_;
	return result;
}

void Status::PrintMessage(bool to_user, const char fmt[], ...)
{
	try {
		va_list args;
		char buffer[2048];
		va_start(args, fmt);
		vsprintf(buffer, fmt, args);
		va_end(args);
		if (to_user)
			QplugAppendTextDisplay(buffer);
		printf("%s", buffer);
		fflush(stdout);
	}
	catch (...) {
	}
}
