/****************************************************************************
							LICENSE

  sf_PommeForChamp -  plug-in for the QSAS software.
  Copyright (C) 2010 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file sf_pomme_champ.cc
 * 
 * \brief
 * 
 * 
 */

#include <stdio.h>
#include <math.h>
#include <string.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"

#include "./pomme.h"
#include "./POMME7_ext.h"

using namespace QSAS;
using namespace std;

/**
 * The class Status has the purpose to print messages either to user or to the Console.
 */
class Status
{
public:
	static const bool kPrintToUserAndConsole = true;
	static const bool kPrintToConsoleOnly = false;
	static void PrintMessage(bool to_user, const char fmt[], ...);
public:
	enum ReturnStatus
	{
		kOk = 0, kUnknownError, kWrongInput, kNoInput, kVectorFieldRequired
	} code;
	std::string message() const;
	inline void set_message(const std::string message)
	{
		message_ = message;
	}
	;
	Status();
	virtual ~Status();
private:
	std::string message_;
};


// ---------------------------------------------------------------------

const int kCallListIndexOfAlt  		= 0;
const int kCallListIndexOfLat 		= 1;
const int kCallListIndexOfLon 		= 2;
const int kCallListIndexOfTimeRange = 3;
const int kCallListIndexOfIMFBy		= 4;
const int kCallListIndexOfEm		= 5;
const int kCallListIndexOfEst 		= 6;
const int kCallListIndexOfIst 		= 7;
const int kCallListIndexOff107 		= 8;
const int kCallListIndexOfMaxOrder	= 9;

struct TExternalParameters
{
	//printf("IMF-By 35 minutes earlier (enter '0' if unknown)? ");
	double IMF_By;
	//printf("Merging electric field, 60 minutes earlier (enter '0.5' if unknown)? ");
	double Em;
	//printf("E_st (enter '0' if unknown)? ");
	double E_st;
	//printf("I_st (enter '0' if unknown)? ");
	double I_st;
	//printf("F10.7, 81-day average, 20 months earlier (enter '120' if unknown)? ");
	double f107;
	int oder_internal_field;
};

QplugReturnStatus getPositions(vector<DvObject_var>& inputs, DvEvent &start_end
		, DvObject_var &geo_alt, DvObject_var &geo_lat, DvObject_var &geo_lon);

QplugReturnStatus compute_pomme6model_in_nec_frame(DvObject_var &geo_alt, DvObject_var &geo_lat
		, DvObject_var &geo_lon, TExternalParameters &parameters, DvObject_var &B_field);

QplugReturnStatus get_external_parameters(vector<DvObject_var>& inputs, TExternalParameters &parameters);

// ---------------------------------------------------------------------
extern "C" QplugReturnStatus Dvsf_pomme_champ(vector<DvObject_var>& inputs, vector<DvObject_var>& outputs )
{
//	   INPUT   SCALAR_TIME_SERIES geo_alt,
//	   INPUT   SCALAR_TIME_SERIES geo_lat,
//	   INPUT   SCALAR_TIME_SERIES geo_lon,
//	   INPUT  TIME_INTERVAL      start_end,
//	   INPUT  FLOAT              IMF_By,
//	   INPUT  FLOAT              Em,
//	   INPUT  FLOAT              E_st,
//	   INPUT  FLOAT              I_st,
//	   INPUT  FLOAT              f107,
//	   INPUT  INTEGER            model_no,
//	   OUTPUT VECTOR_TIME_SERIES B_field
	QplugReturnStatus status = QPLUG_SUCCESS;

	// 1. Importing data into memory --------------------------------------------------------------
	Status::PrintMessage(Status::kPrintToConsoleOnly, "1. Extracting data from the working list\n");

	DvEvent start_end = inputs[kCallListIndexOfTimeRange]->getTimeRange();
	
	DvObject_var geo_alt;
    DvObject_var geo_lat;
    DvObject_var geo_lon;
    
    status = getPositions(inputs, start_end, geo_alt, geo_lat, geo_lon);
    if (status == QPLUG_FAILURE) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: could not read data.\n");
        return status;
	}
    
	// Extract external parameters
	TExternalParameters external_params;
    
    status = get_external_parameters(inputs, external_params);

    if (status == QPLUG_FAILURE) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: could not read external parameters.\n");
        return status;
	}

	// 2. Computing model -------------------------------------------------------------------------
	Status::PrintMessage(Status::kPrintToConsoleOnly, "2. Computing the POMME model.\n");

	DvObject_var B_field;
    
    status = compute_pomme6model_in_nec_frame(geo_alt, geo_lat, geo_lon, external_params, B_field);
 

	// 3. Placing output into the working list ----------------------------------------------------
	if (status == QPLUG_SUCCESS && B_field.is_ok())
	{
		Status::PrintMessage(Status::kPrintToConsoleOnly, "3. Placing output into the working list\n");
        //Units and SI conversion:
        B_field->change_xref(FRAME, "vector>nec_xyz");
        B_field->change_xref(UNITS, "nT");
        B_field->change_xref(SI_CONVERSION, "1.0E-9>T");
        B_field->change_xref(LABLAXIS, "POMME6 B field");
        B_field->change_xref(FIELDNAM, "POMME6 B field");
    
        Status::PrintMessage(Status::kPrintToConsoleOnly, "4. Push back\n");
        outputs.push_back(B_field);
	}
	return status;
}


QplugReturnStatus getPositions(vector<DvObject_var>& inputs, DvEvent &start_end
		, DvObject_var &geo_alt, DvObject_var &geo_lat, DvObject_var &geo_lon)
{
	//get altitude
    geo_alt = inputs[kCallListIndexOfAlt];
    geo_alt = geo_alt->subSequence(start_end);
    
    if (geo_alt.is_nil() || geo_alt->seqSize() <= 1) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not understand altitude time series.\n");
        return QPLUG_FAILURE;
    }
    
	//get latitude
    geo_lat = inputs[kCallListIndexOfLat];
    geo_lat = geo_lat->Join(geo_alt);

    if (geo_lat.is_nil() || geo_lat->seqSize() <= 1) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not understand latitude time series.\n");
        return QPLUG_FAILURE;
    }
    
	//get longitude
    geo_lon = inputs[kCallListIndexOfLon];
    geo_lon = geo_lon->Join(geo_alt);

    if (geo_lon.is_nil() || geo_lon->seqSize() <= 1) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not understand longitude time series.\n");
        return QPLUG_FAILURE;
    }
   
	return QPLUG_SUCCESS;
}

QplugReturnStatus get_external_parameters(vector<DvObject_var>& inputs, TExternalParameters &parameters)
{
	//IMF By
	DvObject_var temp_scalar  = inputs[kCallListIndexOfIMFBy];
	if (temp_scalar.is_nil()) {
		Status::PrintMessage(Status::kPrintToUserAndConsole,
				"error: Could not get IMF By. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
	}
    
    parameters.IMF_By = temp_scalar->asDouble();
    
	//Em
    temp_scalar = inputs[kCallListIndexOfEm];
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Em. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
    } 
    
    parameters.Em = temp_scalar->asDouble();

	//Est
    temp_scalar = inputs[kCallListIndexOfEst];
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Est. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
    } 
    
    parameters.E_st = temp_scalar->asDouble();
    
	//Ist
    temp_scalar = inputs[kCallListIndexOfIst];
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Ist. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
    } 
    
    parameters.I_st = temp_scalar->asDouble();
            
	//F107
    temp_scalar = inputs[kCallListIndexOff107];
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get F107. Check that it is a scalar.\n");
        return QPLUG_FAILURE;
    }
    
    parameters.f107 = temp_scalar->asDouble();
		
	//Main field order.
    temp_scalar = inputs[kCallListIndexOfMaxOrder];
    if (temp_scalar.is_nil()) {
        Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get the max order of the main field. Check that it is an integer.\n");
        return QPLUG_FAILURE;
    } 
    
    parameters.oder_internal_field = static_cast<unsigned int>(temp_scalar->asInt());

	return QPLUG_SUCCESS;
}

// ---------  Model functions  ------------------------------------------------

extern void pomme_geocentric (int control[MAXCONTROL],
		double fday, double r, double lat, double lon,
		double est, double ist, double imf_by, double f107, double Em,  /* time lagged but not peaked or subtracted mean */
		double *bx, double *by, double *bz);

void convert_geoxyz2long_lat_r(double x, double y, double z, double &lon, double &lat, double &r)
{
	double temp = x*x + y*y;
	r = sqrt(temp + z*z);
	lon = atan2(y, x);
	lat = atan2(z, sqrt(temp));
}

QplugReturnStatus compute_pomme6model_in_nec_frame(DvObject_var &geo_alt, DvObject_var &geo_lat
		, DvObject_var &geo_lon, TExternalParameters &parameters, DvObject_var &B_field)
{
	if (geo_alt.is_nil() || geo_lat.is_nil() || geo_lon.is_nil())
	{
		return QPLUG_FAILURE;
	}
	else
	{
		size_t nr_measurements = geo_lat->seqSize();
		B_field = new DvObject(0.0, DIM3, nr_measurements);

		//Initialize model's variables:
		tcontrol control;        /* array of user-control parameters */

		/* mainfield_open_file_mp initializes the control parameters. You can now modify these: */
		control[POS_0] =  parameters.oder_internal_field;  //133;   /* default: model.deg_int, 0=off */
		control[POS_1] =         15;    /* default: model.deg_1st, 0=off */

		control[POS_SM] =        DEG_SM;       /* degree of SM field, default = (*model).deg_sm, 0=off */
		control[POS_SM_IND] =    DEG_SM;       /* degree of field induced by SM field, default = (*model).deg_sm, 0=off */
		control[POS_F107] =      DEG_F107;     /* degree of SM field, default = (*model).deg_sm, 0=off */
		control[POS_F107_IND] =  DEG_F107;     /* degree of field induced by SM field, default = (*model).deg_sm, 0=off */
		control[POS_EST] =       1;                  /* Est dependent field: 1=on, 0=off */
		control[POS_IST] =       1;                  /* Ist dependent field: 1=on, 0=off */
		control[POS_EST_FAC] =   1;                  /* 1 = use model Est factor, 0 = use factor 1.0 */

		control[POS_GSM] =       DEG_GSM;     /* degree of GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_GSM_IND] =   DEG_GSM;     /* degree of field induced by GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_IMFBY] =     1;                 /* IMF-By dependent field: 1=on, 0=off */
		control[POS_IMFBY_IND] = 1;                 /* Induction by IMF-By dependent field: 1=on, 0=off */
		control[POS_EM] =        DEG_EM;      /* degree of GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_EM_IND] =    DEG_EM;      /* degree of field induced by GSM field, default = (*model).deg_gsm, 0=off */


		// Computation of the model for each position
		DvTime day_Jan2000(2000, 1, 1, 0, 0, 0.0);
		DvObject_var tt = geo_lat->getTimeTags();
		double r = 0, fday = 0.0;
		double B_nec[3];   /* Resulting field */
		for (size_t t = 0; t < nr_measurements; ++t) {
			// Universal time in floating point days since 00:00:00 on 1/1/2000
			fday = (tt->time(t) - day_Jan2000) / 86400.0;
			r = geo_alt->dbl(t) + 6371.2;

			pomme_geocentric (control, fday, r, geo_lat->dbl(t), geo_lon->dbl(t), parameters.E_st
					, parameters.I_st, parameters.IMF_By, parameters.f107
					, parameters.Em  /* time lagged but not peaked or subtracted mean */
					, &(B_nec[0]), &(B_nec[1]), &(B_nec[2]) );
			B_field->dbl(t,0) = B_nec[0];
			B_field->dbl(t,1) = B_nec[1];
			B_field->dbl(t,2) = B_nec[2];
		}
        
        B_field->change_xref(DEPEND_0, tt);
        
	}
	return QPLUG_SUCCESS;
}

// ---------  class Status  ------------------------------------------------=

Status::Status()
{
	message_ = "";
}

Status::~Status()
{
}

string Status::message() const
{
	string result = "sf_PommeForChamp: ";
	result += message_;
	return result;
}

void Status::PrintMessage(bool to_user, const char fmt[], ...)
{
	try {
		va_list args;
		char buffer[2048];
		va_start(args, fmt);
		vsprintf(buffer, fmt, args);
		va_end(args);
		if (to_user)
			QplugAppendTextDisplay(buffer);
		printf("%s", buffer);
		fflush(stdout);
	}
	catch (...) {
	}
}
