/************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Bow Shock Models
* 
* MODULE:       qbowshock.c
*
* LANGUAGE:     C
*
* PURPOSE:
*
* DESCRIPTION:
* Routines to calculate model bow shock (and magnetopause) surfaces, normals
* based on ISSI Methods for analysing data from multi-satellites book Ch10
* section 4.6, with extensions. See qbowshock.h for commented prototypes.
* Unless stated below, most routines return an error code, passing back
* science values in arguments.
*
* FUNCTIONS: 
*
* High Level Routines:
* ====================
* QBSNormalAtSurfaceGSE  
*      In: Position in Re_GSE
*          Vsw (x) in km/s 
*          Model to use
*      Out:Outward GSE normal
*      Remarks: Aberrates, calls QBSNormalAtSurface, unaberrates
*               Surface is scaled to pass through Pos prior to normal calc.
*
* QBSDistanceAlongNormal 
*      In: Position in Re_GSE
*          Vsw (x) in km/s
*          Model to use
*          Scaling factor to model
*      Out:Signed Distance along outward normal to surface in Re 
*          Position in Re_GSE where normal intersects shock
*
* QBSDistanceAlongVec
*      In:  Position in Re_GSE
*           Vsw (x) in km/s 
*           model to use 
*           scaling to apply to model
*           vector direction to follow
*      Out: returned signed dist along vec 
*           returned GSE pos at shock
*
* Information Routines:
* =====================
*
* QBSListModels:  In: FILE stream (e.g., stdout)
*                 Out:List of models, internal numbering, params to FILE
*
* QBSCheckModelNumber:   In: model number
*                        Returns: QBS_OK or out-of-range error
*                        Remarks: All routines which return an error code
*                                 and take model no. as input call this check.
*
* QBSErrorMessage:In: FILE stream (e.g., stdout) where message printed
*                     error number returned as function value 
*                 Out:Text string describing error code
*
* The static const char* BSMP_name[] holds a string of the name of model
* n_model in BSMP_name[n_model];
*
* Useful Routines:
* ================
* QBSSurface:     In: Position r_vec in Re  [typically rab_vec aberrated]
*                     Model to use
*                     Scale factor for distances in model
*                 Out:Value of S = (r+eps x)^2 - L^2 as returned function value
*
* QBSInside_Outside: In: Position r_vec in Re [typically rab_vec aberrated]
*                        Model to use
*                        Scale factor for distances in model
*                    Out:Inside or outside indicator as returned function value
*
* QBSScaleToXing: In: Position r_vec in Re  [typically rab_vec aberrated]
*                     Model to use
*                 Out: Scaling factor sigma for L and ro
*                 Return: Error if can't fit
*
* QBSNormalAtSurface  In: Position r_vec in Re   [typically rab_vec aberrated]
*                         Model to use
*                     Out:n_vec outward pointing in same frame as input
*                     Remarks: Surface is SCALED to pass through position
*
*
* Utility Routines:
* =================
* QBSAberrate: In: vec in GSE
*                  V_solar_wind in km/s (def = 450 if |input| < 1km/s)
*              Out:vec_ab aberrated by tan^-1(29.8058/V_solar_wind)
*              Returns void
*
* QBSUnAberrate:In: vec 
*                   V_solar_wind in km/s (def = 450 if input < 1km/s)
*               Out:vec de-aberrated by tan^-1(V_Earth/V_solar_wind)
*               Returns void
*
* QBSBuildRotateXY: In: angle in radians for RH rotation about z-axis
*                   Out:rot rotation matrix
*                   Returns void
*
* QBSRotate:    In:  vec
*                    rot rotation matrix
*               Out: vec_rot = (rot)(vec)
*               Returns void
*
* QBSVectorMagnitude:   In:   vec
*                       Out:  |vec| as returned function value
*
* QBSVectorScalarProd:  In:   vec1
*                             vec2
*                       Out:  vec1 dot vec2 as returned function value
*
* QBSDisplace: In: Position r_vec in Re    [typically rab_vec]
*                  Displacement ro_vec in Re_GSE
*              Out:rdispl_vec = r_vec - ro_vec
*              Returns void
* 
* QBSCalcNormalDist:  In:   position (pre-aber and displ)
*                           model eps and L - pre-scaled 
*                     Out:  signed distance along normal in Re 
*                           pt of intersect n,shock; in frame of R
*
* QBSPositionAtShockRoot:  In:  eps (model parameters) 
*                               L [ NB can pass L=1 and R normalised to L ]
*                               position in question (pre-aber, displ)
*                               root of quartic, t,  for R = r+tN, r on shock
*                          Out: returned pos vector r; same units/frame as
*
* QBSChooseDistanceRoot: In: On-axis position, normalised by L
*                            Dist to axis, normalised by L
*                            eccentricity
*                            roots t of r + tN = R vec eqn
*                                (only real ones; others garbage
*                            nreal
*                        Out:chosen root (closest, correct branch, ... )
*
* QBSSolveQuartic:  In: coefficients of quartic; c[0] const term
*                   Out:sorted array of returned real roots
*                       number of real roots returned
*
* QBSBracketQuarticRoots: In: coefficients of quartic;c[0] const term
*                         Out:low and high values bracketing root
*                         Remarks: Solves cubic for max/min
*
* poly:  In: value of argument
*            highest pwr of t inpolynomial
*            coefficients, c[0] constant term
*        Out:evaluated polynomial
*            evaluated derivative
*        Returns void
*
* rtsafe_poly:  In: polynomial coeffs. c[0] constant term
*                   highest power of x
*                   lower bound for bracketed root
*                   upper bound for bracked root
*                   absolute accuracy for root
*               Out:found root
*               Remarks: polynomial implementaion of Numerical Recipees rtsafe
*
* QBSSolveCubic: In: coeffs of cubic; c[0] = constant term
*                Out:returned real roots - sorted
*                    no of real roots; complex roots NOT retnd
*                Remarks: uses rtsafe_poly, divides out root, quadratic formula
*
* QBSDoubleCompare: In:two double values to compare (for qsort)
*                   Returns: -1 of 1st arg less than 2nd
*                            0 if equal
*                            +1 if 1st arg larger than 2nd
*
* QBSCalcDistAlongVec: In: position (pre-aber and displaced)
*                          model eps and L - pre-scaled
*                          vector direction to follow (pre-aberrated) 
*                      Out:signed distance along vector in Re 
*                          pos at shock in frame of R
*
* QBSChooseAlongVecRoot: In: pre-aber, displaced position
*                            model eps and  scaled L 
*                            unit direction vector to follow (pre-aber) 
*                            roots to quadratic for t soln R = r +t vec
*                        Out:selected root
*                        Remarks: rejects root on wrong branch, more distant,...
*
* NOTES:
*
* COPYRIGHT:
* ---------------------------------------------------------------------
* Copyright (c) 1999 Steve Schwartz
*
* Astronomy Unit, Queen Mary and Westfield College, London, UK.
* This software is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
* If you are associated with the Cluster project, you may copy, or
*   modify this software freely, solely for purposes associated with
*   the Cluster project, and provided that you maintain a complete
*   copy of this copyright message with the software.
* If this software is distributed to you at the same time as a separate
*   file called COPYRIGHT, then the terms of copyright listed in that
*   file apply to this software.
* ---------------------------------------------------------------------
*
*
* HISTORY: SJS April 1999
* Small mods July 2001 for use with QSAS Boundary Normal Coord Plugin
* and changed arg of QBSChooseDistanceRoot to XX to avoid clash on Solaris
*
***********************************************************************/


#include "qbowshock.h"  

/************************************************* QBSAberrate *******/

/* aberrate a vector according to solar wind velocity */

void QBSAberrate( float* vec_gse,              /* input vec in gse xyz */
                  float vsw,                   /* sw speed in km/s     */
                  float *vec_ab)               /* vec in aberrated frame*/
{

float angle;
float rot[3][3];

if ( fabs(vsw) < VSW_MIN ) vsw = VSW_DEFAULT;

angle = - atan (V_EARTH/vsw);   /* NB aberration <=> negative rot about Z */
QBSBuildRotateXY(angle,rot);
QBSRotate(vec_gse, rot, vec_ab);

return;
}

/************************************************* QBSUnAberrate *******/

/* un-aberrate a vector according to solar wind velocity */

void QBSUnAberrate( float* vec,              /* input vec in gse xyz */
                    float vsw,               /* sw speed in km/s     */
                    float *vec_unab)         /* vec in aberrated frame*/
{

float angle;
float rot[3][3];

if ( fabs(vsw) < VSW_MIN ) vsw = VSW_DEFAULT;

angle = + atan (V_EARTH/vsw);   /* NB aberration <=> negative rot about Z */
QBSBuildRotateXY(angle,rot);
QBSRotate(vec, rot, vec_unab);

return;
}

/*************************************************QBSBuildRotateXY ******/

/* take angle for rotation about z-axis and build rotation angle  */

void QBSBuildRotateXY( float angle, /* angle in RADIANS: + in RH sense wrt z*/
                       float rot[3][3] ) /* rotation matrix rot[row][col]*/
{

rot[0][0] = cos(angle);    /* rot[row][column] */
rot[0][1] = sin(angle);
rot[0][2] = 0.0;
rot[1][0] = -sin(angle);
rot[1][1] = cos(angle);
rot[1][2] = 0.0;
rot[2][0] = 0.0;
rot[2][1] = 0.0;
rot[2][2] = 1.0;

return;
}

/******************************************** QBSRotate ****************/

/* Rotate input vector using rotation matrix   */

void QBSRotate( float* vec,            /* input vector */
                float rot[3][3],            /* rotation matrix */
                float* vec_rot)        /* rotated vector */
{

int i,j;

/* reminder: array storage is rot[row][column]  */

for( i=0; i<3; i++){
   vec_rot[i] = 0;
   for ( j=0; j<3; j++){
      vec_rot[i] += rot[i][j] * vec[j];
   }
}
return;
}

/******************************************** QBSDisplace *************/

/* displace vector by -ro */

void QBSDisplace( float* vec,
                  float* ro_vec,
                  float* rdispl_vec )     /* rdispl_vec = vec - ro_vec */
{

int i;

for (i=0; i<3; i++){ 
  rdispl_vec[i] = vec[i] - ro_vec[i];
}

return;

}

/*****************************************************QBSVectorMagnitude */

float QBSVectorMagnitude (float *vec)
{

float v_mag;

v_mag = QBSVectorScalarProduct(vec,vec);

return sqrt (v_mag);
}

/***************************************************QBSVectorScalarProduct */

float QBSVectorScalarProduct(float *vec1,
                             float *vec2)
{
int i;
float prod;

prod = 0.0;
for (i = 0; i<3; i++){
   prod += vec1[i] * vec2[i];
}
return prod;
}


/******************************************************** QBSSurface ******/

/* For given input position (aberrated if required by CALLING routine)
*  calculate the value of the surface S = ( r + eps x )^2 - L^2 */


float QBSurface( float* r_vec,     /* position, already aberrated if req'd */
               QBSMP_models model, /* which model to use */
               float sigma )       /* scaling for L, ro; 1 means no scaling */
{


float eps, L, ro[3];
float  s,rdisp[3], rmod, scratch;

eps   = BSMP_eps[model];         /* get model parameters and */
L     = sigma * BSMP_L[model];   /* scale distances */
ro[0] = sigma * BSMP_xo[model];
ro[1] = sigma * BSMP_yo[model];
ro[2] = 0.0;
         
QBSDisplace ( r_vec, ro, rdisp);  /* displace origin according to model */

rmod = QBSVectorMagnitude(rdisp);

scratch = rmod + eps * rdisp[0];

s = scratch*scratch - L*L;

return s;
}


/************************************************** QBSInside_Outside ***/

/* Determine whether a given position (pre-aberrated by calling routine if
*  required, is inside or outside a (scaled) model surface */

QBSMP_IN_OR_OUT QBSInside_Outside(  
           float*        r_vec,   /* pos (pre-aberrated) */
           QBSMP_models  model,   /* which model to use */
           float         sigma )  /* scaling for L, ro; 1 means no scaling */
{
float eps, L, ro[3];
float rdisp[3], rmod, scratch;
float s;

eps   = BSMP_eps[model];         /* get model parameters and */
L     = sigma * BSMP_L[model];   /* scale distances */
ro[0] = sigma * BSMP_xo[model];
ro[1] = sigma * BSMP_yo[model];
ro[2] = 0.0;

QBSDisplace ( r_vec, ro, rdisp);  /* displace origin according to model */

rmod = QBSVectorMagnitude(rdisp);

scratch = rmod + eps * rdisp[0];  /* if this is negative, we MUST be inside */
                                  /* This occurs only for hyperb (eps > 1)  */
                                  /* when behind line through O || asymptote*/

if (scratch < 0.0 ) return INSIDE;

s = QBSurface (r_vec, model, sigma) ;  /* All other cases value of S enough */

if (s <  0.0) return INSIDE;
if (s >  0.0) return OUTSIDE;
if (s == 0.0) return ON_SURFACE;

return CONFUSED;    /* Hmmm. should never get here! */
}
           
/******************************************************* QBSScaleToXing ***/

/* take position (pre-aberrated by calling routine if required) and scale
*  given bow shock model to fit it. Scaling parameter sigma is largest
*  positive root of quadratic:
*
*    A sigma**2 + B sigma + C = 0 with
*
*  A = ro**2 - (eps xo + L)**2
*  B = -2 [ r dot ro - eps x ( eps xo +L) ]
*  C = r**2 - (eps x)**2
*
* Analysis shows that, at least in the simple case of ro = 0,
*
* HYPERBOLIC models (eps > 1) have 2 roots which can be both +ve, one +ve or
* neither positive (this latter arises essentially if the point is inside
* the lines passing through the origin parallel to the asymptotes). In the
* case of two positive roots, the small one corresponds to shrinking until
* the opposite branch of the hyperbola passes through the point.
*
* PARABOLIC models ( eps = 1) generally have only 1 positive root except
* if point is along axis, in which case the second root is zero
*
* ELLIPTIC models (eps < 1)  only have one positive root.
*
* Thus when the roots of the above equation are found, the decision depends
* on the NUMBER of POSITIVE ROOTS:
*
* 0        no scaling will make the model pass through the given point
* 1        take it
* 2        take the larger root
*/

int QBSScaleToXing (float* r_vec,         /* position, pre-aberrated if req'd */
                   QBSMP_models  model,   /* which model to use */
                   float* sigma )         /* returned scaling for L, ro; */
                         /* Multiply L,ro by sigma to make model fit point */
{                    
float eps, L, ro[3];
float ro_mag, eps_xo_plus_L, r_mag;
float A,B,C, discrim, root1, root2;
int check;

check = QBSCheckModelNumber(model);
if (check != QBS_OK) return check;   /* check model number */

eps   = BSMP_eps[model];         /* get model parameters */
L     = BSMP_L[model];  
ro[0] = BSMP_xo[model];
ro[1] = BSMP_yo[model];
ro[2] = 0.0;

ro_mag = QBSVectorMagnitude(ro);
r_mag  = QBSVectorMagnitude(r_vec);
eps_xo_plus_L = eps*ro[0] + L;

A = ro_mag*ro_mag - (eps_xo_plus_L)*(eps_xo_plus_L);
B = -2 * ( QBSVectorScalarProduct(r_vec,ro) - eps*r_vec[0]*(eps_xo_plus_L) );
C = r_mag*r_mag - eps*eps*r_vec[0]*r_vec[0];

discrim = B*B - 4.0 * A * C;

if (discrim < 0) return DISCRIM_NEGATIVE_IN_SCALE; /* can't find real roots */

discrim = sqrt(discrim);

root1 = (-B + discrim)/(2.0 * A);
root2 = (-B - discrim)/(2.0 * A);

*sigma = (root1 > root2) ? root1 : root2 ; /* max of root1, root2 */

if ( *sigma > 0) return QBS_OK;
else {
   *sigma = 1.0; /* set to something safe just in case the user doesn't check*/
   return LARGEST_SCALING_ROOT_NOT_POSITIVE;
}

} /* end QBSScaleToXing */

/********************************************** QBSNormalAtSurface *********/  

/* Given (pre-aberrated) position vector and model, scale the model to
*  pass through the given point and calculate the outward normal to 
*  the model surface at that point. The normal is given in the same
*  frame (the routine has no aberration information) and needs to be
*  unaberrated if required.
*
* Algorithm:
*	Scale model, and displace origin.
*       Calculate grad S = [ xdisp(1-eps**2) + eps*L,
*                            ydisp,
*                            zdisp ]
*       Normalise, and flip so nx>0.
*/

int QBSNormalAtSurface ( 
                   float* r_vec,         /* position, pre-aberrated if req'd */
                   QBSMP_models  model,  /* which model to use */
                   float* normal)        /* same system as r_vec */
{

float eps, L, ro[3];
float sigma, rdisp[3], nmag;
int i;
int check;

check = QBSCheckModelNumber(model);
if (check != QBS_OK) return check;   /* check model number */

check = QBSScaleToXing (r_vec, model, &sigma);  /* get scaling */
if (check != QBS_OK) return check;

eps   = BSMP_eps[model];         /* get model parameters */
L     = sigma * BSMP_L[model];   /* scale distances */
ro[0] = sigma * BSMP_xo[model];
ro[1] = sigma * BSMP_yo[model];
ro[2] = 0.0;

QBSDisplace ( r_vec, ro, rdisp);  /* displace origin according to model */

normal[0] = rdisp[0] *( 1-eps*eps) + eps * L;  /* calc normal */
normal[1] = rdisp[1];
normal[2] = rdisp[2];

nmag = QBSVectorMagnitude(normal);            /* normalise */
for ( i=0; i<3; i++) normal[i] /= nmag;

/* code to flip normal if nx < 0 removed. This is always the OUTWARD normal*/

return QBS_OK;
}


/********************************************* QBSNormalAtSurfaceGSE ****/

/*  Put together routines to take input as:
*        r_vec in Re_GSE
*        Vsw   in km/s
*        model
*
*  Aberrate, find normal (scales surface to pass through r_vec), Unaberrate
*/

int QBSNormalAtSurfaceGSE( float* r_vec,   /* Position in Re_GSE */
                           float  vsw,     /* Vsw (x) in km/s */
                           QBSMP_models nmodel, /* model to use */
                           float* normal_vec) /*returned outward GSE normal */
{

float rab_vec[3], aber_normal[3], nmag;
int check, i;

check = QBSCheckModelNumber(nmodel);
if (check != QBS_OK) return check;   /* check model number */

QBSAberrate(r_vec, vsw, rab_vec);    /* pos vector in aberrated system */

check = QBSNormalAtSurface (rab_vec, nmodel, aber_normal);
if (check != QBS_OK) return check;

QBSUnAberrate(aber_normal, vsw, normal_vec); /* get normal in unAberrated GSE*/
nmag = QBSVectorMagnitude(normal_vec);
for (i=0; i<3; i++) normal_vec[i] /= nmag;/*renormalise to reduce rounding err*/

return QBS_OK;

}

/********************************************* QBSDistanceAlongNormal ****/

/*
*  Aberrate Position
*  Scale L, ro
*  Displace Pos
*  Calc Distance
*  
*/

int QBSDistanceAlongNormal(float* r_vec,   /* Position in Re_GSE */
                           float  vsw,     /* Vsw (x) in km/s */
                           QBSMP_models model, /* model to use */
                           float  sigma,    /* mult scaling to apply to model*/
                           float* distance,/*returned signed distance along n */
                           float* r_at_shock)/* returned GSE pos at shock */
{

float rab_vec[3], rabd_vec[3],ro[3],L, eps;
float rabd_atshock[3], rab_atshock[3];
float minus_ro[3];
int i,check;

check = QBSCheckModelNumber(model);
if (check != QBS_OK) return check;   /* check model number */

QBSAberrate(r_vec, vsw, rab_vec);  /* pos vector in aberrated coords */

eps   = BSMP_eps[model];         /* get model parameters */
L     = sigma * BSMP_L[model];   /* scale distances */
ro[0] = sigma * BSMP_xo[model];
ro[1] = sigma * BSMP_yo[model];
ro[2] = 0.0;

QBSDisplace (rab_vec, ro, rabd_vec);  /* displace position according to model*/

check = QBSCalcNormalDist(rabd_vec, eps, L, distance, rabd_atshock); 
                           /* calc distance and shock intersect with normal */

for(i=0; i<3; i++) minus_ro[i] = - ro[i];   /* un-displace r_at_shock */
QBSDisplace (rabd_atshock, minus_ro, rab_atshock);

QBSUnAberrate ( rab_atshock, vsw, r_at_shock ); /* and unaberrate it */

return check;

}

/********************************************** QBSCalcNormalDist *****/
/*
*   Take position and model (pre-aberrated and dispalced)
*   Normalise positions to L and calc coefficients of quartic
*   Find real roots of quartic
*   Choose "correct" root
*   Calculate distance, pos at shock (same frame as input) and return
*/

int QBSCalcNormalDist( float* R,          /* position (pre-aber and displ)*/
                       float eps,         /* model eps and L - pre-scaled */
                       float L, 
                       float* distance,  /*signed distance along normal in Re*/
                       float* r_at_shock) /* intersect n,shock; in frame of R*/

{


float XX, rho, rhosq, chisq, epssq; /* set up vars */
double c[5];
double droots[4];                  /* returns from quartic solver */
float roots[4], chosen_root;       /* single precision equivs */        
float dr[3];
int nreal, check,i;

XX    = R[0]/L;
rhosq = (R[1]*R[1] + R[2]*R[2])/(L*L);
rho   = sqrt(rhosq);
chisq = XX*XX - (1-eps*XX)*(1-eps*XX);
epssq = eps*eps;
c[4]  = epssq - 1;
c[3]  = 2.0*(epssq-2.0);
c[2]  = epssq - 5.0 + chisq + rhosq*(1.0-epssq)*(1.0-epssq);
c[1]  = 2.0 * ( -1.0 + chisq + rhosq*(1.0 - epssq) );
c[0]  = chisq + rhosq;

check = QBSSolveQuartic(c, droots, &nreal);

if (check != QBS_OK){
     if (check == DEGENERATE_1_ROOT_AT_TANGENCY){
         QBSErrorMessage (stdout, check);  /* continuing */
     } else { return check;}
}

for (i=0; i<nreal; i++) roots[i] = (float)droots[i]; /*convert to single prec*/

check = QBSChooseDistanceRoot(XX,rho,eps,roots,nreal,&chosen_root); 
if (check != QBS_OK) return check;

check = QBSPositionAtShockRoot(eps,L,R,chosen_root,r_at_shock);
if (check != QBS_OK) return check;

QBSDisplace ( R, r_at_shock, dr);         /* find dr = R - r_at_shock */
*distance = QBSVectorMagnitude (dr);      /* distance from shock to pt */
if(chosen_root < 0.0) *distance *= -1.0;  /* make negative if root < 0 */

return QBS_OK;
}

/****************************************** QBSPositionAtShockRoot*********/
/*
* Take model parameters (pre-scaled)
* position R in question (pre aberrated and displaced)
* root t of quartic which is solution to vector equation
*
* r + tN(r) = R   with 
*
* r laying on shock surface
* N = r + eps |r| x_hat the unnormalised normal vector
*
* Find r
*/

int QBSPositionAtShockRoot(float eps,  /* model parameters */
                    float L,    /* NB can pass L=1 and R normalised to L */
                    float *R,   /* position in question (pre-aber, displ)*/
                    float t,    /* root of quartic */
                    float* r)   /*returned pos vector; same units/frame as R*/
{

r[0] = (R[0] - eps*t*L)/(1.0 + t*(1.0-eps*eps));
r[1] = R[1]/(1.0 + t);
r[2] = R[2]/(1.0 + t);

return QBS_OK;
}

/*******************************************QBSChooseDistanceRoot **********/

/* Given position along axis and perp (X,rho), normalised to L
*  roots [0,1,2,3] of quartic which solves r+tN = R (see QBSPositionAtShockRoot)
*  of which there are nreal roots, choose the root which corresponds to
*  desired point on the bow shock s.t. the normal passes through this point
*  and goes through (X,rho). Eliminate roots on wrong branches, more
*  distant roots, etc.
*
*  Return "chosen" root in *chosen_root
*
*  Analysis shows that for both ellipsoidal surfaces and hyperbolic ones
*  there can be either two real roots (and 2 complex) or 4 real roots.
*  The analysis depends on what kind of conic:
*
*  Hyperbolae
*  ==========
*  Reject roots on the wrong branch. In the conic representation 
*     L/r = 1 + eps cos theta this branch has r<0. Re-writing this
*     and solving for r yields r = L - eps x, so test for (L - eps x) < 0
*  If more than one real root remains, take the smallest. This occurs for
*     points deep enough inside the conic (for which the "correct" root
*     may not be the most sensible, but what the heck).
*
*  Ellipses
*  ========
*  Take the smallest distance amongst the real roots. For things downtail,
*     this may turn out to be even further downtail rather than upwind/
*     on the flanks, but this is an indication tha tthe model is inappropriate
*     rather than the choice incorrect.
*
*  Parabolae
*  =========
*  These only have one "branch" (the other is at infinity) and "r" never
*     goes negative, so like ellipse
*
*  ALL cases can therefore be treated by:
*   Reject roots with (L - eps x < 0)
*   Take smallest distance from those remaining
*/

int QBSChooseDistanceRoot(float  XX,      /* On-axis position, normalised by L */
                          float  rho,    /* Dist to axis, normalised by L */
                          float  eps,    /* eccentricity */
                          float* roots,  /* roots t of r + tN = R vec eqn */
                                         /* (only real ones; others garbage */
                          int    nreal,  /* number of real roots */
                          float* chosen_root) /* chosen root */
{

float r, rs[3], R[3], dr[3], dist[4], poss_roots[4];
int n_poss;
int i, i_choice;

/* make sure something to work with */

if (nreal <= 0) return NO_REAL_ROOTS_FOR_CLOSEST_DIST_ALONG_NORMAL;
if (nreal > 4) return TOO_MANY_REAL_ROOTS_FOR_DIST_ALONG_NORMAL;

/* loop over all real roots and reject those on wrong branch. Calculate
*  distance for possible ones */

R[0] = XX;      /* Fake position vector for local use */
R[1] = rho;
R[2] = 0.0;

for (i = 0, n_poss = 0; i< nreal; i++){

   QBSPositionAtShockRoot(eps, 1.0, R, roots[i], rs); /* find rs for roots[i] */
   r = 1.0 - eps * rs[0];
   if (r>= 0.0){                              /* only accept correct branch */
      poss_roots[n_poss] = roots[i];          /* store possible root*/
      QBSDisplace ( R,  rs, dr );             /* find dr = R - rs */
      dist[n_poss] =  QBSVectorMagnitude (dr);/* distance from shock to pt */
      n_poss++;                               /* increment index/counter */
   }
}

if (n_poss == 0) return NO_POSSIBLE_ROOTS_FOR_DIST_ALONG_NORMAL;
i_choice = 0;

for ( i = 1 ; i<n_poss; i++){
   if ( dist[i] < dist[i_choice] ) i_choice = i;
}

*chosen_root = poss_roots[i_choice];
return QBS_OK;
}

/******************************************** QBSSolveQuartic ************/

/* Solve quartic eqn c4 t^4 + c3 t^3 + c2 t^2 + c1 t + co = 0
*
*  Find root[0] using safe Newton-Raphson algorithm 
*  Factor out t-root[0] and solve remaining cubic 
*
*  Return roots sorted in roots[0-3] and the number of real ones in *nreal
*  Only values in roots[0 -> *nreal -1] are meaningful
*/

int QBSSolveQuartic(double* c,     /* coefficients of quartic; c[0] const term*/
                    double* roots, /*array of returned roots; only real ones ok*/
                    int*   nreal) /* number of real roots returned */
{
double t_low, t_high;
int check;
double acc = ABSOLUTE_ROOT_ACCURACY, c3[4], rem;
int i;


check = QBSBracketQuarticRoots(c, &t_low, &t_high);
if (check != QBS_OK) {
      if (check == DEGENERATE_1_ROOT_AT_TANGENCY){
           roots[0] = t_low;
           *nreal   = 1;
      }
      return check;
}

/* use rtsafe algorithm to find bracketed root */

check = rtsafe_poly (c , 4, t_low, t_high, acc, &roots[0]);
if (check != QBS_OK) return check;

/* divide out 1st root to leave cubic */
rem = c[4];
for (i = 3; i>=0; i--) {
   c3[i] = rem;
   rem = c[i] + rem*roots[0];
}
check = QBSSolveCubic ( c3, &roots[1], nreal );     /* solve cubic */
if (check != QBS_OK) return check;
*nreal += 1;  /* add root count to include roots[0] */
qsort (roots, *nreal, sizeof(double), QBSDoubleCompare); /* sort roots */
return QBS_OK;
}




/************************************* QBSBracketQuarticRoots ************/

/* Find values which brackets roots of quartic eqn 
*  c4 t^4 + c3 t^3 + c2 t^2 + c1 t + co = 0
*
*  Solve cubic equation for max and min
*  If only 1, 3 cases:
*     It is also a root (within tolerance)
*     There are no roots
*     At least one root lies between it and both of +infnty, - infnty
*  If 3, either:
*     Root lies between max and min
*     At least one root lies between outer max/min and correspond infnty
*

*/

int QBSBracketQuarticRoots(double* c,/* coefficients of quartic;c[0] const term*/
                    double* t_low, /*   low and high brackets*/
                    double* t_high )
{
double cubic_coef[4], roots[3];
double t_guess, t_step;  /* root-bracketing values */
double p_guess, dp_guess, p_low, p_high, dp_low, dp_high;
int i, iter, nreal, check;

/* to find brackets for roots, find max and min */

for (i = 1; i < 5; i++) cubic_coef[i-1] = (double)(i)*c[i];
check = QBSSolveCubic( cubic_coef, roots , &nreal);
if (check != QBS_OK) return check;
switch (nreal) {
   case 1:    /* only 1 max/min; root to quartic (if exists) between here*/
              /* and infinity (in both directions). Go towards + infnty*/
      *t_low = roots[0];
      poly(*t_low, 4, c, &p_low, &dp_low);
      if (fabs(p_low )< QUARTIC_EPS) {  /* only max/min also a root! DONE! */
          *t_high = *t_low;
          return  DEGENERATE_1_ROOT_AT_TANGENCY;
      }     
      if (p_low * c[4] > 0.0){
         return NO_REAL_ROOTS_1_MAX_MIN;
                      /* f(turn_pt) same sign as infnty; no root */
      }
      t_step = p_low/(12.0*c[4]*(*t_low)*(*t_low)+6.0*c[3]*(*t_low)+2.0*c[2]);
      t_step = sqrt ( fabs (t_step) );  /* use curvature to guess step */
      *t_high = *t_low + t_step;
      poly (*t_high, 4, c, &p_high, &dp_high);
      iter = 0;
      while ( (iter < N_MAX_ITER) && (p_low * p_high > 0.0) ){
              t_step *= 1.4; *t_high = *t_low + t_step;
              (*poly) (*t_high, 4, c, &p_high, &dp_high);
              iter++;
      }
      break;
   case 3:  /* there are max and min; see if root lies between them */
      *t_low = roots[0]; t_guess = roots[1]; *t_high = roots[2];
      poly(*t_low, 4, c, &p_low, &dp_low);
      poly(t_guess, 4, c, &p_guess, &dp_guess);
      poly(*t_high, 4, c, &p_high, &dp_high);
      if ( p_low * p_guess < 0.0 ) {
           *t_high = t_guess; p_high = p_guess; 
      } else if (p_guess * p_high < 0.0 ) {
           *t_low  = t_guess; p_low = p_guess; 
      } else {    /* no root between max and min. Must be -> infnty */
           if (p_high * c[4] > 0.0 ){
               return NO_REAL_ROOTS_3_MAX_MIN;  /* Nope. No real roots */
           }
           t_step = *t_high - *t_low;
           *t_low = *t_high; p_low = p_high;
           *t_high = *t_low + t_step;
           poly ( *t_high, 4, c, &p_high, &dp_high );
           iter = 0;
           while ( (iter < N_MAX_ITER) && (p_low * p_high > 0.0) ){
              t_step *= 1.4; *t_high = *t_low + t_step;
              poly(*t_high, 4, c, &p_high, &dp_high);
              iter++;
           }
      }
      break;
   default:  /* help! */
      return ERROR_FINDING_MAX_MIN_FOR_BRACKET;
      break;
}
/* Now t_low and t_high should bracket roots. Let's check for safety */
poly (*t_low, 4, c, &p_low, &dp_low);
poly ( *t_high, 4, c, &p_high, &dp_high );
if (p_low * p_high > 0.0 ) {   /* something wrong */
    return QUARTIC_BRACKETING_CONFUSED;
} else {
return  QBS_OK;
}

}


/********************************************************** poly ******/
/*
* Evaluate a polynomial of any degree and its derivative
*
* Numerical Recipes Sect 5.3 says everyone should know this!
*/

void poly( double  t,      /* value of argument */
              int    degree, /* highest pwr of t inpolynomial */
              double* c,      /* coefficients, c[0] constant term */
              double* f,      /* evaluated polynomial */
              double* df )    /* evaluated derivative */
{
*f  = c[degree];
*df = 0.0;
while (degree > 0){

    *df = *df*t + *f;
    *f  = *f * t + c[--degree];
}
return;
}


/****************************************************** rtsafe_poly *****/

/* Use safe Newton-Raphson algorithm, specialised for case of polynomial,
*  to find a (previously bracketed) root of a polynomial.
*
*  After algorithm rtsafe described in Numerical Recipes
*/

int rtsafe_poly( double* c,     /* polynomial coeffs. c[0] constant term */
                 int  deg,   /* highest power of x */
                 double  x1,    /* lower bound for bracketed root */
                 double  x2,    /* upper bound for bracked root */
                 double  xacc,  /* absolute accuracy for root */
                 double* root )       /* found root */
{
int j;
double df, dx, dxold, f, fh, fl;
double swap, temp, xh, xl;

poly(x1, deg, c, &fl, &df);
poly(x2, deg, c, &fh, &df);
if (fl*fh >= 0.0) return RTSAFE_INPUTS_DONT_BRACKET_ROOT;
if (fl < 0.0){    /* orient search s.t. f(xl < 0 */
   xl = x1;
   xh = x2;
} else {
   xh = x1;
   xl = x2;
   swap = fl;
   fl = fh;
   fh = swap;
}
*root = 0.5 * (x1 + x2);
dxold = fabs (x2 - x1);
dx    = dxold;
poly(*root, deg, c, &f, &df);
for (j=1; j< N_MAX_ITER; j++){
    if (((( *root-xh)*df - f)*((*root-xl)*df-f)>= 0.0)  /*Bisect if Newt */
        || (fabs(2.0*f) > fabs(dxold*df))) {  /* out or range or too slow */
        dxold = dx;
        dx    = 0.5 * (xh - xl);
        *root = xl + dx;
        if (xl == *root) return QBS_OK;  /* neglib. change; take root */
     } else {
        dxold = dx;
        dx    = f/df;             /* take Newton step */
        temp  = *root;
        *root -= dx;
        if (temp == *root) return QBS_OK;
     }
     if (fabs(dx) < xacc ) return QBS_OK;  /* normal convergence acceptance */
     poly(*root, deg, c, &f, &df);
     if (f < 0.0) {            /* maintain bracket on root */
        xl = *root;
        fl = f;
     } else {
        xh = *root;
        fh = f;
     }
}

return RTSAFE_POLY_N_MAX_ITER_EXCEEDED;
}

/************************************************* QBSSolveCubic *****/
/*
*   Use rtsafe_poly to find one root
*   Divide out and use quadratic formula on result
*/

int QBSSolveCubic ( double* c3,   /* coeffs of cubic; c[0] = constant term */
                    double* roots, /* returned real roots */
                    int*   nreal)/* no of real roots; complex roots NOT retnd*/
{
double x1, x2, f1, f2, df1, df2;
double step, rem, temp;
int i,iter, check;
double c2[3], discrim;

/* need to bracket root; start at arbitrary location; try 0.0 and
*  a value based on abs values of coeffs */

x1 = 0.0; 

temp = fabs( c3[2]/c3[3] );
x2 = temp*temp*temp ;
temp = fabs (c3[1]/c3[3]);
x2 += temp*sqrt(temp) + fabs ( c3[0]/c3[3] );
x2 = pow ( x2, 0.33333);
poly(x1, 3, c3, &f1, &df1);
poly(x2, 3, c3, &f2, &df2);
if (fabs(f2) > fabs(f1) ){        /* make sure we aren't near a root */
    x1 = x2; f1 = f2; df1 = df2;
}
step = fabs ( f1/df1 ); /* use derivative at x1 to guess step */
if ( (step < 10.0 * ABSOLUTE_ROOT_ACCURACY ) /* if not sensible, set to 1.0 */
     || (step > 1.0/ ABSOLUTE_ROOT_ACCURACY) ) step = 1.0;

if ( f1 * c3[3] > 0.0 ) step *= -1.0;  /* if x1 and x->+infnty same side */
                                       /* root lies toward -infnty */
x2 = x1 + step;
poly(x2, 3, c3, &f2, &df2);
iter = 0;
while ( (iter < N_MAX_ITER) && (f1 * f2 )>= 0.0 ){
   step *= 1.4;
   x2 = x1+step;
   poly(x2, 3, c3, &f2, &df2);
   iter++;
}
if ( f1*f2 >= 0.0 ) return CUBIC_BRACKET_FAILED;
check = rtsafe_poly (c3, 3, x1, x2, ABSOLUTE_ROOT_ACCURACY, &roots[0]);
if (check != QBS_OK) return check;

/* divide out found root to leave quadratic */

rem = c3[3];
for (i = 2; i>=0; i--) {
   c2[i] = rem;
   rem = c3[i] + rem*roots[0];
}
discrim = ( c2[1]*c2[1] - 4.0* c2[2]*c2[0] );
if ( discrim < 0.0 ){   /* other roots complex */
   *nreal = 1;
   return QBS_OK;
} else{
   discrim = sqrt ( discrim );
   roots[1] = ( -c2[1] + discrim )/ (2.0 * c2[2]);
   roots[2] = ( -c2[1] - discrim )/ (2.0 * c2[2]);
   *nreal = 3;
   qsort( roots, 3, sizeof(double), QBSDoubleCompare); /* sort roots */
   return QBS_OK;
}

}

/************************************************ QBSFloatCompare ***********/
/*
* Compare function for qsort. Returns:
*
* -1 if el1 < el2
* 0  if el1 == el2
* +1 if el1 > el2
*
*/

int QBSDoubleCompare (const void* el1, const void* el2){

double f1, f2;

f1 = * ((double*) el1);
f2 = * ((double*) el2);

if (f1 < f2) return -1;
if (f1 == f2) return 0;
return 1;
}

/********************************************* QBSDistanceAlongVec ****/

/*
*  Aberrate Position and Vec
*  Scale L, ro
*  Displace Pos
*  Calc Distance along vec to intersection with shock, or to pt of
*        closest approach if no intersection - flag passed as return arg
*  Undisplace and unaberrate
*/

int QBSDistanceAlongVec(float* r_vec,   /* Position in Re_GSE */
                           float  vsw,     /* Vsw (x) in km/s */
                           QBSMP_models model, /* model to use */
                           float  sigma,    /* mult scaling to apply to model*/
                           float* vec,      /* vector direction to follow */
                           float* distance, /*returned signed dist along vec */
                           float* r_at_shock)/* returned GSE pos at shock */
{

float rab_vec[3], rabd_vec[3],ro[3],L, eps;
float rabd_atshock[3], rab_atshock[3];
float vecab[3];
float minus_ro[3];
int i,check;

check = QBSCheckModelNumber(model);
if (check != QBS_OK) return check;   /* check model number */

QBSAberrate(r_vec, vsw, rab_vec);  /* pos vector in aberrated coords */
QBSAberrate(vec,   vsw, vecab  );  /* aberrate direction vector      */

eps   = BSMP_eps[model];         /* get model parameters */
L     = sigma * BSMP_L[model];   /* scale distances */
ro[0] = sigma * BSMP_xo[model];
ro[1] = sigma * BSMP_yo[model];
ro[2] = 0.0;

QBSDisplace (rab_vec, ro, rabd_vec);  /* displace position according to model*/

check = QBSCalcDistAlongVec(rabd_vec, eps, L, vecab, distance, rabd_atshock); 
                           /* calc distance and shock intersect with normal */

for(i=0; i<3; i++) minus_ro[i] = - ro[i];   /* un-displace r_at_shock */
QBSDisplace (rabd_atshock, minus_ro, rab_atshock);

QBSUnAberrate ( rab_atshock, vsw, r_at_shock ); /* and unaberrate it */

return check;  /* nb returns flag if no intersection. dist & r_at shock in */
               /* this case corresponds to nearest approach - para to tangent*/

}


/********************************************** QBSCalcDistAlongVec *****/
/*
*   Take position and model (pre-aberrated and displaced)
*   Normalise vec
*   Solve quadratic for roots
*   Choose "correct" root; if no intersect calc nearest approach and
*      set return flag
*   Calculate distance, pos at shock/nearest (same frame as input) and return
*/

int QBSCalcDistAlongVec(float* R,          /* position (pre-aber and displ)*/
                       float eps,         /* model eps and L - pre-scaled */
                       float L,
                       float* vec, /* vector direction to follow (pre-aber)*/ 
                       float* distance, /*signed distance along vector in Re */
                       float* r_at_shock) /* intersect n,shock; in frame of R*/

{

float epssq;
float A, B, C;
float roots[2], discrim;       
float vec_norm[3], vec_mag;
int   check,i;

vec_mag = QBSVectorMagnitude(vec);
for (i=0; i<3; i++) vec_norm[i] = vec[i]/vec_mag;  /* normalise vector */

epssq = eps*eps;

A = 1.0 - epssq * vec_norm[0]*vec_norm[0];
B =  -QBSVectorScalarProduct (R, vec_norm) - eps * L * vec_norm[0];
  B = 2.0 * ( B + epssq * R[0] * vec_norm[0]);
C =  (L - eps*R[0]);
  C = C*C; 
  C = QBSVectorScalarProduct(R,R) - C;

discrim = B*B - 4.0*A*C;
if (discrim < 0.0){   /* no intersection; set flags, etc. */
    *distance = -B/(2.0 * A);
    check = NO_INTERSECTION_WITH_SURFACE; /* proceed; find closest dist */
} else {
    discrim = sqrt(discrim);
    roots[0] = (-B + discrim)/(2.0 * A);
    roots[1] = (-B - discrim)/(2.0 * A);
    check = QBSChooseAlongVecRoot(R, eps, L, vec_norm, roots, distance);
} /* above returns QBS_OK if found else flag if no root on right branch */

for (i = 0; i<3; i++){
    r_at_shock[i] = R[i] - *distance * vec_norm[i];
}
return check;
}

/*********************************************QBSChooseAlongVecRoot****/
/* Given roots t to vector eqn R = r + t b_hat
*
*  Reject roots on wrong hyperbola branch
*  Of remaining, take smallest abs. value
*  Set return flag
*/


int QBSChooseAlongVecRoot(float* R,   /* pre-aber, displaced position */
                          float  eps,  /* model, scaled L */
                          float  L,
                          float* vec_norm,  /* unit direction vector */ 
                          float* roots, /* roots to quadratic for t */
                          float* chosen_root) /* returned selected root */
{
float x1, x2, r1, r2;
int check;

x1 = R[0] - roots[0]*vec_norm[0];
x2 = R[0] - roots[1]*vec_norm[0];

r1 = L - eps*x1;
r2 = L - eps*x2;

if ( ( r1 < 0.0 ) && ( r2 < 0.0 ) ) {
     *chosen_root =  (fabs(roots[0]) < fabs(roots[1]) ) ? roots[0] : roots[1];
     check = BOTH_INTERSECTIONS_ON_WRONG_BRANCH;
                           /* return smallest dist, though it is nonsense */
} else if ( r1*r2 < 0.0 ){                          /* only one good root */
     *chosen_root =  (r1 > 0.0) ?  roots[0] : roots[1];
     check = QBS_OK;
} else {                 /* two possible roots; take smallest magnitude */
     *chosen_root =  (fabs(roots[0]) < fabs(roots[1]) ) ? roots[0] : roots[1];
     check = QBS_OK;
}
return check;
}
 
/**************************************** QBSListModels **************/

/* Send a list of BS and MP models to an i/o stream */

void QBSListModels(FILE* fp)
{
int i;

fprintf(fp,"Model                     No.   eps    L      xo     yo  \n");
fprintf(fp,"========================================================\n");

for(i=0; i<N_MODELS; i++){
   fprintf(fp,"%-25s %2d %6.2f %6.2f %6.2f %6.2f\n", 
       BSMP_name[i], i, BSMP_eps[i],BSMP_L[i], BSMP_xo[i], BSMP_yo[i]);
}
return;
}

/***************************************** QBSCheckModelNumber ********/

/* ensure model number is in range 0 thru N_MODELS-1) */

int QBSCheckModelNumber(int n_model){

if ( (n_model < 0) || ( n_model > (N_MODELS -1) ) ){
    return MODEL_NUMBER_OUT_OF_RANGE;
} else {
    return QBS_OK;
}
}

/******************************************** QBSErrorMessages ******/
/* Take error code and print string message to stream */

void QBSErrorMessage(FILE *fp, int  qbserrno) /* i/o stream, e.g., stdout, error code number */
{
fprintf(fp,"QBS return code %d:",qbserrno);
switch (qbserrno){
      case QBS_OK:
         fprintf(fp,"Successful return");
         break;
      case DISCRIM_NEGATIVE_IN_SCALE:
         fprintf(fp,"Negative discriminant in scaling routine (?)");
         break;
      case LARGEST_SCALING_ROOT_NOT_POSITIVE:
         fprintf(fp,"Scaling roots not positive (?)");
         break;
      case NO_REAL_ROOTS_FOR_CLOSEST_DIST_ALONG_NORMAL:
         fprintf(fp,"No real roots for closest distance along normal. Unreachable?");
         break;
      case TOO_MANY_REAL_ROOTS_FOR_DIST_ALONG_NORMAL: 
         fprintf(fp,"Root count for distance along normal > 4 (?)");
         break;
      case NO_POSSIBLE_ROOTS_FOR_DIST_ALONG_NORMAL:
         fprintf(fp,"No possible roots for distance along normal; all on wrong branch of hyperbola (?)");
         break;
      case DEGENERATE_1_ROOT_AT_TANGENCY:
         fprintf(fp,"Quartic for distance along normal has no crossing, but root at tangency found and returned.");
         break;
      case NO_REAL_ROOTS_1_MAX_MIN:
         fprintf(fp,"Quartic for distance along normal has 1 turning pt and no real roots.");
         break;
      case NO_REAL_ROOTS_3_MAX_MIN:
         fprintf(fp,"Quartic for distance along normal has 3 turning pts and no real roots.");
         break;
      case ERROR_FINDING_MAX_MIN_FOR_BRACKET:
         fprintf(fp,"Failed to converge to brackets for root of quartic for distance along normal.");
         break;
      case QUARTIC_BRACKETING_CONFUSED:
         fprintf(fp,"Routine to bracket root of quartic got confused.");
         break;
      case RTSAFE_INPUTS_DONT_BRACKET_ROOT:
         fprintf(fp,"Polynomial rtsafe routine passed starting values which don't bracket a root.");
         break;
      case RTSAFE_POLY_N_MAX_ITER_EXCEEDED:
         fprintf(fp,"Maximum number of iterations (%d) exceeded in polynomial rtsafe routine", N_MAX_ITER);
         break;
      case CUBIC_BRACKET_FAILED:
         fprintf(fp,"Failed to bracket root for cubic equation.");
         break;
      case NO_INTERSECTION_WITH_SURFACE:
         fprintf(fp,"Tracing along vector doesn't intersect shock; returned value is real part of root which corresponds to nearest approach.");
         break;
      case BOTH_INTERSECTIONS_ON_WRONG_BRANCH:
         fprintf(fp,"Tracing along vector intersects only the wrong branch of the hyperboloid; smallest distance is returned, but not physical.");
         break;
      case MODEL_NUMBER_OUT_OF_RANGE:
         fprintf(fp,"Model number outside range of known models");
         break;
      default:
         fprintf(fp,"Unknown error return code");
         break;
}
fprintf(fp,"\n");
return;
}

