// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Tensor diagonalization
// 
// Used test various methods to calculate the pressures (P/T_perp, P.T_par)
// from temperature or pressure tensors. 
//
// Inputs :
// 	bfield		- vector time series with magnetic field
//      density         - plasma density (may be needed for P=nkT
// 	temp	        - time series with plasma temp as vector or tensor
//      time_interval   - time interval for MVAB (used to establish n)
//      method          - 0,1,2 to select calculation/diagonalization method
//
// Outputs
//      P_p             - full plasma pressure [nkT_perp, nkT_perp, nkT_par] 
//      T               - full plasma temperature [T_perp, Tperp, T_par]  
//      Alpha           - anisotropy (P||-Pperp)/(B^2/u0)
//      Angular         - angle between B and paralell direction [deg]
//
//
// SEH, 08 Jan, 2014 - based on PlasmaParams
//      20 Jan, 2014 - added diagnostics output (angles, eigvecs)
//      05 Mar, 2014 - added sqrt(1-alpha) and beta + use joined |B| for these
//      17 Mar, 2014 - added stress calculation
//      10 Oct, 2016 - based on THEMIS pressure calculator
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "TensorDiag.h"
#include "DvObj.h"


DvObject_var Projection(DvObject_var);
DvObject_var DiagonalizeSort(DvObject_var,DvObject_var);
DvObject_var DiagonalizeProject(DvObject_var,DvObject_var);
DvObject_var DiagonalizeSonnerup(DvObject_var,DvObject_var);

DvObject_var angleBetween(DvObject_var, DvObject_var);
DvObject_var projectPtens(DvObject_var,DvObject_var);

extern "C" QplugReturnStatus  DvTensorDiag  (vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;  

QplugReturnStatus  DvTensorDiag  (vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)

{

  bool    Bsup=false;
  long    nestsize;
  char    txtbuf[200];
  DvObject_var  tt_out;                                  // tt of masked V and B


  // --- establish what inputs are -------


  DvObject_var P_in = inputs[0];
  DvEvent start_end = P_in->getTimeRange();
  if (P_in.is_ok()) {
    QplugAppendTextDisplay("Pressure tensor supplied..\n");
  } else {
    QplugAppendTextDisplay("No pressure tensor supplied\n");
    return(QPLUG_FAILURE);
  }
  P_in=P_in->subSequence(start_end); 
  tt_out=P_in->getTimeTags();


  DvObject_var B_in =  inputs[1];

  bool bok = B_in->isVectorXYZ();
  bool bol = B_in->isThreeVector();

  printf("B is %d %d\n",bok,bol);

  if (B_in.is_ok() & B_in->isVectorXYZ()) {
      Bsup=true;
      QplugAppendTextDisplay("B-field supplied\n");
      B_in = B_in->Join(tt_out);
  } else {
    QplugAppendTextDisplay("No B-field or vector error(MMS 4comp?) \n");
    return(QPLUG_FAILURE);
  }
  nestsize = P_in->seqSize();

 

  // ---- get method 

  int method = -1;
  DvObject_var in_method = inputs[2];
  if (in_method->totalSize() == 1 ) {
    method = in_method->asInt();
  } else {
    QplugAppendTextDisplay("Failure : No calculation method given");
    return(QPLUG_FAILURE);
  }
  if (method > 2) method = 0;


  // ---- declare and calculate the output objects -----

  DvObject_var P_out = new DvObject();
  DvObject_var A_out = new DvObject(0.0,nestsize);
      

  switch(method) {
  case 0: 
    if (!Bsup) {
      QplugAppendTextDisplay("B-field not given (required for diagnostics)\n");
      return(QPLUG_FAILURE);
    }
    P_out = DiagonalizeSort(P_in, B_in);
    for (int i=0; i<7; i++) {
      std::string str = GetSlotText(i);
      str+="_Diag";
      SetSlotText(i,str.c_str());
    }
    break;
  case 1:
    if (!Bsup) {
      QplugAppendTextDisplay("Projection along B impossible: B-field not given\n");
      return(QPLUG_FAILURE);
    }
    P_out = DiagonalizeProject(P_in, B_in);
    for (int i=0; i<7; i++) {
      std::string str = GetSlotText(i);
      str+="_Mag";
      SetSlotText(i,str.c_str());
    }
    break;
  case 2:
    if (!Bsup) {
      QplugAppendTextDisplay("x3 method impossible: B-field not given\n");
      return(QPLUG_FAILURE);
    }
    P_out = DiagonalizeSonnerup(P_in, B_in);
    for (int i=0; i<7; i++) {
      std::string str = GetSlotText(i);
      str+="_xB";
      SetSlotText(i,str.c_str());
    }
    break;
  }


  P_out->copy_xrefs_from(P_in);
  P_out->change_xref(DEPEND_0, tt_out);
  outputs.push_back(P_out);


  float e = 1.60217649e-19;      //  elementary charge C (Vs)
  float Mu_0 = 1.25664E-06; 
  DvObject_var alpha = new DvObject(0.0, nestsize);
  DvObject_var beta = new DvObject(0.0, nestsize);
  DvObject_var traceP = P_in->trace();
  DvObject_var B2 = B_in->abs()->power(2.0);

  for (int i=0; i<nestsize; i++) {
    double p_par = P_out->dbl(i,2);  
    double p_perp = 0.5*(P_out->dbl(i,0) + P_out->dbl(i,1));
    double p_tot = traceP->dbl(i)/3;
    alpha->dbl(i) = 1e9 * Mu_0 * (p_par - p_perp) /  B2->dbl(i);
    beta->dbl(i) =  2e9 * Mu_0 * p_tot / B2->dbl(i);
  }


  if (alpha.is_ok()){
    alpha->change_xref(DEPEND_0, tt_out);
    outputs.push_back(alpha);
  } else {
    DvObject_var empty = new DvObject();
    outputs.push_back(empty);
  }

  if (beta.is_ok()){
    beta->change_xref(DEPEND_0, tt_out);
    outputs.push_back(beta);
  } else {
    DvObject_var empty = new DvObject();
    outputs.push_back(empty);
  }

  return QPLUG_SUCCESS;
  
}


// -----------------------------------------------------------------------------------------
//
// Diagonalizes the pressure tensor, sort, check similarity
// and returns a reduced pressure vectortime series with [P_perp, P_perp, P_par]
//
// Input:
//      Pts = time series with 3x3 pressure tensor
//      Bts = time series with B-field vector
//
// SEH, Jan 10, 2014
// 
// -----------------------------------------------------------------------------------------
DvObject_var DiagonalizeSort(DvObject_var Pts,  DvObject_var Bts) 
{
  printf("0 - Diagionalize and sort...\n");
  QplugAppendTextDisplay("Method 0 : diagnalize and sort..");
  int nestsize = Pts->seqSize();
  DvObject_var P_out = new DvObject(0.0,DIM3,nestsize);
  DvObject_var P =  new DvObject(0.0,DIM33);     // Pressure tensor
  DvObject_var Q =  new DvObject(0.0,DIM33);     // Q = projection matrix for constrain
  DvObject_var EigenVec = new DvObject(0.0, DIM33);  
  DvObject_var EigenValues = new DvObject(0.0, DIM3);
  DvObject_var B = new DvObject(0.0, DIM3);      // Unit vector along B
  DvObject_var A = new DvObject(0.0, DIM3);      // Angle btw B and the 3 Eigvecs [deg]
  char    txtbuf[255];

  for (int j=0;j<3; j++) Q->dbl(0,j,j) = 1.0;

  for (int i=0; i<nestsize; i++) {
    for (int j=0;j<3; j++){
      B->dbl(0,j) = Bts->dbl(i,j);
      for (int k=0;k<3; k++) {
  	P->dbl(0,j,k)= Pts->dbl(i,j,k);
      }
    }

    ThemisDiagonalize diag(0,P,Q);  
    EigenVec = diag.eigenvectors();
    EigenValues = diag.eigenvalues();
   
    double babs  = B->abs()->dbl(0);
    for (int j=0;j<3; j++) {
      P_out->dbl(i,2-j) = EigenValues->dbl(0,j);
      B->dbl(0,j)/=babs;
    }   
//     A = angleBetween(EigenVec,B);
//     for (int k=0; k<3; k++) {
//       A_out->dbl(i,2-k) = A->dbl(0,k);
//       for (int l=0;l<3;l++) Evecs->dbl(i,2-k,l) = EigenVec->dbl(0,k,l);
//     }

    double d21 = EigenValues->dbl(0,2) - EigenValues->dbl(0,1);
    double d10 = EigenValues->dbl(0,1) - EigenValues->dbl(0,0);

    // -- any need to swap ?
    if (d21 > d10) {
      P_out->dbl(i,2) = EigenValues->dbl(0,2); //Ppar
      P_out->dbl(i,1) = EigenValues->dbl(0,0); //Pperp
      P_out->dbl(i,0) = EigenValues->dbl(0,1); //Pperp
//       for (int l=0;l<3;l++) {
// 	Evecs->dbl(0,i,1,l) = EigenVec->dbl(0,0,l);
// 	Evecs->dbl(0,i,0,l) = EigenVec->dbl(0,1,l);
//       }

//       printf("setting A_out\n");
//       A_out->dbl(0,i,2) = A->dbl(0,2);
//       A_out->dbl(0,i,1) = A->dbl(0,0);
//       A_out->dbl(0,i,0) = A->dbl(0,1);
    } 
  
  }
  return(P_out);
};


// -----------------------------------------------------------------------------------------
//
// Diagonalizes the pressure tensor, projects it along B 
// and returns a reduced pressure vectortime series with [P_perp, P_perp, P_par]
//
// SEH, Jan 10, 2014
// 
// -----------------------------------------------------------------------------------------
DvObject_var DiagonalizeProject(DvObject_var Pts, DvObject_var Bts)
{
  printf("2 - Mag...\n");
  QplugAppendTextDisplay("Method 1 : Mag : Project along B as in Sonnerup document..");

  int nestsize = Pts->seqSize();
  DvObject_var P_out = new DvObject(0.0,DIM3,nestsize);
  DvObject_var P =  new DvObject(0.0,DIM33);     // Pressure tensor
  DvObject_var Q =  new DvObject(0.0,DIM33);     // Q = projection matrix for constrain
  DvObject_var EigenVec = new DvObject(0.0, DIM33);  
  DvObject_var EigenValues = new DvObject(0.0, DIM3);
  DvObject_var B = new DvObject(0.0, DIM3);      // Unit vector along B
  DvObject_var A = new DvObject(0.0, DIM3);      // Angle btw B and the 3 Eigvecs [deg]
  char    txtbuf[255];


  for (int j=0;j<3; j++) Q->dbl(0,j,j) = 1.0;

  for (int i=0; i<nestsize; i++) {
    for (int j=0;j<3; j++){
      B->dbl(0,j) = Bts->dbl(i,j);
      for (int k=0;k<3; k++) {
  	P->dbl(0,j,k)= Pts->dbl(i,j,k);
      }
    }

    double babs  = B->abs()->dbl(0);
    // double trace =0.0;
    for (int j=0;j<3; j++) {
      B->dbl(0,j)/=babs;
      //trace += P->dbl(0,j,j);
    }
    DvObject_var bp_temp = B * P;
    DvObject_var ppar = bp_temp * B;

    Q = Projection(B);

    ThemisDiagonalize diag(1,P,Q);  
    EigenVec = diag.eigenvectors();
    EigenValues = diag.eigenvalues();
    
    for (int j=0;j<3; j++){
      P_out->dbl(i,j) = EigenValues->dbl(0,j);
      //      for (int l=0;l<3;l++) Evecs->dbl(0,i,j,l) = EigenVec->dbl(0,j,l);
    }
    P_out->dbl(i,2) = ppar->dbl(0);
//     A = angleBetween(EigenVec,B);
//     for (int k=0; k<3; k++) A_out->dbl(0,i,k) = A->dbl(0,k);
  }
  return(P_out);
};



// -----------------------------------------------------------------------------------------
//
// Diagonalizes the pressure tensor, use Sonnerup method to find directions
// and returns a reduced pressure vectortime series with [P_perp, P_perp, P_par]
//
// SEH, Jan 10, 2014
// 
// -----------------------------------------------------------------------------------------
DvObject_var DiagonalizeSonnerup(DvObject_var Pts, DvObject_var Bts) 
{
  printf("0 - Diagionalize and sort...\n");
  QplugAppendTextDisplay("Method 2 : Mag : Project to minimize angle along B for ppar..");
  int nestsize = Pts->seqSize();
  DvObject_var P_out = new DvObject(0.0,DIM3,nestsize);
  DvObject_var P =  new DvObject(0.0,DIM33);     // Pressure tensor
  DvObject_var Q =  new DvObject(0.0,DIM33);     // Q = projection matrix for constrain
  DvObject_var EigenVec = new DvObject(0.0, DIM33);  
  DvObject_var EigenValues = new DvObject(0.0, DIM3);
  DvObject_var B = new DvObject(0.0, DIM3);      // Unit vector along B
  DvObject_var A = new DvObject(0.0, DIM3);      // Angle btw B and the 3 Eigvecs [deg]
  char    txtbuf[255];

 

  for (int j=0;j<3; j++) Q->dbl(0,j,j)=1.0;


  double babs;
  int indices[]={0,1,2,0,1};

  for (int i=0; i<nestsize; i++) {

    // --- extract Bfield and construct 3x3 tensor ---
    for (int j=0;j<3; j++){
      B->dbl(0,j) = Bts->dbl(i,j);
      for (int k=0;k<3; k++) {
  	P->dbl(0,j,k) = Pts->dbl(i,j,k);
      }
    }
    // --- diagonalize: find eigenvectors and eigenvalues of tensor ---
    ThemisDiagonalize diag(0,P,Q);  
    EigenVec = diag.eigenvectors();
    EigenValues = diag.eigenvalues();
   
    double babs  = B->abs()->dbl(0,0);
    for (int j=0;j<3; j++) {
      B->dbl(0,j)/=babs;
      //      for (int l=0;l<3;l++) Evecs->dbl(0,i,j,l) = EigenVec->dbl(0,j,l);
    }   


    A = angleBetween(EigenVec,B);

   // -- now, find the index of the element with smallest angle == tpar;
    int index=0;
    for (int l=1;l<3; l++) {
      if (A->dbl(0,index) > A->dbl(0,l)) index=l;
    }

    P_out->dbl(i,2) = EigenValues->dbl(0,index); //Ppar
    P_out->dbl(i,1) = EigenValues->dbl(0,indices[index+1]); //Pperp
    P_out->dbl(i,0) = EigenValues->dbl(0,indices[index+2]); //Pperp

//     A_out->dbl(0,i,2) = A->dbl(0,index);
//     A_out->dbl(0,i,1) = A->dbl(0,indices[index+1]);
//     A_out->dbl(0,i,0) = A->dbl(0,indices[index+2]);

  
//     printf("i=%3d A = %10.3f %10.3f %10.3f, index=%d\n",i, A->dbl(0,0),A->dbl(0,1),A->dbl(0,2),index);
  }
    
  return(P_out);

};



// -----------------------------------------------------------------------------------------
//
// Calculates the angles between a unit vector and each of 3 eigenvectors;
//
// Inputs :
//     B = unit magnetic field vector  (1 x 3)  
//     E = eigenvectors (3 x 3)
//
// SEH, Jan 10, 2014
// 
// -----------------------------------------------------------------------------------------
DvObject_var angleBetween(DvObject_var evec, DvObject_var B)
{
  DvObject_var ang = new DvObject(0.0, DIM3);      // angle between B and evecs

  for (int j=0; j<3; j++) {
    double dotprod=0.0;
    for(int i=0; i<3;i++)  dotprod+=B->dbl(0,i) * evec->dbl(0,j,i);
    ang->dbl(0,j) = 57.29*acos(abs(dotprod));	      
  }
  return(ang);
}


// -----------------------------------------------------------------------------------------
//
// Calculate Projection matrix dij -bibj
//
// Inputs :
//     B = unit magnetic field vector  (1 x 3)  for Bn=0 constrain
//
// SEH, Jan 10, 2014
// 
// -----------------------------------------------------------------------------------------
DvObject_var Projection(DvObject_var A) 
{
  DvObject_var C  = new DvObject(0.0, DIM33);
 
  DvObject_var A2 = A->dot(A);
  double sz2 = A2->dbl();
  
  C->dbl(0,0,0) = 1.0 - (A->dbl(0)*A->dbl(0)) / sz2;
  C->dbl(0,0,1) = - (A->dbl(0)*A->dbl(1)) / sz2;
  C->dbl(0,0,2) = - (A->dbl(0)*A->dbl(2)) / sz2;

  C->dbl(0,1,0) = - (A->dbl(1)*A->dbl(0)) / sz2;
  C->dbl(0,1,1) = 1.0 - (A->dbl(1)*A->dbl(1)) / sz2;
  C->dbl(0,1,2) = - (A->dbl(1)*A->dbl(2)) / sz2;
  
  C->dbl(0,2,0) = - (A->dbl(2)*A->dbl(0)) / sz2;
  C->dbl(0,2,1) = - (A->dbl(2)*A->dbl(1)) / sz2;
  C->dbl(0,2,2) = 1.0 - (A->dbl(2)*A->dbl(2)) / sz2;
  
  return(C);
}


