/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	tsyganenko.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS tsyganenko wrapper
*
* PURPOSE:      Tsyganenko Earth magnetic field calculation.
* 
* DESCRIPTION:  Plug In module checking input parameters and calling c-code.
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*
*   OUT:
*         float tsyganenko_n       vector time series of function values		
*
* GLOBALS: see geopack.c
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: QrGetFrameFromAttr,QrIsPositionVector,SIConv_factor (appended)
*                 from Tsyganenko's GEOPACK.FOR and T96.FOR packages:
*                 external C calcps_,t96_01__,igrf_,recalc_,geogsm_,bspcar_
*                 The external functions are translated by f2c, thus libf2c.a
*                 must be loaded.
*
* HISTORY: created 2001 Markus Fraenz 
*          UNITS checking repaired wjmf 12/07/01
*          rotation into GSM included wjmf 12/07/01
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"

#include "qdutil.h"
#include "DvObj.h"
#include "qrtn.h"

extern "C"  int calcps_(long int *iyr, long int *iday, long int *ihour, 
	long int *min__, long int *isec, float *ps);
extern "C"  int t96_01__(long int *iopt, float *parmod, float *ps, float *x, 
	float *y, float *z, float *bx, float *by, float *bz);
extern "C" int recalc_(long int *iyr,long  int *iday,long  int *ihour,
        long  int *min__,long  int *isec);
extern "C" int igrf_(long int *iy, long int *nm, float *r__, float *t, float 
	*f, float *br, float *bt, float *bf);
extern "C" int geogsm_(float *xgeo, float *ygeo, float *zgeo, float *xgsm, 
	float *ygsm, float *zgsm, long int *j);
extern "C" int bspcar_(float *teta, float *phi, float *br, float *btet, 
	float *bphi, float *bx, float *by, float *bz);


using namespace QSAS;  

extern "C" QplugReturnStatus  DvTsyganenko( vector<DvObject_var>& inputs, vector<DvObject_var>& outputs )
{

// define some C variables

long   check;
size_t Ndata;
int   MJDdate, MJD0;
int   year, month, day, hour, minute;
double second;
long  iopt=0,iyr,iday,ihour,imin,isec,jswitch;
long int nIGRF = 10;
bool  extset = true,rotateset = false;

float parmod[10],ps=0.0,GSMx,GSMy,GSMz,bx=0.0,by=0.0,bz=0.0;
float pdyn,dst,byimf,bzimf,xgsm,ygsm,zgsm,bxgeo=0.0,bygeo=0.0,bzgeo=0.0;
float xgeo=0.0,ygeo=0.0,zgeo=0.0,rgeo=0.0,tgeo=0.0,pgeo=0.0,brgeo=0.0,btgeo=0.0,bpgeo=0.0;
char	str_out[150];


RtnVector vec;

DvObject_var  pdyn_obj;
DvObject_var  dst_obj;
DvObject_var  byimf_obj;
DvObject_var  bzimf_obj;
DvObject_var  nIGRF_obj;
DvObject_var  ts_in;
DvObject_var  tsyganenko_out;

QplugAppendTextDisplay ("Tsyganenko Model Plugin V1.1 M.Fraenz 31/07/2001.\n");

// ...first item is input Time Series
ts_in = inputs[0];

//  check QSAS datatype

if( ts_in.is_nil() || ts_in->arraySize() != 3 ) {
    QplugAppendTextDisplay ("tsyganenko>> Input not Vector Sequence.\n");
    return QPLUG_FAILURE;
}

// Find original frame
 
vec.in_frame = vec.QrGetFrameFromAttr(ts_in);

if( vec.in_frame == UNKNOWN ) {
    QplugAppendTextDisplay ("tsyganenko>>WARNING: Input Vector Frame not defined, assuming GSM.\n");
    vec.in_frame = GSM;
}
else if( vec.in_frame != GSM ) {
    rotateset = true;
}

// ensure input is position in Re

vec.IsPositionVector = vec.QrIsPositionVector(ts_in);
if(vec.IsPositionVector != 1){
   QplugAppendTextDisplay ("tsyganenko>> Input Vector not position vector.\n");
   return QPLUG_FAILURE;
}

sprintf(str_out, "tsyganenko>> Converting Input Vector to Re \n");
QplugAppendTextDisplay (str_out);

ts_in = ts_in->changeUnitsTo("6.37104e6>m", "Re");

// get timetags
vec.t = ts_in->getTimeTags();
if( vec.t.is_nil() ) {
    QplugAppendTextDisplay ("Cannot find time tags, some frames are time dependent\n");
    return QPLUG_FAILURE;
}

Ndata = ts_in->seqSize();

if (Ndata < 1) {
	QplugAppendTextDisplay ("tsyganenko>> ERROR - Series must have at least 1 data point.\n");
	return QPLUG_FAILURE; 
}

//	 option pdyn

pdyn_obj = inputs[1];

if ( pdyn_obj.is_nil() ) { 
	QplugAppendTextDisplay ("tsyganenko>> ERROR - pdyn not scalar.\n");
	return QPLUG_FAILURE; }

pdyn = pdyn_obj->asDouble();

//	 option dst

dst_obj = inputs[2];

if ( dst_obj.is_nil() ) { 
	QplugAppendTextDisplay ("tsyganenko>> ERROR - DST not scalar.\n");
	return QPLUG_FAILURE; 
}

dst = dst_obj->asDouble();

//	 option byimf

byimf_obj = inputs[3];

if ( byimf_obj.is_nil() ) { 
	QplugAppendTextDisplay ("tsyganenko>> ERROR - byimf not scalar.\n");
	return QPLUG_FAILURE; 
}

byimf = byimf_obj->asDouble();

//	 option bzimf

bzimf_obj = inputs[4];

if ( bzimf_obj.is_nil() ) { 
	QplugAppendTextDisplay ("tsyganenko>> ERROR - bzimf not scalar.\n");
	return QPLUG_FAILURE; }

bzimf = bzimf_obj->asDouble();

//	 option nIGRF

nIGRF_obj = inputs[5];

if ( nIGRF_obj.is_nil() ) { 
	QplugAppendTextDisplay ("tsyganenko>> ERROR - nIGRF undefined.\n");
	return QPLUG_FAILURE; 
}

nIGRF = (long)nIGRF_obj->asInt();

if ( nIGRF < 0 ) { 
	QplugAppendTextDisplay ("tsyganenko>> No external field added.\n");
	extset=false;
	nIGRF = -nIGRF;
}
if ( nIGRF > 10 ) { 
	QplugAppendTextDisplay ("tsyganenko>>  WARNING - nIGRF max=10.\n");
	nIGRF =10;
}
if ( nIGRF == 0 ) { 
	QplugAppendTextDisplay ("tsyganenko>>  No internal field added.\n");
}
else {
    sprintf(str_out, "tsyganenko>> Calculating internal field order:%ld.\n", nIGRF);
    QplugAppendTextDisplay (str_out);
}

// now rotate into GSM if necessary
if (rotateset) {
  QplugAppendTextDisplay ("tsyganenko>>  Rotating into GSM.\n");
  vec.n_recs = (int) Ndata;
  vec.ptr = ts_in;
  vec.out_frame = GSM; 
  vec.IsPositionVector = false;  // this is to avoid rejection
  check = vec.QrDoRotation();
  printf("Frame_in:%i\n",vec.in_frame);
  printf("Frame_out:%i\n",vec.out_frame);
  fflush(stdout);

  if (check > QRTN_OK) {
    QplugAppendTextDisplay ("tsyganenko>> ERROR rotating into GSM:\n ");
    QplugAppendTextDisplay (vec.QrErrorMessage(check));
    return QPLUG_FAILURE; 
  }	  
}

  //	..create the output data series objects


  sprintf(str_out, "tsyganenko>> Calculating t96 function for %ld data points.\n", Ndata);
  QplugAppendTextDisplay (str_out);

  tsyganenko_out = new DvObject(0.0, DIM3, Ndata);
  
  iopt=0;
  for (int i=0; i < 10; i++) parmod[i]=1.0;
  parmod[0] = pdyn;
  parmod[1] = dst;
  parmod[2] = byimf;
  parmod[3] = bzimf;

for(size_t i=0; i<Ndata; i++)  {

    vec.t->time(i).breakDownMJD(&year, &month, &day, &hour, &minute, &second);
    MJD0 = vec.QrMJD(year, 1, 1);
    MJDdate = vec.QrMJD(year, month, day);
    iday  = (long) (MJDdate - MJD0);
    iyr   = (long) year;
    ihour = (long) hour;
    imin  = (long) minute;
    isec  = (long) second;  

    xgsm = (float) vec.ptr->dbl(i,0);
    ygsm = (float) vec.ptr->dbl(i,1);
    zgsm = (float) vec.ptr->dbl(i,2);

    bx=0.0;
    by=0.0;
    bz=0.0;
    
    if (extset) {
        // first calculate dipole angle:
        calcps_(&iyr,&iday,&ihour,&imin,&isec,&ps);
        // then model Bxyz:
        t96_01__(&iopt,&parmod[0],&ps,&xgsm,&ygsm,&zgsm,&bx,&by,&bz);
    }
    
    // now the internal field:
    if (nIGRF > 0) {
    
        recalc_(&iyr,&iday,&ihour,&imin,&isec);
        jswitch = -1;
        geogsm_(&xgeo,&ygeo,&zgeo,&xgsm,&ygsm,&zgsm,&jswitch);
    
        rgeo = sqrt(xgeo*xgeo + ygeo*ygeo + zgeo*zgeo);
        tgeo = acos(zgeo/rgeo); 
        pgeo = atan2(ygeo,xgeo);
        igrf_(&iyr,&nIGRF,&rgeo,&tgeo,&pgeo,&brgeo,&btgeo,&bpgeo);
        bspcar_(&tgeo,&pgeo,&brgeo,&btgeo,&bpgeo,&bxgeo,&bygeo,&bzgeo);
      /*if (i < 30){
        printf("iyr:%i,iday:%i,ihour:%i,imin:%i,isec:%i\n",iyr,iday,ihour,imin,isec);
        printf("bx:%g,by:%g,bz:%g\n",bx,by,bz);
        printf("rgeo:%g,tgeo:%g,pgeo:%g\n",rgeo,tgeo,pgeo);
        printf("bxgeo:%g,bygeo:%g,bzgeo:%g\n",brgeo,btgeo,bpgeo);
      }*/
        jswitch = 1;
        geogsm_(&bxgeo,&bygeo,&bzgeo,&xgsm,&ygsm,&zgsm,&jswitch);

        bx+=xgsm;
        by+=ygsm;
        bz+=zgsm;
    }
    tsyganenko_out->dbl(i,0) = (double) bx;
    tsyganenko_out->dbl(i,1) = (double) by;
    tsyganenko_out->dbl(i,2) = (double) bz;
  }

//	...set some attributes

  tsyganenko_out->change_xref(DEPEND_0, vec.t);
  
  tsyganenko_out->change_xref("SI_conversion", "1.0e-9>T");
  sprintf(str_out, "B_xyz_t96");
  tsyganenko_out->change_xref("LABLAXIS", str_out);
  sprintf(str_out, "B_xyz Tsyganenko96");
  tsyganenko_out->change_xref("FIELDNAM", str_out);
  tsyganenko_out->change_xref("UNITS", "nT");
  
  // now rotate back into original frame if necessary
  if (rotateset) {
    QplugAppendTextDisplay ("tsyganenko>>  Rotating into original frame.\n");
    vec.n_recs = Ndata;
    vec.ptr = tsyganenko_out;
    vec.out_frame = vec.in_frame; 
    vec.QrSetFrameAttr(GSM, tsyganenko_out);
    vec.in_frame = GSM;
    vec.IsPositionVector = false;  // this is to avoid rejection
    check = vec.QrDoRotation();

    if (check > QRTN_OK) {
        QplugAppendTextDisplay ("tsyganenko>> ERROR rotating into GSM:\n ");
        QplugAppendTextDisplay (vec.QrErrorMessage(check));
        return QPLUG_FAILURE; 
    }
    tsyganenko_out = vec.ptr;
  } 
  else vec.QrSetFrameAttr(GSM, tsyganenko_out);


    QplugAppendTextDisplay ("tsyganenko>> finished...\n");

    //	..add them to the list

    outputs.push_back(tsyganenko_out);     

    return QPLUG_SUCCESS;

}






