// qpui_page.cc
//
// Plot Interface Implementation - Page Object
//
// Janet Barnes
//
// July 2002
//
// Tony Allen, Jan 2014

#include <stdlib.h>

#include "qpui.h"


//     PageItem CLASS
//     --------------

// static members
double PageItem::_width_in_inches[6]={8.3,8.5,7.5,8.5,8.5,0.};
double PageItem::_height_in_inches[6]={11.694,9,10.,14,11,0.};

PageItem::PageItem(const QString& name, PlotList* parent, int position) : QpuiItem(name, parent)
{
	if(parent==NULL)
	{
		std::cerr << "PageItem::PageItem: creating PageItem with NULL parent" << std::endl;
		return;
	}
	
		
	tabWidget=parent->tabWidget;
	
	// Constructor
	I_Type = QP_PAGE;
	_print_id = 0;

	if(name.isEmpty())
	{
		setNextName();
	}
	
	// initialise page tab widgets
	PageTab = new QpPageTab(tabWidget, "PageTab", this);
	PageTab->hide();
	
	LayoutTab = new QpLayoutTab(tabWidget, "LayoutTab", this);
	LayoutTab->hide();
		
	ValueReturnTab = new QpValueReturnTab(tabWidget, "ValueReturnTab", this);
	ValueReturnTab->hide();

    _conversion_factors[1]=1;
    _conversion_factors[2]=2.54;
    _conversion_factors[3]=25.4;
    
	setIcon(QP_PAGE);
}

PageItem::PageItem(PageItem* page, PlotList* parent, int position) : QpuiItem("", parent)
{
	if(parent==NULL)
	{
		std::cerr << "PageItem::PageItem: creating PageItem with NULL parent" << std::endl;
		return;
	}
	
		
	tabWidget=parent->tabWidget;
	// Constructor
	I_Type = QP_PAGE;
	_print_id = page->_print_id;

	setNextName();
		
	// initialise page tab widgets
	PageTab = new QpPageTab(page->PageTab, this);
	PageTab->hide();
	

	LayoutTab = new QpLayoutTab(page->LayoutTab, this);
	LayoutTab->hide();
		
	ValueReturnTab = new QpValueReturnTab(page->ValueReturnTab, this);
	ValueReturnTab->hide();
	
	for(int i=0; i<page->childCount(); i++)
	{
		FrameItem* frameItem=dynamic_cast<FrameItem*>(page->child(i));
		if(frameItem!=NULL)
		{
			new FrameItem(frameItem, this);
		}
	}
    _conversion_factors[1]=1;
    _conversion_factors[2]=2.54;
    _conversion_factors[3]=25.4;
	
	setIcon(QP_PAGE);


}
//-------------

PageItem::~PageItem()
{
	delete ValueReturnTab;
	delete PageTab;
	delete LayoutTab;
}
//-------------

void PageItem::setName(const QString& newName)
{

	QpuiItem::setName(newName);
	
	if(PageTab->_xwin_num_fld->currentWindow){
#ifdef Q_OS_MACX
        PageTab->_xwin_num_fld->currentWindow->setTitle(QString("(double click on plot to launch interval editor)                     %1    %2   ").arg(getName()).arg(PageTab->_xwin_num_fld->getName().c_str()));
#else
        PageTab->_xwin_num_fld->currentWindow->setTitle(QString("(click on plot to launch interval editor)                      %1    %2   ").arg(getName()).arg(PageTab->_xwin_num_fld->getName().c_str()));
#endif
    }
	
}


// ----------------------------------

void PageItem::showTabs()
{
	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--) 
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}

	tabWidget->addTab(PageTab, "Page");
    
	tabWidget->addTab(LayoutTab, "Layout");
	tabWidget->addTab(ValueReturnTab, "Value Return");

 	if(this->Tree()->I_LastTab == QP_LAYOUT_TAB)  tabWidget->setCurrentIndex(tabWidget->indexOf(LayoutTab)); 
 	else	tabWidget->setCurrentIndex(tabWidget->indexOf(PageTab));

}

// ----------------------------------

void PageItem::setDefaultOptions()
{
	
	PageTab->_page_size_fld->setCurrentIndex(QPUI_A4);
	PageTab->_units_fld->setCurrentIndex(QPUI_PIXELS);
	PageTab->_orientation_fld->setCurrentIndex(QPUI_PORTRAIT);
	PageTab->_show_ident_fld->setChecked(true);
	PageTab->_xwin_num_fld->setCurrentIndex(PageTab->_xwin_num_fld->size()); // new window is last item, size is number of actual windows
		
	ValueReturnTab->_xaxis_rescale_fld->setChecked(true);
	ValueReturnTab->_xaxis_swap_fld->setChecked(true);
	ValueReturnTab->_yaxis_rescale_fld->setChecked(true);
	ValueReturnTab->_yaxis_swap_fld->setChecked(true);
	ValueReturnTab->_replot_fld->setChecked(true);

	LayoutTab->_autolayout_fld->setChecked(true);
	LayoutTab->_hgap_fld->setValue(QPUI_PAGE_HGAP);
	LayoutTab->_vgap_fld->setValue(QPUI_PAGE_VGAP);
	LayoutTab->_lmargin_fld->setValue(QPUI_PAGE_LMARGIN);
	LayoutTab->_rmargin_fld->setValue(QPUI_PAGE_RMARGIN);
	LayoutTab->_tmargin_fld->setValue(QPUI_PAGE_TMARGIN);
	LayoutTab->_bmargin_fld->setValue(QPUI_PAGE_BMARGIN);
}

double PageItem::aspectRatio() const
{
    double page_width=PageTab->_width_fld->text().toDouble();
    double page_height=PageTab->_height_fld->text().toDouble();
    double page_ratio=sqrt(2.);

    if(page_width>0. && page_height>0.)
    {
        page_ratio=page_height/page_width;
    }
    if(PageTab->_orientation_fld->currentIndex()==QPUI_PORTRAIT)
    {
        page_ratio=1./page_ratio;
    }

    return page_ratio;
}




// ----------------------------------

void PageItem::setSmooth(bool smooth)
{
	PlotWindow* _plotWindow = PageTab->_xwin_num_fld->currentWindow;
	if(_plotWindow == 0)  return;

	_plotWindow->setSmooth(smooth);
	
}


// ----------------------------------

PlotWindow* PageItem::focusWindow()
{
    
   double width, height;
    double conv;
    width = PageTab->_width_fld->text().toDouble();
    height =PageTab->_height_fld->text().toDouble();
    
    int _units = PageTab->_units_fld->currentIndex();
    conv=_conversion_factors[_units];
    width=width/conv; //conversion to inches
    height=height/conv; //conversion to inches
    
    conv=_conversion_factors[0]; // conv factor to pixels
    width=width*conv; //px
    height=height*conv; //px
    
    if(width<=0. || height<=0.)
    {
        width=800.;
        height=600.;
    }
    QPrinter::Orientation orientation=(PageTab->_orientation_fld->currentIndex() == QPUI_LANDSCAPE ? QPrinter::Landscape : QPrinter::Portrait);

    QPrinter::PageSize pageSize=QPrinter::A4;
    int resolution=72;
    switch(PageTab->_page_size_fld->currentIndex())
    {
        case QPUI_EXECUTIVE:
            pageSize=QPrinter::Executive;
            break;
        case QPUI_LEGAL:
            pageSize=QPrinter::Legal;
            break;
        case QPUI_LETTER:
            pageSize=QPrinter::Letter;
            break;
        case QPUI_CUSTOM:
            pageSize=QPrinter::Custom;
            break;
        default: break;
    }
    
    switch(PageTab->_dpi_fld->currentIndex())
    {
        case QPUI_150_DPI: resolution=150; break;
        case QPUI_300_DPI: resolution=300; break;
        case QPUI_600_DPI: resolution=600; break;
        case QPUI_1200_DPI: resolution=1200; break;
        default: resolution=72; break;
    }

    PlotWindow *win = PageTab->_xwin_num_fld->focusWindow(width, height, resolution, pageSize, orientation);

#ifdef Q_OS_MACX
	win->setTitle(QString("(double click on plot to launch interval editor)                     %1    %2   ").arg(getName()).arg(PageTab->_xwin_num_fld->getName().c_str()));
#else
	win->setTitle(QString("(click on plot to launch interval editor)                      %1    %2   ").arg(getName()).arg(PageTab->_xwin_num_fld->getName().c_str()));
#endif
	win->show();

	return win;
}


// ----------------------------------

QpuiReturnStatus PageItem::plot(PlotWindow* win)
{

    if(QuSlotDebug())
	{
		printf("QpPageItem::plot\n");
		fflush(stdout);
	}
	if(!win) return QPUI_FAILURE;
	
	win->setPLFocus(); // just to be safe
	
	static PLFLT default_xwidth=0.;
	static PLFLT default_xheight=0.;
	
	plspause(0);
	
	// To make sure the new width will be validated:
	// plwid only stores changes in width
	// If no change from the previous plot, the default
	// width is not stored in the device buffer.
	// Thus we force a change, to store it.
	plwid(QtWidth2PlWidth(10));
	
	pladv(0);
	plvpor(0., 1., 0., 1.);

	// set background and default colour
	plscol0(0,255,255,255); // background temporarily set to white
	plscol0(1,0,0,0); // default set to black
	
	plwid(QtWidth2PlWidth(1));
	
	// clear colour table 
	
	getPGPColourIndex(QColor(0,0,0), true);
	
	win->setSmooth(true);
	
	// now handle each frame within this page
	
	if(this->hasChildren())
	{
		FrameItem* frameItem;
		for(int i=0; i<this->childCount(); ++i)
		{
			frameItem = dynamic_cast<FrameItem*>(this->child(i));
			if(QPUI_FAILURE == frameItem->plot(win))
			{
				QuMsgBar(("Problem plotting frame "+frameItem->getName()).toLatin1().data(),true);
			}
		}
	}

	if(PageTab->_show_ident_fld->isChecked()) print_iden();
	
	qApp->processEvents();

	return QPUI_SUCCESS; 
}

// ---------------

void PageItem::set_colour_table(const QpuiColourTable& colour_table)
{
//     PLFLT *h, *s, *l, *i;
//     QString str;
//     QStringList sl, sl2;
//     int npoints;
//     
//     switch(colour_table)
//     {
// 
//         case QPUI_RAINBOW_EXT:
//             h=new PLFLT[4]; h[0]=300.; h[1]=300.; h[2]=-10.; h[3]=-10.;
//             s=new PLFLT[4]; s[0]=1.; s[1]=1.; s[2]=1.; s[3]=1.;
//             l=new PLFLT[4]; l[0]=1.; l[1]=.5; l[2]=.5; l[3]=0.;
//             
//             i=new PLFLT[4]; i[0]=0.; i[1]=0.1; i[2]=0.9; i[3]=1.;
//             
//             plscmap1l(0, 4, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             break;
//             
//         case QPUI_GRAY_WHITE_HIGH:
//             h=new PLFLT[2]; h[0]=0.; h[1]=0.;
//             s=new PLFLT[2]; s[0]=0.; s[1]=0.;
//             l=new PLFLT[2]; l[0]=0.; l[1]=1.;
//             
//             i=new PLFLT[2]; i[0]=0.; i[1]=1.;
//             
//             plscmap1l(0, 2, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
//             break;
//             
//         case QPUI_GRAY_WHITE_LOW:
//             h=new PLFLT[2]; h[0]=0.; h[1]=0.;
//             s=new PLFLT[2]; s[0]=0.; s[1]=0.;
//             l=new PLFLT[2]; l[0]=1.; l[1]=0.;
//             
//             i=new PLFLT[2]; i[0]=0.; i[1]=1.;
//             
//             plscmap1l(0, 2, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
//             break;
//             
//         case QPUI_INV_RED_BLUE:
//             h=new PLFLT[4]; h[0]=230.; h[1]=230.; h[2]=-10.; h[3]=-10.;
//             s=new PLFLT[4]; s[0]=1.; s[1]=1.; s[2]=1.; s[3]=1.;
//             l=new PLFLT[4]; l[0]=.9; l[1]=.2; l[2]=.2; l[3]=.9;
//             
//             i=new PLFLT[4]; i[0]=0.; i[1]=0.499; i[2]=0.5; i[3]=1.;
//             
//             plscmap1l(0, 4, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
// 
//             break;
//             
//         case QPUI_RED_BLUE:
//             h=new PLFLT[4]; h[0]=230.; h[1]=230.; h[2]=-10.; h[3]=-10.;
//             s=new PLFLT[4]; s[0]=1.; s[1]=1.; s[2]=1.; s[3]=1.;
//             l=new PLFLT[4]; l[0]=.25; l[1]=.9; l[2]=.9; l[3]=.25;
//             
//             i=new PLFLT[4]; i[0]=0.; i[1]=0.499; i[2]=0.5; i[3]=1.;
//             
//             plscmap1l(0, 4, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
//             
//             break;
//             
//         case QPUI_EARTH:
//             h=new PLFLT[2]; h[0]=50.; h[1]=50.;
//             s=new PLFLT[2]; s[0]=0.2; s[1]=1.;
//             l=new PLFLT[2]; l[0]=0.9; l[1]=0.;
//             
//             i=new PLFLT[2]; i[0]=0.; i[1]=1.;
//             
//             plscmap1l(0, 2, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
//             break;
// 	    
// 	    case QPUI_USER_TABLE:
// 		    str=QSAS_profiles->current()->ColourTable;
// 		    str=str.trimmed();
// 		    bool isRGB;
// 		    if(str.isEmpty())
// 		    {
// 			    set_colour_table(QPUI_RAINBOW);
// 			    return;
// 		    }
// 
// 		    sl=str.split(QRegExp("\\(|\\)"), QString::SkipEmptyParts);
// 		    if(sl.size()<3)
// 		    {
// 			    set_colour_table(QPUI_RAINBOW);
// 			    return;
// 		    }
// 		    if(sl[0].contains("RGB", Qt::CaseInsensitive))
// 		    {
// 			    isRGB=true;
// 		    }
// 		    else if(sl[0].contains("HSV", Qt::CaseInsensitive))
// 		    {
// 			    isRGB=false;
// 		    }
// 
// 		    else
// 		    {
// 			    set_colour_table(QPUI_RAINBOW);
// 			    return;
// 		    }
// 		    npoints=sl.size()-1;
// 		    h=new PLFLT[npoints];
// 		    s=new PLFLT[npoints];
// 		    l=new PLFLT[npoints];
// 		    i=new PLFLT[npoints];
// 		    for(int index=0; index<npoints; ++index)
// 		    {
// 			    sl2=(sl[index+1]).split(",",  QString::SkipEmptyParts, Qt::CaseInsensitive);
// 
// 			    if(sl2.size()!=4)
// 			    {
// 				h[index]=0.; s[index]=0.; l[index]=0.; i[index]=0.;
// 			    }
// 			    else
// 			    {
// 				i[index]=sl2[0].toDouble();
// 				h[index]=sl2[1].toDouble();
// 				if(isRGB)
// 				{
// 					l[index]=sl2[2].toDouble();
// 					s[index]=sl2[3].toDouble();
// 				}
// 				else
// 				{
// 					s[index]=sl2[2].toDouble();
// 					l[index]=sl2[3].toDouble();
// 				}
// 			    }
// 		    }
// 		    plscmap1l(isRGB, npoints, i, h, l, s, NULL);
//             
// 		    delete[] h;
// 		    delete[] s;
// 		    delete[] l;
// 		    delete[] i;
// 		    break;
// 	    
// 	case QPUI_RAINBOW:
// 	    default:
// 		    h=new PLFLT[2]; h[0]=300.; h[1]=-10.;
// 		    s=new PLFLT[2]; s[0]=1.; s[1]=1.;
// 		    l=new PLFLT[2]; l[0]=.5; l[1]=.5;
//             
// 		    i=new PLFLT[2]; i[0]=0.; i[1]=1.;
//             
// 		    plscmap1l(0, 2, i, h, l, s, NULL);
//             
// 		    delete[] h;
// 		    delete[] s;
// 		    delete[] l;
// 		    delete[] i;
//             
// 		    break;
//     }
}

// -----------------


void PageItem::updateIsotropicFrames()
{
    // update the sizes of the frames with isotropic or square ratios
    if(this->hasChildren())
    {
        FrameItem* frameItem;
        for(int i=0; i<this->childCount(); ++i)
        {
           frameItem=dynamic_cast<FrameItem*>(this->child(i));

            double page_ratio=aspectRatio();
            
            if(frameItem->PropertiesTab->_aspect_ratio_fld->currentIndex()==QPUI_SQUARE_ASPECT)
            {
                frameItem->PropertiesTab->_height_fld->setValue(frameItem->PropertiesTab->_width_fld->value()*page_ratio);
            } 
            else if(frameItem->PropertiesTab->_aspect_ratio_fld->currentIndex()==QPUI_ISOTROPIC)
            {
				frameItem->updateIsotropic();
            }
        }
    }
}

void PageItem::setLayout()
{
	// does the auto-layout for a page. Ignored for User layout.
	// update layout for page
	if(QuSlotDebug())
	{
		printf("PageItem::setLayout \n");
		fflush(stdout);
	}

	if( !(LayoutTab->_autolayout_fld->isChecked()) ) return;
	
	double _tmargin = LayoutTab->_tmargin_fld->value();
	double _lmargin = LayoutTab->_lmargin_fld->value();
	double _rmargin = LayoutTab->_rmargin_fld->value();
	double _bmargin = LayoutTab->_bmargin_fld->value();
	double _hgap = LayoutTab->_hgap_fld->value();
	double _vgap = LayoutTab->_vgap_fld->value();
    double _cbar = 0.;
    

	// calculate number of rows and columns
	int num_rows = LayoutTab->_row_num_fld->value();
	int num_cols = LayoutTab->_col_num_fld->value();
	// get widths of panels 
	float frameWidth = 0.98 - (_lmargin + _rmargin + _hgap*(num_cols - 1) );
	frameWidth /= num_cols;
    
    int cnt = 0;
	int col = 0;
	int row = 0;
	float height = 0.;
	float width = 0.;
	
	// calculate left offest and top offset for frames in page
	
	double frameLoffset = _lmargin;
	double frameToffset = _tmargin;


    
	FrameItem* frameItem;
    
	for(int i=0; i<this->childCount(); ++i)
	{
		frameItem=dynamic_cast<FrameItem*>(this->child(i));
		// fetch number of panels in this frame
		int numPanels = frameItem->numberInStack();
		numPanels = (numPanels>1)?numPanels:1;
			
		// calculate height of this frame so that all panels are equal height
		float frameHeight = 1. - (_tmargin + _bmargin + _vgap*(num_rows - 1));
		frameHeight /= panelsInColumn(col);
		frameHeight *= numPanels;
			
		// set panel loffset, width, toffset, height
        frameItem->setPanelProperties(frameLoffset, frameWidth, frameToffset, frameHeight, num_rows, num_cols);
			
		int rowLast = row;
		int colLast = col;
		cnt++;
		col = (int)cnt  / num_rows;
		row = cnt - col * num_rows;
			
		// add widths of stack panels to loffset
		if(col > colLast) frameLoffset += frameWidth + _hgap;
		else if(col == 0)  frameLoffset = _lmargin; // reset for next row
			
		// add height of stack panel to toffset
		if(row > rowLast) frameToffset += frameHeight + _vgap;
		else if(row ==0) frameToffset = _tmargin;
	}
    
}

// -----------------------------------
	
int PageItem::panelsInColumn(int colno)
{
	// number of panels for all frames in column
	if(QuSlotDebug())
	{
		printf("QpPageItem::panelsInColumn \n");
		fflush(stdout);
	}

  	int cnt = 0;
	int row = 0;
	int col = 0;
	int numPanels = 0;
	int num_rows =  LayoutTab->_row_num_fld->value();
	
	if(this->hasChildren())
	{
		FrameItem* frameItem;
		for(int i=0; i<this->childCount(); ++i)
		{
			frameItem = dynamic_cast<FrameItem*>(this->child(i));
			col = (int)cnt  / num_rows;
			row = cnt - col * num_rows;
			cnt++;
			if(colno == col) 
			{	
				// increment by number of panels in this frame
				numPanels += frameItem->childCount();
			}
		}
	}
	
	return (numPanels>0)?numPanels:1;
}


// -----------------------------------
	
void PageItem::setFramesAuto(bool active)
{

	FrameItem* frameItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		frameItem=dynamic_cast<FrameItem*>(this->child(i));
			
		if(active)
		{
			LayoutTab->updateRowCol(this->childCount());
			frameItem->PropertiesTab->_autolayout_lbl->setText("Autoset by Page Layout ");
		}
		else
		{
			frameItem->PropertiesTab->_autolayout_lbl->setText(" Page Layout not Autoset");
		}
		frameItem->PropertiesTab->_loffset_fld->setDisabled(active);
		frameItem->PropertiesTab->_toffset_fld->setDisabled(active);
		frameItem->PropertiesTab->_width_fld->setDisabled(active);
		frameItem->PropertiesTab->_height_fld->setDisabled(active);

	}
	
	LayoutTab->_hgap_fld->setEnabled(active);
	LayoutTab->_vgap_fld->setEnabled(active);
	LayoutTab->_lmargin_fld->setEnabled(active);
	LayoutTab->_rmargin_fld->setEnabled(active);
	LayoutTab->_tmargin_fld->setEnabled(active);
	LayoutTab->_bmargin_fld->setEnabled(active);
}
// -------------------------------

bool PageItem::containsSpectrum()
{
	if(this->hasChildren())
	{
		FrameItem* frameItem;
		for(int i=0; i<this->childCount(); ++i)
		{
			frameItem = dynamic_cast<FrameItem*>(this->child(i));
			if(frameItem->containsSpectrum()) return true;
		}
	}
	return false;
}


// -------------------------------

QDomElement PageItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Page_Item");
    xml.appendChild(createTextXMLNode(doc, "userSetName", userSetName));
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));
    xml.appendChild(PageTab->toXMLElement(doc));
    xml.appendChild(LayoutTab->toXMLElement(doc));

    int cnt = childCount();
    xml.setAttribute("size", QString::number(cnt));
   
    FrameItem* frameItem;
    for(int i=0; i<cnt; ++i)
    {
        frameItem=dynamic_cast<FrameItem*>(child(i));
        xml.appendChild(frameItem->toXMLElement(doc));
    }
    return xml;
}
// -------------------------------

void PageItem::fromXMLElement(QDomElement& el)
{
    QString s;
	
    QDomElement page_tab=el.firstChildElement("Page_Tab");
    if(!page_tab.isNull()) PageTab->fromXMLElement(page_tab);

    QDomElement layout_tab=el.firstChildElement("Layout_Tab");
    if(!layout_tab.isNull()) LayoutTab->fromXMLElement(layout_tab);

    int cnt=el.attribute("size", "0").toInt();

    QDomElement frameElement=el.firstChildElement("Frame_Item");

    while(!frameElement.isNull())
    {
        FrameItem *frameItem = new FrameItem("", this);
        frameItem->fromXMLElement(frameElement);

        frameElement=frameElement.nextSiblingElement("Frame_Item");
    }

    if(getXMLNodeText(layout_tab, "Num_Rows", s))
    {
        LayoutTab->_row_num_fld->setValue(s.toInt());
    }
    if(getXMLNodeText(layout_tab, "Num_Cols", s))
    {
        LayoutTab->_col_num_fld->setValue(s.toInt());
    }
	
	setFramesAuto(LayoutTab->_autolayout_fld->isChecked());
    
    if(getXMLNodeText(el, "Name", s) )
    {
		setName(s);
    }
    
    if(getXMLNodeText(el, "userSetName", s) )
    {
		userSetName = s.toInt();
    } 

}

